#include "DriverPropertyDialog.h"
#include "core/DriverSetupData.h"
#include "mylib/wtlext/textextent.h"

DriverPropertyView::DriverPropertyView(DriverSetupData const &driver_data)
: driver_data(driver_data)
{
}

BOOL DriverPropertyView::OnInitDialog(HWND /*wndFocus*/, LPARAM /*lInitParam*/)
{
	// 
	::SetWindowText(this->GetDlgItem(IDC_STATIC_PROVIDER),
		driver_data.getProviderName().c_str());
	// t
	uint32_t inf_date = this->driver_data.getInfFileDate();
	if (inf_date != 0) {
		std::wstringstream ss;
		ss.imbue(std::locale(""));
		tm t = {0};
		t.tm_year = (inf_date >> 16) - 1900;
		t.tm_mon = ((inf_date >> 8) & 0xFF) - 1;
		t.tm_mday = (inf_date & 0xFF);
		const std::time_put<TCHAR>& tp =
			std::use_facet<std::time_put<TCHAR> >(ss.getloc());
		LPCTSTR fmt = _T("%x");
		tp.put(ss, ss, ss.fill(),
			&t, fmt, fmt + 2);
		::SetWindowText(this->GetDlgItem(IDC_STATIC_DATE), ss.str().c_str());
	} else {
		::SetWindowText(this->GetDlgItem(IDC_STATIC_DATE), L"s");
	}
	// o[W
	uint64_t inf_ver = this->driver_data.getInfFileVersion();
	std::wstringstream ss;
	ss << ((inf_ver >> 48) & 0xFFFF) << _T('.') << ((inf_ver >> 32) & 0xFFFF) << _T('.') <<
		((inf_ver >> 16) & 0xFFFF) << _T('.') << (inf_ver & 0xFFFF);
	::SetWindowText(this->GetDlgItem(IDC_STATIC_VERSION), ss.str().c_str());
	// INF t@C̃pX
	CStatic inf_path_label = this->GetDlgItem(IDC_STATIC_INF_PATH);
	TextExtent inf_path_label_te(inf_path_label);
	boost::filesystem::wpath const &inf_path = this->driver_data.getInfFilePath();
	std::wstring const inf_path_str = inf_path.native_file_string();
	SIZE text_extent = inf_path_label_te.getTextExtent(inf_path_str);
	inf_path_label.SetWindowText(inf_path_str.c_str());
	int const inf_path_label_width = text_extent.cx;
	// Xg{bNX
	CListBox list_box = this->GetDlgItem(IDC_LIST1);
	int list_box_width = 0;
	int list_box_height = 0;
	TextExtent list_box_te(list_box);
	const DriverSetupData::DDInstallCollection& ddic = driver_data.getDDInstallSectionCollection();
	for (DriverSetupData::DDInstallCollection::const_iterator it = ddic.begin(); it != ddic.end(); ++it) {
		for (DriverSetupData::DescriptionCollection::const_iterator it2 = it->second.descriptions.begin(); it2 != it->second.descriptions.end(); ++it2) {
			SIZE text_extent = list_box_te.getTextExtent(*it2);
			if (text_extent.cx > list_box_width) {
				list_box_width = text_extent.cx;
			}
			list_box_height += text_extent.cy;
			list_box.AddString(it2->c_str());
		}
	}

	CRect rc_dlg;
	this->GetClientRect(rc_dlg);

	CRect rc_inf_path_label;
	inf_path_label.GetWindowRect(rc_inf_path_label);
	this->ScreenToClient(rc_inf_path_label);
	int const offset_inf_path = inf_path_label_width - rc_inf_path_label.Width();

	CRect rc_list;
	list_box.GetWindowRect(rc_list);
	this->ScreenToClient(rc_list);
	CRect rc_list_c;
	list_box.GetClientRect(rc_list_c);
	int const offset_list_box_w = list_box_width - rc_list_c.Width() + 5;
	int const offset_list_box_h = list_box_height - rc_list_c.Height() + 3;

	int offset = std::max(offset_inf_path, offset_list_box_w);
	if (offset > 0) {
		// _CAO̕
		rc_dlg.right += offset;
	}
	if (offset_inf_path > 0) {
		// uINFt@C̃pXv\̕
		rc_inf_path_label.right += offset_inf_path;
	}
	if (offset_list_box_w > 0) {
		// Xg{bNX̕
		rc_list.right += offset_list_box_w;
	}
	if (offset_list_box_h > 0) {
		rc_dlg.bottom += offset_list_box_h;
		rc_list.bottom += offset_list_box_h;
	}
	this->MoveWindow(rc_dlg, false);
	list_box.MoveWindow(rc_list, false);
	inf_path_label.MoveWindow(rc_inf_path_label, false);

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////

ScrollView::ScrollView(const DriverSetupData &driver_data)
: view(driver_data)
{
}

int ScrollView::OnCreate(CREATESTRUCT * /*create_struct*/)
{
	int const ret = static_cast<int>(this->DefWindowProc());

	this->view.Create(this->m_hWnd);
	//CRect rc1;
	//this->view.GetWindowRect(rc1);
	CRect rc;
	this->GetClientRect(rc);
	CRect rc_child;
	this->view.GetClientRect(rc_child);
	//this->view.ClientToScreen(rc_child);
	//this->view.ShowWindow(SW_SHOW);
	int const width = rc.Width();
	int const height = rc.Height();
	int const child_width = rc_child.Width();
	int const child_height = rc_child.Height();
	if (width >= child_width) {
		rc_child.right = rc.right;
	}
	if (height >= child_height) {
		rc_child.bottom = rc.bottom;
	}
	if (width >= child_width || height >= child_height) {
		this->view.MoveWindow(rc_child, false);
	}
	this->SetScrollSize(child_width, child_height);
	this->SetScrollLine(10, 10);
	this->SetScrollPage(100, 100);
	this->SetMsgHandled(false);

	return ret;
}

void ScrollView::DoPaint(CDCHandle dc)
{
}

/////////////////////////////////////////////////////////////////////////////

DriverPropertyDialog::DriverPropertyDialog(std::wstring const &title, DriverSetupData const &driver_data)
: title(title)
, view(driver_data)
{
}

BOOL DriverPropertyDialog::OnInitDialog(HWND /*wndFocus*/, LPARAM /*lInitParam*/)
{
	// ^Cgݒ
	::SetWindowText(this->m_hWnd, this->title.c_str());
	// hCovpeB\̈ʒuݒ
	CWindow frame = this->GetDlgItem(IDC_DRIVER_PROPERTY_PLACE);
	CRect rc;
	frame.GetWindowRect(rc);
	this->ScreenToClient(rc);
	this->view.Create(this->m_hWnd, rc, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN, /*WS_EX_CLIENTEDGE |*/ WS_EX_CONTROLPARENT);
	frame.SetWindowPos(HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	//this->view.Create(this->m_hWnd);

	//CRect view_rc;
	//this->view.GetWindowRect(view_rc);
	//int const view_offset = this->view.getOffset();
	//int const offset = view_rc.Width() + view_offset - rc.Width();
	//int offset2 = 0;
	//if (offset > 0) {
	//	CRect dlg_rect;
	//	this->GetWindowRect(dlg_rect);
	//	dlg_rect.right += offset;

	//	// Ώۃj^̏擾
	//	HMONITOR hMonitor = MonitorFromRect(
	//		&dlg_rect, MONITOR_DEFAULTTONEAREST);
	//	MONITORINFO mi;
	//	mi.cbSize = sizeof(mi);
	//	GetMonitorInfo(hMonitor, &mi);
	//	int const monitor_width = mi.rcMonitor.right - mi.rcMonitor.left;
	//	int const n = dlg_rect.Width() - monitor_width;
	//	if (n > 0) {
	//		offset2 = offset - n;
	//		dlg_rect.right -= n;
	//	} else {
	//		offset2 = offset;
	//	}
	//	// ʒu␳
	//	if (dlg_rect.right > mi.rcMonitor.right) {
	//		dlg_rect.left -= dlg_rect.right - mi.rcMonitor.right;
	//		dlg_rect.right = mi.rcMonitor.right;
	//	}
	//	if (dlg_rect.left < mi.rcMonitor.left) {
	//		dlg_rect.right += mi.rcMonitor.left - dlg_rect.left;
	//		dlg_rect.left = mi.rcMonitor.left;
	//	}

	//	this->MoveWindow(dlg_rect, false);
	//}
	//if (offset2 > 0) {
	//	// uv{^̈ʒu
	//	CButton close_button = this->GetDlgItem(IDOK);
	//	CRect close_button_rect;
	//	close_button.GetWindowRect(close_button_rect);
	//	this->ScreenToClient(close_button_rect);
	//	close_button_rect.OffsetRect(offset2, 0);
	//	close_button.MoveWindow(close_button_rect, false);
	//	// r[̕vZ
	//	rc.right += offset2;
	//}
	//this->view.MoveWindow(rc, false);
	//this->view.setWidth(view_offset);
	//this->view.ShowWindow(SW_SHOW);

	return TRUE;
}

void DriverPropertyDialog::OnCloseCmd(unsigned int /*uNotifyCode*/, int nID, HWND /*wndCtl*/)
{
	this->EndDialog(nID);
}

/////////////////////////////////////////////////////////////////////////////

DriverReplaceDialog::DriverReplaceDialog(DriverSetupData const &driver_data_exist, DriverSetupData const &driver_data_new)
: view_exist(driver_data_exist)
, view_new(driver_data_new)
{
}

BOOL DriverReplaceDialog::OnInitDialog(HWND, LPARAM)
{
	CWindow group_new = this->GetDlgItem(IDC_DRIVER_REPLACE_GROUP_NEW);
	CRect group_new_rect;
	group_new.GetWindowRect(group_new_rect);
	this->ScreenToClient(group_new_rect);
	group_new_rect.DeflateRect(10, 10);
	group_new_rect.top += 10;
	this->view_new.Create(this->m_hWnd,
	                      group_new_rect,
	                      NULL,
	                      WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
	                      /*WS_EX_CLIENTEDGE |*/ WS_EX_CONTROLPARENT);
	//this->view_new.Create(this->m_hWnd);

	//CRect view_new_rect;
	//view_new.GetWindowRect(view_new_rect);

	//CRect view_new_rect2 = group_new_rect;
	//view_new_rect2.DeflateRect(10, 10);
	//view_new_rect2.top += 10;
	//int const view_new_offset = this->view_new.getOffset();
	//int const offset_new = view_new_rect.Width() + view_new_offset - view_new_rect2.Width();


	CWindow group_exist = this->GetDlgItem(IDC_DRIVER_REPLACE_GROUP_EXIST);
	CRect group_exist_rect;
	group_exist.GetWindowRect(group_exist_rect);
	this->ScreenToClient(group_exist_rect);
	group_exist_rect.DeflateRect(10, 10);
	group_exist_rect.top += 10;
	this->view_exist.Create(this->m_hWnd,
	                        group_exist_rect,
	                        NULL,
	                        WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
	                        /*WS_EX_CLIENTEDGE |*/ WS_EX_CONTROLPARENT);

	group_new.SetWindowPos(HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	group_exist.SetWindowPos(HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	//this->view_exist.Create(this->m_hWnd);

	//CRect view_exist_rect;
	//view_exist.GetWindowRect(view_exist_rect);

	//CRect view_exist_rect2 = group_exist_rect;
	//view_exist_rect2.DeflateRect(10, 10);
	//view_exist_rect2.top += 10;
	//int const view_exist_offset = this->view_exist.getOffset();
	//int const offset_exist = view_exist_rect.Width() + view_exist_offset - view_exist_rect2.Width();


	//int const view_offset = std::max(view_new_offset, view_exist_offset);
	//int const offset = std::max(offset_new, offset_exist);
	//int offset2 = 0;
	//if (offset > 0) {
	//	CRect dlg_rect;
	//	this->GetWindowRect(dlg_rect);
	//	dlg_rect.right += offset;

	//	// Ώۃj^̏擾
	//	HMONITOR hMonitor = MonitorFromRect(
	//		&dlg_rect, MONITOR_DEFAULTTONEAREST);
	//	MONITORINFO mi;
	//	mi.cbSize = sizeof(mi);
	//	GetMonitorInfo(hMonitor, &mi);
	//	int const monitor_width = mi.rcMonitor.right - mi.rcMonitor.left;
	//	int const n = dlg_rect.Width() - monitor_width;
	//	if (n > 0) {
	//		offset2 = offset - n;
	//		dlg_rect.right -= n;
	//	} else {
	//		offset2 = offset;
	//	}
	//	// ʒu␳
	//	if (dlg_rect.right > mi.rcMonitor.right) {
	//		dlg_rect.left -= dlg_rect.right - mi.rcMonitor.right;
	//		dlg_rect.right = mi.rcMonitor.right;
	//	}
	//	if (dlg_rect.left < mi.rcMonitor.left) {
	//		dlg_rect.right += mi.rcMonitor.left - dlg_rect.left;
	//		dlg_rect.left = mi.rcMonitor.left;
	//	}

	//	this->MoveWindow(dlg_rect, false);
	//}
	//if (offset2 > 0) {
	//	// 
	//	group_new_rect.right += offset2;
	//	group_new.MoveWindow(group_new_rect, false);
	//	// 
	//	group_exist_rect.right += offset2;
	//	group_exist.MoveWindow(group_exist_rect, false);
	//	// uOKv{^̈ʒu
	//	CButton ok_button = this->GetDlgItem(IDOK);
	//	CRect ok_button_rect;
	//	ok_button.GetWindowRect(ok_button_rect);
	//	this->ScreenToClient(ok_button_rect);
	//	ok_button_rect.OffsetRect(offset2, 0);
	//	ok_button.MoveWindow(ok_button_rect, false);
	//	// uLZv{^̈ʒu
	//	CButton cancel_button = this->GetDlgItem(IDCANCEL);
	//	CRect cancel_button_rect;
	//	cancel_button.GetWindowRect(cancel_button_rect);
	//	this->ScreenToClient(cancel_button_rect);
	//	cancel_button_rect.OffsetRect(offset2, 0);
	//	cancel_button.MoveWindow(cancel_button_rect, false);

	//	view_new_rect2.right += offset2;
	//	view_exist_rect2.right += offset2;
	//}

	//this->view_new.MoveWindow(view_new_rect2, false);
	//this->view_new.setWidth(view_offset);
	//this->view_new.ShowWindow(SW_SHOW);

	//this->view_exist.MoveWindow(view_exist_rect2, false);
	//this->view_exist.setWidth(view_offset);
	//this->view_exist.ShowWindow(SW_SHOW);

	return TRUE;
}

void DriverReplaceDialog::OnCloseCmd(unsigned int /*uNotifyCode*/, int nID, HWND /*wndCtl*/)
{
	this->EndDialog(nID);
}
