﻿#pragma once

#include "resource.h"
#include "DriversListViewCtrl.h"
#include "core/DriverSetupData.h"
#include "core/DriverFileData.h"
#include "core/DriverServiceData.h"


class SourceSetting;

struct DriverSetting
{
	std::vector<DriverSetupData> add_drivers;
	DriverFileDataMap leave_drivers;
	DriverFileDataMap del_drivers;
	DriverServiceData leave_services;
	DriverServiceData del_services;
};

class DriverSettingPage :
	public CDialogImpl<DriverSettingPage>,
	public CDialogResize<DriverSettingPage>,
	public CCustomDraw<DriverSettingPage>
{
	typedef DriverSettingPage ThisType;

	int architecture_id;
	uint32_t os_version;
	//CButton installed_driver_button;
	CButton add_button;
	CButton remove_button;
	CButton property_button;
	DriversListViewCtrl drivers_list;
	boost::filesystem::wpath drivers_dir_path_;

public:
	enum { IDD = IDD_DRIVER_SETTING_PAGE };
	//enum {
	//	UWM_DRIVER_LIST_IMPORT = WM_APP + 0x20,
	//	UWM_DRIVER_LIST_EXPORT
	//};
	enum {
		UWM_POPUP_MENU_SELECT = WM_APP + 0x20
	};

	DriverSettingPage();
	DriverSetting* getSetting();
	void loadSourceSetting(SourceSetting const *source_setting);

	DWORD OnPrePaint(int nID, NMCUSTOMDRAW *lpnmcd);
	DWORD OnItemPrePaint(int nID, NMCUSTOMDRAW *lpnmcd);
	DWORD OnItemPostPaint(int nID, NMCUSTOMDRAW *lpnmcd);

	BEGIN_DLGRESIZE_MAP(ThisType)
		DLGRESIZE_CONTROL(IDC_BUTTON_REMOVE, DLSZ_MOVE_Y)
		DLGRESIZE_CONTROL(IDC_BUTTON_PROPERTY, DLSZ_MOVE_Y)
		DLGRESIZE_CONTROL(IDC_BUTTON_EXPORT, DLSZ_MOVE_X | DLSZ_MOVE_Y)
		DLGRESIZE_CONTROL(IDC_LIST1, DLSZ_SIZE_X | DLSZ_SIZE_Y)
	END_DLGRESIZE_MAP()

	BEGIN_MSG_MAP_EX(ThisType)
		MSG_WM_INITDIALOG(OnInitDialog)
		MSG_WM_CONTEXTMENU(OnContextMenu)
		COMMAND_HANDLER_EX(IDC_BUTTON_ADD, BN_CLICKED, OnButtonAdd)
		COMMAND_ID_HANDLER_EX(IDM_ADD_DRIVERS_INF, OnAddDriversInf)
		COMMAND_ID_HANDLER_EX(IDM_ADD_DRIVERS_FOLDER, OnAddDriversFolder)
		COMMAND_ID_HANDLER_EX(IDM_ADD_DRIVERS_INSTALLED, OnAddDriversInstalled)
		COMMAND_ID_HANDLER_EX(IDM_ADD_DRIVERS_IMPORT, OnAddDriversImport)
		//COMMAND_HANDLER_EX(IDC_BUTTON_FILE, BN_CLICKED, OnButtonFile)
		//COMMAND_HANDLER_EX(IDC_BUTTON_FOLDER, BN_CLICKED, OnButtonFolder)
		//COMMAND_HANDLER_EX(IDC_BUTTON_INSTALLED_DRIVER, BN_CLICKED, OnButtonInstalledDriver)
		COMMAND_HANDLER_EX(IDC_BUTTON_EXPORT, BN_CLICKED, OnButtonExport)
		COMMAND_HANDLER_EX(IDC_BUTTON_REMOVE, BN_CLICKED, OnButtonRemove)
		COMMAND_HANDLER_EX(IDC_BUTTON_PROPERTY, BN_CLICKED, OnButtonProperty)
		COMMAND_ID_HANDLER_EX(IDM_DRIVERS_LIST_REMOVE, OnDriversListRemove)
		COMMAND_ID_HANDLER_EX(IDM_DRIVERS_LIST_PROPERTY, OnDriversListProperty)
		NOTIFY_CODE_HANDLER_EX(LVN_GETINFOTIP, OnGetInfoTip)
		NOTIFY_CODE_HANDLER_EX(LVN_DELETEITEM, OnDeleteItem)
		NOTIFY_CODE_HANDLER_EX(LVN_ITEMCHANGED, OnItemChanged)
		MESSAGE_HANDLER_EX(WM_MENUSELECT, OnMenuSelect)
		//MESSAGE_HANDLER_EX(UWM_DRIVER_LIST_IMPORT, OnDriverListImport)
		//MESSAGE_HANDLER_EX(UWM_DRIVER_LIST_EXPORT, OnDriverListExport)
		CHAIN_MSG_MAP(CCustomDraw<ThisType>)
		CHAIN_MSG_MAP(CDialogResize<ThisType>)
	END_MSG_MAP_EX()

	BOOL OnInitDialog(HWND /*wndFocus*/, LPARAM /*lInitParam*/);
	void OnContextMenu(CWindow /*wnd*/, CPoint /*point*/);
	void OnButtonAdd(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	void OnAddDriversInf(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	void OnAddDriversFolder(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	void OnAddDriversInstalled(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	void OnAddDriversImport(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	//void OnButtonFile(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	//void OnButtonFolder(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	//void OnButtonInstalledDriver(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	void OnButtonExport(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	void OnButtonRemove(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	void OnButtonProperty(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	void OnDriversListRemove(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	void OnDriversListProperty(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/);
	LRESULT OnGetInfoTip(NMHDR * /*pnmh*/);
	LRESULT OnDeleteItem(NMHDR * /*pnmh*/);
	LRESULT OnItemChanged(NMHDR * /*pnmh*/);
	LRESULT OnMenuSelect(unsigned int /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/);
	//LRESULT OnDriverListImport(unsigned int /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/);
	//LRESULT OnDriverListExport(unsigned int /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/);
};

