#include "DriversListViewCtrl.h"
#include "DriverPropertyDialog.h"
#include "core/DriverSetupData.h"

namespace fs {
using namespace boost::filesystem;
};


ItemAssociatedData::ItemAssociatedData(ItemState item_state, DriverSetupData const &item_data)
: item_state(item_state)
, item_data_ptr(new DriverSetupData(item_data))
{
}
ItemAssociatedData::~ItemAssociatedData()
{
	delete this->item_data_ptr;
}
void ItemAssociatedData::setItemState(ItemState item_state)
{
	this->item_state = item_state;
}

ItemAssociatedData::ItemState ItemAssociatedData::getItemState() const
{
	return this->item_state;
}

DriverSetupData* ItemAssociatedData::getItemData() const
{
	return this->item_data_ptr;
}

/////////////////////////////////////////////////////////////////////////////

DriversListViewCtrl::DriversListViewCtrl()
: architecture_id(-1)
, os_version(0)
{
}

DriversListViewCtrl::~DriversListViewCtrl()
{
}

void DriversListViewCtrl::setParameter(int architecture_id, uint32_t os_version)
{
	this->architecture_id = architecture_id;
	this->os_version = os_version;
}

void DriversListViewCtrl::setDeviceClassImageList()
{
	// foCXNX̃C[WXg擾
	this->image_list_data.cbSize = sizeof(SP_CLASSIMAGELIST_DATA);
	::SetupDiGetClassImageList(&image_list_data);
	this->SetImageList(image_list_data.ImageList, LVSIL_SMALL);
}

DriversListViewCtrl::ItemAdditionResult
DriversListViewCtrl::checkDuplication(DriverSetupData const &driver_data)
{
	ItemAdditionResult result = ITEM_ADDED;
	int item_count = this->GetItemCount();
	for (int i = 0; i < item_count; ++i) {
		ItemAssociatedData *item_data = reinterpret_cast<ItemAssociatedData*>(this->GetItemData(i));
		ItemAssociatedData::ItemState item_state = item_data->getItemState();
		if (item_state != ItemAssociatedData::ADDED_BEFORE_REMOVE) {
			DriverSetupData *ptr = item_data->getItemData();
			fs::wpath const new_inf_path = driver_data.getInfFilePath();
			fs::wpath const existing_inf_path = ptr->getInfFilePath();
			if (MyLib::compareStringNoCase(new_inf_path.string(), existing_inf_path.string()) == 0) {
				return NO_CHANGE;
			}
			if (driver_data.isForSameDevice(*ptr)) {
				// t@Ceꂩ`FbN
				fs::ifstream new_inf(new_inf_path, std::ios::in | std::ios::binary);
				fs::ifstream existing_inf(existing_inf_path, std::ios::in | std::ios::binary);
				if (new_inf && existing_inf) {
					size_t const new_inf_size = static_cast<size_t>(fs::file_size(new_inf_path));
					size_t const existing_inf_size = static_cast<size_t>(fs::file_size(existing_inf_path));
					if (new_inf_size == existing_inf_size) {
						boost::scoped_array<char> new_inf_buf(new char[new_inf_size]);
						boost::scoped_array<char> existing_inf_buf(new char[existing_inf_size]);
						new_inf.read(new_inf_buf.get(), new_inf_size);
						existing_inf.read(existing_inf_buf.get(), existing_inf_size);
						if (std::memcmp(new_inf_buf.get(), existing_inf_buf.get(), new_inf_size) == 0) {
							return NO_CHANGE;
						}
					}
				}
				DriverReplaceDialog dlg(*ptr, driver_data);
				if (dlg.DoModal() == IDOK) {
					// ǉhCo̕Vꍇ
					ItemAssociatedData::ItemState item_state = item_data->getItemState();
					if (item_state == ItemAssociatedData::ADDED_BEFORE) {
						item_data->setItemState(ItemAssociatedData::ADDED_BEFORE_REMOVE);
					} else if (item_state == ItemAssociatedData::ADDED_BEFORE_REMOVE) {
						item_data->setItemState(ItemAssociatedData::ADDED_BEFORE);
					} else {
						this->DeleteItem(i);
						--item_count;
						--i;
					}
					result = ITEM_REPLACED;
				} else {
					return NO_CHANGE;
				}
			}
		}
	}
	return result;
}

DriversListViewCtrl::ItemAdditionResult
DriversListViewCtrl::addDriversListItem(DriverSetupData const &driver_data, wchar_t const *str)
{
	DriverSetupData::DDInstallCollection const &ddic = driver_data.getDDInstallSectionCollection();
	DriverSetupData::DDInstallCollection::const_iterator it = ddic.begin();
	// DDInstallZNV  OSp
	if (it == ddic.end()) {
		return NO_CHANGE;
	}
	// d`FbN
	ItemAdditionResult result = this->checkDuplication(driver_data);
	if (result != NO_CHANGE) {
		// foCX̎ނ̕\擾
		wchar_t class_name[256];
		if (!::SetupDiGetClassDescription(&driver_data.getClassUUID(), class_name, 256, NULL)) {
			class_name[0] = L'\0';
		}
		// foCX̃ACR擾
		int nnn;
		if (!::SetupDiGetClassImageIndex(&image_list_data, &driver_data.getClassUUID(), &nnn)) {
			nnn = 0;
		}
		// ɒǉ
		int const insert_pos = this->GetItemCount();
		// ڂ̒ǉ
		DriverSetupData::DescriptionCollection::const_iterator it2 =
			it->second.descriptions.begin();
		this->InsertItem(insert_pos,
						 (it2 != it->second.descriptions.end()) ? it2->c_str() : L"No Description",
						 nnn);
		// foCX̎
		this->SetItemText(insert_pos, 1, class_name);
		// INFt@C̃pX
		this->SetItemText(insert_pos, 2, str ? str : driver_data.getInfFilePath().parent_path().native_directory_string().c_str());
		// ڂƁAhCõf[^֘At
		ItemAssociatedData *ptr = new ItemAssociatedData(ItemAssociatedData::NEWLY_ADDED, driver_data);
		this->SetItemData(insert_pos, reinterpret_cast<uintptr_t>(ptr));
		// ǉꂽڂ悤ɃXN[
		this->EnsureVisible(insert_pos, FALSE);
	}
	return result;
}

DriversListViewCtrl::ItemAdditionResult
DriversListViewCtrl::addDriversListItemFromInfFile(const boost::filesystem::wpath &filename,
                                                   int target_architecture,
                                                   boost::uint32_t target_version, wchar_t const *str)
{
	ItemAdditionResult result = NO_CHANGE;
#ifdef NDEBUG
	std::wostringstream ss;
	try {
		try {
#endif
			DriverSetupData data(filename.string(), target_architecture, target_version);
			result = this->addDriversListItem(data, str);
#ifdef NDEBUG
		} catch (...) {
			ss << filename.string() << L" ̃G[\n";
			throw;
		}
	} catch (MyLib::FileTransferException<wchar_t>& e) {
		ss << L"From: " << e.getSourceFileName() << L"\n";
		ss << L"To: " << e.getDestinationFileName() << L"\n";
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText();
		this->MessageBox(ss.str().c_str(), NULL, MB_ICONERROR);
	} catch (MyLib::FileException<wchar_t>& e) {
		ss << e.getFileName() << L"\n";
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText();
		this->MessageBox(ss.str().c_str(), NULL, MB_ICONERROR);
	} catch (MyLib::Exception<wchar_t>& e) {
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText() << L'[' << e.getErrorCode() << L']';
		this->MessageBox(ss.str().c_str(), NULL, MB_ICONERROR);
	}
#endif
	return result;
}

DriversListViewCtrl::ItemAdditionCount
DriversListViewCtrl::addDriversListItems(std::vector<DriverSetupData> const &driver_data, wchar_t const *str)
{
	typedef std::vector<DriverSetupData>::const_iterator DataIterator;
	ItemAdditionCount result;
	for (DataIterator it = driver_data.begin(); it != driver_data.end(); ++it) {
		switch (this->addDriversListItem(*it, str)) {
		case ITEM_ADDED:
			++(result.add_count);
			break;
		case ITEM_REPLACED:
			++(result.replace_count);
			break;
		default:
			break;
		}
	}
	return result;
}

DriversListViewCtrl::ItemAdditionCount
DriversListViewCtrl::addDriversFromDir(boost::filesystem::wpath const &path, wchar_t const *str)
{
	ItemAdditionCount result;

	DWORD buf_count = 0;
	std::wstring const path_str = path.string();
	if (::SetupGetInfFileList(path_str.c_str(), INF_STYLE_WIN4, NULL, 0, &buf_count) && buf_count > 1) {
		boost::scoped_array<wchar_t> buf(new wchar_t[buf_count]);
		wchar_t *ptr = buf.get();
		if (::SetupGetInfFileList(path_str.c_str(), INF_STYLE_WIN4, ptr, buf_count, NULL)) {
			size_t n;
			while ((n = wcslen(ptr)) > 0) {
				ItemAdditionResult iar = this->addDriversListItemFromInfFile(path/ptr, this->architecture_id, this->os_version, str);
				switch (iar) {
				case ITEM_ADDED:
					++(result.add_count);
					break;
				case ITEM_REPLACED:
					++(result.replace_count);
					break;
				default:
					break;
				}
				// ptr -> inf file name
				ptr += n + 1;
			}
		}
	}
	fs::wdirectory_iterator it_end;
	for (fs::wdirectory_iterator it(path); it != it_end; ++it) {
		fs::wpath const src = it->path();
		if (fs::is_directory(src)) {
			//ċAĂяo
			ItemAdditionCount const ret = this->addDriversFromDir(src, str);
			result.add_count += ret.add_count;
			result.replace_count += ret.replace_count;
		}
	}

	return result;
}

void DriversListViewCtrl::addIntegratedDrivers(boost::filesystem::wpath const &path)
{
	DWORD buf_count = 0;
	std::wstring const path_str = path.string();
	if (::SetupGetInfFileList(path_str.c_str(), INF_STYLE_WIN4, NULL, 0, &buf_count) && buf_count > 1) {
		boost::scoped_array<wchar_t> buf(new wchar_t[buf_count]);
		wchar_t *ptr = buf.get();
		if (::SetupGetInfFileList(path_str.c_str(), INF_STYLE_WIN4, ptr, buf_count, NULL)) {
			size_t n;
			while ((n = wcslen(ptr)) > 0) {
				try {
					fs::wpath const file_path = path/ptr;
					DriverSetupData driver_data(file_path.string(), this->architecture_id, this->os_version);

					DriverSetupData::DDInstallCollection const &ddic = driver_data.getDDInstallSectionCollection();
					DriverSetupData::DDInstallCollection::const_iterator it = ddic.begin();
					// DDInstallZNV  OSp
					if (it != ddic.end()) {
						// foCX̎ނ̕\擾
						wchar_t class_name[256];
						if (!::SetupDiGetClassDescription(&driver_data.getClassUUID(), class_name, 256, NULL)) {
							class_name[0] = L'\0';
						}
						// foCX̃ACR擾
						int nnn;
						if (!::SetupDiGetClassImageIndex(&image_list_data, &driver_data.getClassUUID(), &nnn)) {
							nnn = 0;
						}
						// ɒǉ
						int const insert_pos = this->GetItemCount();
						// ڂ̒ǉ
						DriverSetupData::DescriptionCollection::const_iterator it2 =
							it->second.descriptions.begin();
						this->InsertItem(insert_pos,
										 (it2 != it->second.descriptions.end()) ? it2->c_str() : L"No Description",
										 nnn);
						// foCX̎
						this->SetItemText(insert_pos, 1, class_name);
						// INFt@C̃pX
						this->SetItemText(insert_pos, 2, L"<ߋɓhCo>");
						// ڂƁAhCõf[^֘At
						ItemAssociatedData *ptr = new ItemAssociatedData(ItemAssociatedData::ADDED_BEFORE, driver_data);
						this->SetItemData(insert_pos, reinterpret_cast<uintptr_t>(ptr));
					}

				} catch (...) {
					// nop
				}
				// ptr -> inf file name
				ptr += n + 1;
			}
		}
	}
	fs::wdirectory_iterator it_end;
	for (fs::wdirectory_iterator it(path); it != it_end; ++it) {
		fs::wpath const src = it->path();
		if (fs::is_directory(src)) {
			//ċAĂяo
			this->addIntegratedDrivers(src);
		}
	}
}

DriversListViewCtrl::ItemAdditionCount
DriversListViewCtrl::addDriversFromFileOrDir(boost::filesystem::wpath const &path, wchar_t const *str)
{
	ItemAdditionCount result;
	if (fs::is_directory(path)) {
		result = this->addDriversFromDir(path, str);
	} else {
		ItemAdditionResult iar = this->addDriversListItemFromInfFile(path, this->architecture_id, this->os_version, str);
		switch (iar) {
		case ITEM_ADDED:
			++(result.add_count);
			break;
		case ITEM_REPLACED:
			++(result.replace_count);
			break;
		default:
			break;
		}
	}
	return result;
}

void DriversListViewCtrl::removeSelectedItem()
{
	int i = this->GetNextItem(-1, LVIS_SELECTED);
	while (i != -1) {
		ItemAssociatedData *ptr = reinterpret_cast<ItemAssociatedData*>(this->GetItemData(i));
		ItemAssociatedData::ItemState item_state = ptr->getItemState();
		if (item_state == ItemAssociatedData::ADDED_BEFORE) {
			ptr->setItemState(ItemAssociatedData::ADDED_BEFORE_REMOVE);
			i = this->GetNextItem(i, LVIS_SELECTED);
		} else if (item_state == ItemAssociatedData::ADDED_BEFORE_REMOVE) {
			// d`FbN
			ItemAdditionResult result = this->checkDuplication(*(ptr->getItemData()));
			if (result != NO_CHANGE) {
				ptr->setItemState(ItemAssociatedData::ADDED_BEFORE);
			}
			i = this->GetNextItem(i, LVIS_SELECTED);
		} else {
			this->DeleteItem(i);
			i = this->GetNextItem(i - 1, LVIS_SELECTED);
		}
	}
	CRect r;
	this->GetClientRect(r);
	this->InvalidateRect(r);
}

void DriversListViewCtrl::showPropertyOfSelectdeItem()
{
	int i = this->GetNextItem(-1, LVIS_SELECTED);
	while (i != -1) {
		TCHAR buf[256];
		this->GetItemText(i, 0, buf, 256);
		ItemAssociatedData *ptr = reinterpret_cast<ItemAssociatedData*>(this->GetItemData(i));
		DriverPropertyDialog dlg(buf, *(ptr->getItemData()));
		dlg.DoModal();
		i = this->GetNextItem(i, LVIS_SELECTED);
	}
}

void DriversListViewCtrl::showItemAdditionMessage(ItemAdditionCount add_count)
{
	bool no_msg = true;
	std::wostringstream ss;
	if (add_count.add_count > 0) {
		ss << add_count.add_count << L" ̃hCoǉ܂B";
		no_msg = false;
	}
	if (!no_msg) {
		ss << L'\n';
	}
	if (add_count.replace_count > 0) {
		ss << add_count.replace_count << L" ̃hCoւ܂B";
		no_msg = false;
	}
	if (!no_msg) {
		this->MessageBox(ss.str().c_str(), L"F6c[", MB_ICONINFORMATION);
	}
}

/////////////////////////////////////////////////////////////////////////////

void DriversListViewCtrl::OnKeyDown(unsigned int nChar, unsigned int /*nRepCnt*/, unsigned int /*nFlags*/)
{
	switch (nChar) {
	case VK_DELETE:
		this->removeSelectedItem();
		break;
	//case VK_RETURN:
	//	if (::GetKeyState(VK_MENU) & 0x80) {
	//		this->showPropertyOfSelectdeItem();
	//		return 0;
	//	}
	//	break;
	}
	this->SetMsgHandled(false);
}

void DriversListViewCtrl::OnDropFiles(HDROP drop_info)
{
	ItemAdditionCount result;
	unsigned int count = ::DragQueryFile(drop_info, 0xFFFFFFFF, NULL, 0);
	for (unsigned int i = 0; i < count; ++i) {
		wchar_t buf[MAX_PATH + 1];
		::DragQueryFile(drop_info, i, buf, MAX_PATH);
		ItemAdditionCount const ret = this->addDriversFromFileOrDir(boost::filesystem::wpath(buf));
		result.add_count += ret.add_count;
		result.replace_count += ret.replace_count;
	}
	::DragFinish(drop_info);
	this->showItemAdditionMessage(result);
}
