#include "InstalledDriverDialog.h"

InstalledDriverDialog::InstalledDriverDialog(int architecture_id, uint32_t os_version)
: architecture_id(architecture_id)
, os_version(os_version)
{
}

std::vector<DriverSetupData> InstalledDriverDialog::getData()
{
	return this->result;
}

//===========================================================================
// MyRegKey
//===========================================================================
class MyRegKey :
	public CRegKey
{
public:
	explicit MyRegKey(HKEY reg_key_handle)
	{
		this->Attach(reg_key_handle);
	}

	std::wstring queryStringValue(LPCTSTR value_name)
	{
		unsigned long required_count = 0;
		this->QueryValue(NULL, value_name, &required_count);
		//this->QueryStringValue(value_name, NULL, &required_count);
		if (required_count == 0) {
			return L"";
		}
		boost::scoped_array<wchar_t> temp_buf(new wchar_t[required_count]);
		//long ret = this->QueryStringValue(
			//value_name, temp_buf.getPointer(), &required_count);
		long ret = this->QueryValue(temp_buf.get(), value_name, &required_count);
		if (ret != NO_ERROR) {
			return L"";
		}
		return temp_buf.get();
	}
};

//===========================================================================
// DeviceEnumerator
//===========================================================================
class DeviceEnumerator
{
public:
	DeviceEnumerator(UUID *class_uuid, LPCTSTR enumerator, unsigned int flags)
		: devinfo_handle(NULL)
		, device_index(0)
	{
		this->devinfo_handle =
			::SetupDiGetClassDevs(class_uuid, enumerator, NULL, flags);
		if (this->devinfo_handle == NULL) {
			throw MY_SETUP_API_EXCEPTION_LAST_ERROR;
		}

		this->devinfo_data.cbSize = sizeof(SP_DEVINFO_DATA);

		this->enumDeviceInfo();
	}

	~DeviceEnumerator()
	{
		this->close();
	}

	void operator++()
	{
		++this->device_index;
		this->enumDeviceInfo();
	}

	std::wstring getDeviceRegistryProperty(unsigned int flag)
	{
		DWORD reg_data_type;
		DWORD required_size = 0;
		::SetupDiGetDeviceRegistryProperty(this->devinfo_handle, &this->devinfo_data,
			flag, &reg_data_type, NULL, 0, &required_size);
		if (required_size == 0) {
			return L"";
		}
		boost::scoped_array<BYTE> temp_buf(new BYTE[required_size]);
		BOOL ret = ::SetupDiGetDeviceRegistryProperty(
			this->devinfo_handle, &this->devinfo_data, flag, &reg_data_type,
			temp_buf.get(), required_size, NULL);
		if (!ret) {
			return L"";
		}
		return reinterpret_cast<wchar_t*>(temp_buf.get());
	}

	HKEY openDevRegKey()
	{
		return ::SetupDiOpenDevRegKey(this->devinfo_handle, &this->devinfo_data,
			DICS_FLAG_GLOBAL, 0, DIREG_DRV, KEY_READ);
	}

	bool empty()
	{
		return this->devinfo_handle == NULL;
	}

private:
	HDEVINFO devinfo_handle;
	SP_DEVINFO_DATA devinfo_data;
	unsigned int device_index;

	void enumDeviceInfo()
	{
		BOOL ret = ::SetupDiEnumDeviceInfo(
			this->devinfo_handle, this->device_index, &this->devinfo_data);
		if (!ret) {
			this->close();
			const DWORD last_err = ::GetLastError();
			if (last_err != NO_ERROR && last_err != ERROR_NO_MORE_ITEMS) {
				throw MY_SETUP_API_EXCEPTION(HRESULT_FROM_SETUPAPI(last_err));
			}
		}
	}

	void close()
	{
		if (this->devinfo_handle != NULL) {
			::SetupDiDestroyDeviceInfoList(this->devinfo_handle);
			this->devinfo_handle = NULL;
		}
		//this->device_index = 0xFFFFFFFF;
	}
};

BOOL InstalledDriverDialog::OnInitDialog(HWND /*wndFocus*/, LPARAM /*lInitParam*/)
{
	this->drivers_list.SubclassWindow(this->GetDlgItem(IDC_LIST1));
	this->drivers_list.SetExtendedListViewStyle(LVS_EX_CHECKBOXES | LVS_EX_INFOTIP);
	// foCXNX̃C[WXg擾
	this->image_list_data.cbSize = sizeof(SP_CLASSIMAGELIST_DATA);
	::SetupDiGetClassImageList(&image_list_data);
	this->drivers_list.SetImageList(image_list_data.ImageList, LVSIL_SMALL);

	//CRect rect;
	//this->drivers_list.GetClientRect(&rect);
	this->drivers_list.InsertColumn(0, _T("O"), 0, 300);
	this->drivers_list.SetColumnSortType(0, LVCOLSORT_TEXTNOCASE);
	this->drivers_list.InsertColumn(1, _T(""), 0, 200);
	this->drivers_list.SetColumnSortType(1, LVCOLSORT_TEXTNOCASE);

	// Device Tree
	std::set<std::wstring, MyLib::LessNoCase> inf_file_titles;
	DeviceEnumerator device_enumerator(NULL, NULL, DIGCF_ALLCLASSES);
	// Enumerate through all devices in Set.
	for (; !device_enumerator.empty(); ++device_enumerator) {
		std::wstring class_guid = 
			device_enumerator.getDeviceRegistryProperty(SPDRP_CLASSGUID);
		if (class_guid.empty()) {
			continue;
		}
		MyUUID class_uuid = class_guid;
		if (class_uuid == GUID_DEVCLASS_LEGACYDRIVER) {
			continue;
		}
		HKEY reg_key_handle = device_enumerator.openDevRegKey();
		if (reg_key_handle == INVALID_HANDLE_VALUE) {
			continue;
		}
		MyRegKey reg_key(reg_key_handle);
		const std::wstring inf_file_title = reg_key.queryStringValue(_T("InfPath"));
		if (!inf_file_title.empty()
		 && inf_file_titles.find(inf_file_title) == inf_file_titles.end()
		 && MyLib::compareStringNoCase(reg_key.queryStringValue(_T("ProviderName")), _T("Microsoft")) != 0) {
			inf_file_titles.insert(inf_file_title);
			// 
			DriverSetupData::InstalledDriverInfo driver_info;
			driver_info.inf_file_title = inf_file_title;
			driver_info.inf_section = reg_key.queryStringValue(_T("InfSection"));
			driver_info.inf_section_ext = reg_key.queryStringValue(_T("InfSectionExt"));
			DriverSetupData *ptr = NULL;
			try {
				ptr = new DriverSetupData(driver_info, this->architecture_id, this->os_version);
			} catch (...) {
				continue;
			}
			if (ptr->getDDInstallSectionCollection().size() > 0) {
				TCHAR class_name[256];
				::SetupDiGetClassDescription(&class_uuid, class_name, 256, NULL);
				int nnn;
				::SetupDiGetClassImageIndex(&image_list_data, &class_uuid, &nnn);

				// \
				std::wstring friendly_name =
					device_enumerator.getDeviceRegistryProperty(SPDRP_FRIENDLYNAME);
				if (friendly_name.empty()) {
					friendly_name =
						device_enumerator.getDeviceRegistryProperty(SPDRP_DEVICEDESC);
				}
				const int insert_pos = this->drivers_list.GetItemCount();
				this->drivers_list.InsertItem(insert_pos, friendly_name.c_str(), nnn);
				this->drivers_list.SetItemText(insert_pos, 1, class_name);

				this->drivers_list.SetItemData(insert_pos, reinterpret_cast<uintptr_t>(ptr));
			} else {
				delete ptr;
			}
		}
	}

	return TRUE;
}

void InstalledDriverDialog::OnOK(unsigned int /*uNotifyCode*/, int nID, HWND /*wndCtl*/)
{
	int const item_count = this->drivers_list.GetItemCount();
	this->result.reserve(item_count);
	for (int i = 0; i < item_count; ++i) {
		DriverSetupData *ptr =
			reinterpret_cast<DriverSetupData*>(this->drivers_list.GetItemData(i));
		if (this->drivers_list.GetCheckState(i)) {
			this->result.push_back(*ptr);
		}
	}
	this->EndDialog(nID);
}

void InstalledDriverDialog::OnCancel(unsigned int /*uNotifyCode*/, int nID, HWND /*wndCtl*/)
{
	this->EndDialog(nID);
}

LRESULT InstalledDriverDialog::OnDeleteItem(NMHDR *pnmh)
{
	NMLISTVIEW *ptr = reinterpret_cast<NMLISTVIEW*>(pnmh);
	delete reinterpret_cast<DriverSetupData*>(this->drivers_list.GetItemData(ptr->iItem));
	return 0;
}
