#include "OutputSettingPage.h"
#include "core/OutputSetting.h"

void OutputSettingPage::loadSourceSetting(SourceSetting const *source_setting)
{
	this->record_setting_page.loadSourceSetting(source_setting);
}

OutputSetting* OutputSettingPage::getSetting()
{
	OutputSetting *result = NULL;
	boost::scoped_ptr<RecordSetting> record_setting(
		this->record_setting_page.getSetting());
	if (!record_setting) {
		this->setCurrentTabPage(TAB_PAGE_RECORD_SETTING);
		return NULL;
	}
	boost::scoped_ptr<USBSetting> usb_setting(
		this->usb_setting_page.getSetting());
	if (!usb_setting) {
		this->setCurrentTabPage(TAB_PAGE_USB_SETTING);
		return NULL;
	}
	result = new OutputSetting(
		record_setting->volume_label,
		record_setting->iso_image_path,
		record_setting->record_device,
		record_setting->record_speed,
		record_setting->erases_media,
		record_setting->ejects_media,
		usb_setting->reserved_drive,
		usb_setting->target_drive,
		usb_setting->setup_drive);
	return result;
}

void OutputSettingPage::setCurrentTabPage(unsigned int page_index)
{
#ifdef DEBUG
	size_t const page_count = this->tab_pages.size();
	ATLASSERT(0 <= page_index && page_index < page_count && page_count == NUM_TAB_PAGES);
#endif
	this->tab_ctrl.SetCurSel(page_index);
	this->showTabPage(page_index);
	this->tab_pages[page_index]->SetFocus();
}

void OutputSettingPage::showTabPage(unsigned int page_index)
{
	//size_t const page_count = this->tab_pages.size();
	ATLASSERT(0 <= page_index && page_index < NUM_TAB_PAGES/*page_count*/);
	for (size_t i = 0; i < NUM_TAB_PAGES/*page_count*/; ++i) {
		this->tab_pages[i]->ShowWindow(page_index == i ? SW_SHOW : SW_HIDE);
	}
}

BOOL OutputSettingPage::OnInitDialog(HWND, LPARAM)
{
#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"OutputSettingPage::OnInitDialog() start", PROFILE_PATH->string().c_str());
#endif
	this->tab_ctrl = this->GetDlgItem(IDC_TAB1);
	// NCAgW擾邽߁AɃ^uRg[̍ڂǉĂ
	this->tab_ctrl.InsertItem(TAB_PAGE_RECORD_SETTING, _T("ISOC[WECDĂ"));
	this->tab_ctrl.InsertItem(TAB_PAGE_USB_SETTING, _T("USBEHDDCXg["));
	// ^uRg[̃NCAgW擾
	this->tab_pages.resize(NUM_TAB_PAGES);
	RECT tab_rect;
	this->tab_ctrl.GetWindowRect(&tab_rect);
	this->tab_ctrl.AdjustRect(FALSE, &tab_rect);
	this->ScreenToClient(&tab_rect);
	// RecordSettingPage
	this->record_setting_page.Create(this->m_hWnd);
	this->record_setting_page.SetDlgCtrlID(IDD_RECORD_SETTING_PAGE);
	this->tab_pages[TAB_PAGE_RECORD_SETTING] = &this->record_setting_page;
	// USBSettingPage
	this->usb_setting_page.Create(this->m_hWnd);
	this->usb_setting_page.SetDlgCtrlID(IDD_USB_SETTING_PAGE);
	this->tab_pages[TAB_PAGE_USB_SETTING] = &this->usb_setting_page;
	// e^uy[ẄʒuZI[_[𒲐
	for (size_t i = 0; i < this->tab_pages.size(); ++i) {
		// Create()RECTw肵Ă҂zuɂȂȂ̂ŁAMoveWindow()gKv
		this->tab_pages[i]->MoveWindow(&tab_rect);
		this->tab_pages[i]->SetWindowPos(HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	}
	// ԍŏ̃y[W\悤ɂ
	this->showTabPage(TAB_PAGE_RECORD_SETTING);

	this->DlgResize_Init(false);

#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"OutputSettingPage::OnInitDialog() end", PROFILE_PATH->string().c_str());
#endif
	return true;
}

LRESULT OutputSettingPage::OnTcnSelchangeTab1(LPNMHDR /*pnmh*/)
{
	this->showTabPage(this->tab_ctrl.GetCurSel());
	return 0;
}
