#include "RecordSettingPage.h"
#include "core/SourceSetting.h"
#include "mylib/exception.h"
#include "mylib/function.h"
#include "mylib/ConsoleProcessExecutor.h"
#include "mylib/inifile.h"

RecordSettingPage::RecordSettingPage()
: is_save_iso_checked(false)
, is_record_checked(false)
{
	IniFile ini_file(*PROFILE_PATH);
	this->erases_media_default =
		(ini_file.getInt(L"Settings", L"erase_media", 0) != 0);
	this->erases_media = this->erases_media_default;
	this->ejects_media_default =
		(ini_file.getInt(L"Settings", L"eject_media", 1) != 0);
	this->ejects_media = this->ejects_media_default;
	int n = ini_file.getInt(L"Settings", L"record_speed", 0);
	this->record_speed_default = ((1 <= n && n <= 56) ? n : 8);
}

void RecordSettingPage::switchControls()
{
	this->DoDataExchange(DDX_SAVE);

	this->iso_path_label.EnableWindow(this->is_save_iso_checked);
	this->iso_path_edit.EnableWindow(this->is_save_iso_checked);
	this->iso_path_button.EnableWindow(this->is_save_iso_checked);

	this->record_drive_label.EnableWindow(this->is_record_checked);
	this->record_drive_list.EnableWindow(this->is_record_checked);
	this->reload_button.EnableWindow(this->is_record_checked);
	this->record_speed_label.EnableWindow(this->is_record_checked);
	this->record_speed_edit.EnableWindow(this->is_record_checked);
	this->record_speed2_label.EnableWindow(this->is_record_checked);
	this->record_speed_updown.EnableWindow(this->is_record_checked);
	this->eject_media_checkbox.EnableWindow(this->is_record_checked);
	this->erase_media_checkbox.EnableWindow(this->is_record_checked);

	this->volume_label_label.EnableWindow(this->is_save_iso_checked || this->is_record_checked);
	this->volume_label_edit.EnableWindow(this->is_save_iso_checked || this->is_record_checked);
}

RecordSetting* RecordSettingPage::getSetting()
{
	this->DoDataExchange(DDX_SAVE);

	std::auto_ptr<RecordSetting> result(new RecordSetting());
	// ISO Image
	if (this->is_save_iso_checked) {
		if (this->iso_path.empty()) {
			this->MessageBox(_T("ISOC[W̃t@Cw肭"));
			return NULL;
		}
		result->iso_image_path = this->iso_path;
	}
	// CD Record
	if (this->is_record_checked) {
		int const selection = this->record_drive_list.GetCurSel();
		if (selection == CB_ERR) {
			return NULL;
		}
		int const n = this->record_drive_list.GetLBTextLen(selection) + 1;
		boost::scoped_array<wchar_t> buf(new wchar_t[n]);
		this->record_drive_list.GetLBText(selection, buf.get());
		result->record_device = buf.get();
		if (result->record_device.empty()) {
			this->MessageBox(_T("CD݂̏ɎgphCuw肭"));
			return NULL;
		}

		IniFile ini_file(*PROFILE_PATH);
		int record_speed = this->record_speed_updown.GetPos();
		if (record_speed != this->record_speed_default) {
			ini_file.setInt(L"Settings", L"record_speed", record_speed);
		}
		result->record_speed = record_speed;

		bool ejects_media = this->ejects_media;
		if ((ejects_media && !this->ejects_media_default) || (!ejects_media && this->ejects_media_default)) {
			ini_file.setInt(L"Settings", L"eject_media", ejects_media);
		}
		result->ejects_media = ejects_media;

		bool erases_media = this->erases_media;
		if ((erases_media && !this->erases_media_default) || (!erases_media && this->erases_media_default)) {
			ini_file.setInt(L"Settings", L"erase_media", erases_media);
		}
		result->erases_media = erases_media;
	}
	// Volume Label
	if (this->is_save_iso_checked || this->is_record_checked) {
		result->volume_label = this->volume_label;
	}
	return result.release();
}

void RecordSettingPage::loadSourceSetting(SourceSetting const *source_setting)
{
	type_info const &t_info = typeid(*source_setting);
	if (t_info == typeid(DiscSourceSetting)) {
		DiscSourceSetting const *ptr = static_cast<DiscSourceSetting const*>(source_setting);
		this->volume_label = ptr->getVolumeLabel();
		this->DoDataExchange(DDX_LOAD);
	}
}

void RecordSettingPage::enumerateRecorders()
{
	try {
		// ݁AIĂ鍀ڂL
		int const selection = this->record_drive_list.GetCurSel();
		std::wstring selected_text;
		if (selection != CB_ERR) {
			int const n = this->record_drive_list.GetLBTextLen(selection) + 1;
			boost::scoped_array<wchar_t> selected_text_buf(new wchar_t[n]);
			this->record_drive_list.GetLBText(selection, selected_text_buf.get());
			selected_text = selected_text_buf.get();
		}
		// ڂS
		this->record_drive_list.ResetContent();

		MyLib::ConsoleProcessExecutor cdrecord_exe;
		cdrecord_exe.run((*BIN_DIR_PATH/L"cdrecord.exe").string() + L" -scanbus");
		unsigned int status;
		std::stringstream ss;
		while ((status = cdrecord_exe.getStatus()) != WAIT_ABANDONED) {
			std::vector<char> buf = cdrecord_exe.readStdOut();
			buf.push_back('\0');
			ss << &buf[0];
			if (status == WAIT_OBJECT_0) break;
		}
		const unsigned int exit_code = cdrecord_exe.getExitCode();
		if (exit_code != 0) {
			char buf[64];
			::wsprintfA(buf, "cdrecord.exe returned %d.", exit_code);
			throw std::runtime_error(buf);
		}

		int new_drive_selection = 0;
		std::string str;
		while (std::getline(ss, str, '\n')) {
			if (str.find("Removable CD-ROM") != std::string::npos) {
				size_t i = 0;
				for (; ::isspace(static_cast<unsigned char>(str[i])); ++i);
				std::wstring const display_text = MyLib::convertFromSJIS(str.substr(i));
				int pos = this->record_drive_list.AddString(display_text.c_str());
				// đIĂhCuƈv鍀
				if (display_text == selected_text) {
					new_drive_selection = pos;
				}
			}
		}
		// đIĂhCuĂёI悤ɂ
		const int record_drive_count = this->record_drive_list.GetCount();
		if (record_drive_count > 0) {
			this->record_drive_list.SetCurSel(new_drive_selection);
		}
	} catch (std::runtime_error const &e) {
		this->MessageBox(MyLib::convertFromUTF8(e.what()).c_str(), L"cdrecord.exẽG[");
	} catch (MyLib::Exception<wchar_t>& e) {
		std::wostringstream ss;
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L'(' << e.getErrorLine() << L')';
		}
		ss << e.getErrorText() << L'[' << e.getErrorCode() << L']';
		this->MessageBox(ss.str().c_str(), L"cdrecord.exẽG[");
	}
}

BOOL RecordSettingPage::OnInitDialog(HWND, LPARAM)
{
#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"RecordSettingPage::OnInitDialog() start", PROFILE_PATH->string().c_str());
#endif
	this->iso_path_label = this->GetDlgItem(IDC_STATIC_ISO_PATH);
	this->iso_path_edit = this->GetDlgItem(IDC_EDIT_FILE);
	this->iso_path_button = this->GetDlgItem(IDC_BUTTON_FILE);
	this->record_drive_label = this->GetDlgItem(IDC_STATIC_RECORD_DRIVE);
	this->record_drive_list = this->GetDlgItem(IDC_COMBO_RECORD_DRIVE);
	this->reload_button = this->GetDlgItem(IDC_BUTTON_RELOAD);
	this->record_speed_label = this->GetDlgItem(IDC_STATIC_RECORD_SPEED);
	this->record_speed2_label = this->GetDlgItem(IDC_STATIC_RECORD_SPEED2);
	this->record_speed_edit = this->GetDlgItem(IDC_EDIT_RECORD_SPEED);
	this->record_speed_updown = this->GetDlgItem(IDC_SPIN1);
	this->eject_media_checkbox = this->GetDlgItem(IDC_CHECKBOX_EJECT_MEDIA);
	this->erase_media_checkbox = this->GetDlgItem(IDC_CHECKBOX_ERASE_MEDIA);
	this->volume_label_label = this->GetDlgItem(IDC_STATIC_VOLUME_LABEL);
	this->volume_label_edit = this->GetDlgItem(IDC_EDIT_VOLUME_LABEL);
	this->switchControls();

	this->enumerateRecorders();

	// ݃hCuꗗ̑IԂ̏
	if (this->record_drive_list.GetCount() > 0) {
		this->record_drive_list.SetCurSel(0);
	}
	// ݑxꗗ
	this->record_speed_updown = this->GetDlgItem(IDC_SPIN1);
	this->record_speed_updown.SetRange(1, 56);
	this->record_speed_updown.SetPos(this->record_speed_default);
	// u݌AfBXNovɃ`FbNĂ
	this->eject_media_checkbox.SetCheck(BST_CHECKED);

	this->DlgResize_Init(false);

#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"RecordSettingPage::OnInitDialog() end", PROFILE_PATH->string().c_str());
#endif
	return true;
}

void RecordSettingPage::OnCheckboxSaveIso(unsigned int, int, HWND)
{
	this->switchControls();
}

void RecordSettingPage::OnCheckboxRecord(unsigned int, int, HWND)
{
	this->switchControls();
}

void RecordSettingPage::OnButtonFile(unsigned int, int, HWND)
{
	SaveFileDialog dlg(_T("iso"), NULL, OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY, 
		_T("ISO C[Wt@C (*.iso)\0*.iso\0ׂẴt@C (*.*)\0*.*\0"));
	if (dlg.DoModal() == IDOK) {
		this->iso_path = dlg.m_szFileName;
		this->DoDataExchange(DDX_LOAD);
	}
}

void RecordSettingPage::OnButtonReload(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	this->enumerateRecorders();
}

void RecordSettingPage::OnEnChangeEditRecordSpeed(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	if (this->record_speed_updown.IsWindow()) {
		// ͈̓`FbN
		BOOL ret;
		this->record_speed_updown.GetPos(&ret);
		if (ret) {
			this->record_speed_updown.SetPos(56);
		}
	}
}
