﻿#include "SPSettingPage.h"
#include "core/SourceSetting.h"
#include "mylib/inifile.h"

namespace fs {
	using namespace boost::filesystem;
};

SPSettingPage::SPSettingPage()
: uses_bigdrive_w2k(false)
//, clean_install_only(false)
//, driver_cab_recompress(false)
{
	IniFile ini_file(*PROFILE_PATH);
	this->clean_install_only_default =
		(ini_file.getInt(L"Settings", L"clean_install_only", 0) != 0);
	this->clean_install_only = this->clean_install_only_default;
	this->driver_cab_recompress_default =
		(ini_file.getInt(L"Settings", L"driver_cab_recompress", 0) != 0);
	this->driver_cab_recompress = this->driver_cab_recompress_default;
	this->sp_file_path[0] = L'\0';
}

void SPSettingPage::loadSourceSetting(SourceSetting const *source_setting)
{
	bool is_w2k = (source_setting->getOSVersion() == 0x00050000);
	this->big_drive_checkbox.EnableWindow(is_w2k);
	this->uses_bigdrive_w2k = is_w2k;

	int sp_number = source_setting->getCurrentSPNumber();
	if (sp_number > 0) {
		TCHAR buf[256];
		::wsprintf(buf, _T("現在、サービスパック %d が適用済みです。"), sp_number);
		this->SetDlgItemText(IDC_STATIC_CURRENT_SP, buf);
	} else {
		this->SetDlgItemText(IDC_STATIC_CURRENT_SP, L"");
	}

	// SP適用の有効・無効切り替え
	// SP適用は F6回避ツールとターゲットOSで CPU Architecture が一致している必要あり
#if defined _M_X64
	bool is_same_architecture =
		(source_setting->getArchitectureID() == PROCESSOR_ARCHITECTURE_AMD64);
#elif defined _M_IA64
	bool is_same_architecture =
		(source_setting->getArchitectureID() == PROCESSOR_ARCHITECTURE_IA64);
#else
	bool is_same_architecture =
		(source_setting->getArchitectureID() == PROCESSOR_ARCHITECTURE_INTEL);
#endif
	this->sp_file_box.EnableWindow(is_same_architecture);
	this->open_file_button.EnableWindow(is_same_architecture);
	this->sp_file_box.SetWindowText(L"");

	this->DoDataExchange(DDX_LOAD);
}

SPSetting* SPSettingPage::getSetting()
{
	this->DoDataExchange(DDX_SAVE);

	SPSetting *result = NULL;
	fs::wpath p = this->sp_file_path;
	if (p.empty() || (fs::exists(p) && fs::is_regular_file(p))) {
		result = new SPSetting;
		result->sp_file_path = p;
		result->uses_bigdrive_w2k = this->uses_bigdrive_w2k;

		IniFile ini_file(*PROFILE_PATH);
		bool const clean_install_only = this->clean_install_only;
		if ((clean_install_only && !this->clean_install_only_default) || (!clean_install_only && this->clean_install_only_default)) {
			ini_file.setInt(L"Settings", L"clean_install_only", clean_install_only);
		}
		result->is_clean_install_only = clean_install_only;
		bool const driver_cab_recompress = this->driver_cab_recompress;
		if ((driver_cab_recompress && !this->driver_cab_recompress_default) || (!driver_cab_recompress && this->driver_cab_recompress_default)) {
			ini_file.setInt(L"Settings", L"driver_cab_recompress", driver_cab_recompress);
		}
		result->executes_driver_cab_recompress = driver_cab_recompress;
	} else {
		this->MessageBox(_T("ご指定のファイルが見つかりません"),
			_T("サービスパックファイルのエラー"));
	}
	return result;
}

BOOL SPSettingPage::OnInitDialog(HWND /*wndFocus*/, LPARAM /*lInitParam*/)
{
#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"SPSettingPage::OnInitDialog() start", PROFILE_PATH->string().c_str());
#endif

	//this->current_sp_number = this->GetDlgItem(IDC_STATIC_CURRENT_SP);
	this->sp_file_box = this->GetDlgItem(IDC_EDIT_FILE);
	this->open_file_button = this->GetDlgItem(IDC_BUTTON_FILE);
	this->big_drive_checkbox = this->GetDlgItem(IDC_CHECKBOX_BIGDRIVE);
	this->clean_install_only_checkbox =
		this->GetDlgItem(IDC_CHECKBOX_CLEAN_INSTALL_ONLY);
	this->recompress_driver_cab_checkbox =
		this->GetDlgItem(IDC_CHECKBOX_DRIVER_CAB_RECOMPRESS);
	this->groupbox_sp.SubclassWindow(this->GetDlgItem(IDC_GROUPBOX_SP));
	this->groupbox_bigdrive.SubclassWindow(this->GetDlgItem(IDC_GROUPBOX_BIGDRIVE));
	this->groupbox_clean_install_only.SubclassWindow(this->GetDlgItem(IDC_GROUPBOX_CLEAN_INSTALL_ONLY));

	this->DlgResize_Init(false);

	this->DoDataExchange(DDX_LOAD);


#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"SPSettingPage::OnInitDialog() end", PROFILE_PATH->string().c_str());
#endif
	return TRUE;
}

void SPSettingPage::OnButtonFile(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	OpenFileDialog dlg(NULL, NULL, OFN_HIDEREADONLY | OFN_FILEMUSTEXIST,
		_T("Service Pack ファイル (*.exe)\0*.exe\0"));
	if (dlg.DoModal() == IDOK) {
		::lstrcpy(this->sp_file_path, dlg.m_szFileName);
		this->DoDataExchange(DDX_LOAD);
	}
}
