#include "SlipStreamSettingPage.h"

#include "core/SourceSetting.h"
#include "core/SlipstreamSetting.h"

SlipstreamSetting* SlipstreamSettingPage::getSetting()
{
	SlipstreamSetting *result = NULL;

	boost::scoped_ptr<WorkDirSetting> work_dir_setting(
		this->work_dir_setting_page.getSetting());
	if (!work_dir_setting) {
		this->setCurrentTabPage(TAB_PAGE_WORK_DIR_SETTING);
		return NULL;
	}
	boost::scoped_ptr<SPSetting> sp_setting(
		this->sp_setting_page.getSetting());
	if (!sp_setting) {
		this->setCurrentTabPage(TAB_PAGE_SP_SETTING);
		return NULL;
	}
	boost::scoped_ptr<DriverSetting> driver_setting(
		this->driver_setting_page.getSetting());
	result = new SlipstreamSetting(
		work_dir_setting->work_dir_path,
		work_dir_setting->creates_temp_dir,
		sp_setting->sp_file_path, sp_setting->uses_bigdrive_w2k,
		sp_setting->is_clean_install_only,
		sp_setting->executes_driver_cab_recompress,
		driver_setting->add_drivers,
		driver_setting->leave_drivers,
		driver_setting->del_drivers,
		driver_setting->leave_services,
		driver_setting->del_services);
	return result;
}

void SlipstreamSettingPage::loadSourceSetting(SourceSetting const *source_setting)
{
	this->work_dir_setting_page.loadSourceSetting(source_setting);
	this->sp_setting_page.loadSourceSetting(source_setting);
	this->driver_setting_page.loadSourceSetting(source_setting);
}

BOOL SlipstreamSettingPage::PreTranslateMessage(MSG * /*pMsg*/)
{
	return FALSE;
}

BOOL SlipstreamSettingPage::OnInitDialog(HWND /*wndFocus*/, LPARAM /*lInitParam*/)
{
#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"SlipstreamSettingPage::OnInitDialog() start", PROFILE_PATH->string().c_str());
#endif

	this->tab_ctrl = this->GetDlgItem(IDC_TAB1);
	// NCAgW擾邽߁AɃ^uRg[̍ڂǉĂ
	this->tab_ctrl.InsertItem(TAB_PAGE_WORK_DIR_SETTING, _T("ƗptH_"));
	this->tab_ctrl.InsertItem(TAB_PAGE_SP_SETTING, _T("T[rXpbNE"));
	this->tab_ctrl.InsertItem(TAB_PAGE_DRIVER_SETTING, _T("foCXhCo"));
//	this->tab_ctrl.InsertItem(3, _T("lCXg["));
	// ^uRg[̃NCAgW擾
	this->tab_pages.resize(NUM_TAB_PAGES);
	RECT tab_rect;
	this->tab_ctrl.GetWindowRect(&tab_rect);
	this->tab_ctrl.AdjustRect(FALSE, &tab_rect);
	this->ScreenToClient(&tab_rect);
	// WorkDirPage
	this->work_dir_setting_page.Create(this->m_hWnd);
	this->work_dir_setting_page.SetDlgCtrlID(IDD_WORK_DIR_SETTING_PAGE);
	this->tab_pages[TAB_PAGE_WORK_DIR_SETTING] = &this->work_dir_setting_page;
	// SPPage
	this->sp_setting_page.Create(this->m_hWnd);
	this->sp_setting_page.SetDlgCtrlID(IDD_SP_SETTING_PAGE);
	this->tab_pages[TAB_PAGE_SP_SETTING] = &this->sp_setting_page;
	// DriverPage
	this->driver_setting_page.Create(this->m_hWnd);
	this->driver_setting_page.SetDlgCtrlID(IDD_DRIVER_SETTING_PAGE);
	this->tab_pages[TAB_PAGE_DRIVER_SETTING] = &this->driver_setting_page;
	//// UnattendSettingPage
	//this->unattend_setting_page.Create(this->m_hWnd);
	//this->unattend_setting_page.SetDlgCtrlID(IDD_UNATTEND_SETTING_PAGE);
	//this->tab_pages.push_back(&this->unattend_setting_page);
	//// RecordPage
	//this->tab_ctrl.InsertItem(3, _T("CD"));
	//this->record_page.Create(this->m_hWnd);
	//this->tab_pages.push_back(&this->record_page);
	// e^uy[ẄʒuZI[_[𒲐
	for (size_t i = 0; i < this->tab_pages.size(); ++i) {
		// Create()RECTw肵Ă҂zuɂȂȂ̂ŁAMoveWindow()gKv
		this->tab_pages[i]->MoveWindow(&tab_rect);
		this->tab_pages[i]->SetWindowPos(HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	}
	// ԍŏ̃y[W\悤ɂ
	this->showTabPage(0);

	this->DlgResize_Init(false);
#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"SlipstreamSettingPage::OnInitDialog() end", PROFILE_PATH->string().c_str());
#endif
	return TRUE;
}

LRESULT SlipstreamSettingPage::OnTcnSelchangeTab1(LPNMHDR /*pnmh*/)
{
	this->showTabPage(this->tab_ctrl.GetCurSel());
	return 0;
}

void SlipstreamSettingPage::showTabPage(unsigned int page_index)
{
	//size_t const page_count = this->tab_pages.size();
	ATLASSERT(0 <= page_index && page_index < NUM_TAB_PAGES/*page_count*/);
	for (size_t i = 0; i < NUM_TAB_PAGES/*page_count*/; ++i) {
		this->tab_pages[i]->ShowWindow(page_index == i ? SW_SHOW : SW_HIDE);
	}
}

void SlipstreamSettingPage::setCurrentTabPage(unsigned int page_index)
{
#ifdef DEBUG
	size_t const page_count = this->tab_pages.size();
	ATLASSERT(0 <= page_index && page_index < page_count && page_count == NUM_TAB_PAGES);
#endif
	this->tab_ctrl.SetCurSel(page_index);
	this->showTabPage(page_index);
	this->tab_pages[page_index]->SetFocus();
}

LRESULT SlipstreamSettingPage::OnPopupMenuSelect(unsigned int uMsg, WPARAM wParam, LPARAM lParam)
{
	return this->GetParent().PostMessage(uMsg, wParam, lParam);
}

//LRESULT SlipstreamSettingPage::OnDriverListImport(unsigned int, WPARAM, LPARAM)
//{
//	return this->driver_setting_page.SendMessage(DriverSettingPage::UWM_DRIVER_LIST_IMPORT);
//}
//
//LRESULT SlipstreamSettingPage::OnDriverListExport(unsigned int, WPARAM, LPARAM)
//{
//	return this->driver_setting_page.SendMessage(DriverSettingPage::UWM_DRIVER_LIST_EXPORT);
//}
