#include "USBSettingPage.h"
#include "mylib/exception.h"
#include "mylib/storagedevicenumber.h"
#include "mylib/drivelayoutinformation.h"
#include "mylib/inifile.h"

USBSettingPage::USBSettingPage()
: is_checked(false)
{
	IniFile ini_file(*PROFILE_PATH);
	// foCXhCoCXg[phCu
	std::wstring const reserved_drive_str = ini_file.getString(L"Settings", L"reserved_drive", L"");
	if (reserved_drive_str.empty()) {
		this->reserved_drive = L'Y';
		ini_file.setString(L"Settings", L"reserved_drive", L"Y:");
	} else {
		size_t buf_count = reserved_drive_str.length() + 1;
		boost::scoped_array<wchar_t> buf(new wchar_t[buf_count]);
		std::memcpy(buf.get(), reserved_drive_str.c_str(), buf_count*sizeof(wchar_t));
		::CharUpper(buf.get());
		wchar_t const drive_letter = buf[0];
		if (L'A' <= drive_letter && drive_letter <= L'Z') {
			this->reserved_drive = drive_letter;
		} else {
			this->reserved_drive = L'Y';
			ini_file.setString(L"Settings", L"reserved_drive", L"Y:");
		}
	}
	// ZbgAbvphCu
	std::wstring const setup_drive_str = ini_file.getString(L"Settings", L"setup_drive", L"");
	if (setup_drive_str.empty()) {
		this->setup_drive_default = ((reserved_drive != L'U') ? L'U' : L'Y');
		wchar_t buf[] = {this->setup_drive_default, L':', L'\0'};
		ini_file.setString(L"Settings", L"setup_drive", buf);
	} else {
		size_t buf_count = setup_drive_str.length() + 1;
		boost::scoped_array<wchar_t> buf(new wchar_t[buf_count]);
		std::memcpy(buf.get(), setup_drive_str.c_str(), buf_count*sizeof(wchar_t));
		::CharUpper(buf.get());
		wchar_t const drive_letter = buf[0];
		if (L'A' <= drive_letter && drive_letter <= L'Z') {
			if (this->setup_drive_default != this->reserved_drive) {
				this->setup_drive_default = drive_letter;
			} else {
				this->setup_drive_default = ((this->setup_drive_default < L'Z') ? ++(this->setup_drive_default) : --(this->setup_drive_default));
				wchar_t buf[] = {this->setup_drive_default, L':', L'\0'};
				ini_file.setString(L"Settings", L"setup_drive", buf);
			}
		} else {
			this->setup_drive_default = ((reserved_drive != L'U') ? L'U' : L'Y');
			wchar_t buf[] = {this->setup_drive_default, L':', L'\0'};
			ini_file.setString(L"Settings", L"setup_drive", buf);
		}
	}
}

USBSetting* USBSettingPage::getSetting()
{
	this->DoDataExchange(DDX_SAVE);

	std::auto_ptr<USBSetting> result(new USBSetting());
	result->reserved_drive = this->reserved_drive;
	if (this->is_checked) {
		int const target_drive_selection = this->target_drive_list.GetCurSel();
		if (target_drive_selection == CB_ERR) {
			this->MessageBox(_T("hCuI"));
			return NULL;
		}
		result->target_drive = static_cast<wchar_t>(this->target_drive_list.GetItemData(target_drive_selection));

		int const setup_drive_selection = this->setup_drive_list.GetCurSel();
		ATLASSERT(setup_drive_selection != CB_ERR);
		wchar_t buf[4];
		this->setup_drive_list.GetLBText(setup_drive_selection, buf);
		result->setup_drive = buf[0];
		if (buf[0] != this->setup_drive_default) {
			IniFile ini_file(*PROFILE_PATH);
			ini_file.setString(L"Settings", L"setup_drive", buf);
		}
	} else {
		result->target_drive = L'\0';
		result->setup_drive = L'\0';
	}
	return result.release();
}

BOOL USBSettingPage::OnInitDialog(HWND, LPARAM)
{
#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"USBSettingPage::OnInitDialog() start", PROFILE_PATH->string().c_str());
#endif
	this->DoDataExchange(DDX_LOAD);

	// Setup Drive List
	this->setup_drive_list = this->GetDlgItem(IDC_COMBO_SETUP_DRIVE);
	for (wchar_t i = L'A'; i <= L'Z'; ++i) {
		if (i == this->reserved_drive) {
			continue;
		}
		wchar_t buf[] = L"A:";
		buf[0] = i;
		this->setup_drive_list.AddString(buf);
	}
	this->setup_drive_list.SetCurSel(this->setup_drive_default - L'A');

	// Target Drive List
	this->target_drive_list = this->GetDlgItem(IDC_LIST_DRIVE);
	this->reload_button = this->GetDlgItem(IDC_BUTTON_RELOAD);
	this->enumerateDrives();
	this->switchControlsByCheckbox();

	this->DlgResize_Init(false);

#ifdef MY_DEBUG
	::WritePrivateProfileString(L"Debug", L"State", L"USBSettingPage::OnInitDialog() end", PROFILE_PATH->string().c_str());
#endif
	return true;
}

void USBSettingPage::OnCheckboxUSB(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	this->switchControlsByCheckbox();
}

void USBSettingPage::OnButtonReload(unsigned int /*uNotifyCode*/, int /*nID*/, HWND /*wndCtl*/)
{
	this->enumerateDrives();
}

namespace {
std::set<std::wstring> enumerateDeviceName()
{
	std::set<std::wstring> result;
	// ݂ĂS{[擾
	// __H_Volume{xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx}_
	wchar_t buf[MAX_PATH];
	HANDLE h = ::FindFirstVolume(buf, MAX_PATH);
	if (h != INVALID_HANDLE_VALUE) {
		do {
			unsigned int const drive_type = ::GetDriveType(buf);
			if (drive_type == DRIVE_FIXED || drive_type == DRIVE_REMOVABLE) {
				buf[wcslen(buf) - 1] = L'\0';
				wchar_t buf2[MAX_PATH];
				if (::QueryDosDevice(buf + 4, buf2, MAX_PATH) != 0) {
					result.insert(buf2);
				}
			}
		} while (::FindNextVolume(h, buf, MAX_PATH));
		::FindVolumeClose(h);
	}
	return result;
}

std::wstring getFriendlyNameFromDeviceName(std::wstring const &device_name)
{
	std::wstring result;
	// USBhCûݗL
	// foCX}l[WufoCX(ڑ)vł̔ėp{[̐e\
	// Volume{xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx} -> Device Friendly Name
	HDEVINFO hdi = ::SetupDiGetClassDevs(NULL, NULL, NULL, DIGCF_PRESENT | DIGCF_ALLCLASSES);
	if (hdi != INVALID_HANDLE_VALUE) {
		SP_DEVINFO_DATA spdi_data = {sizeof(SP_DEVINFO_DATA)};
		bool is_succeeded = false;
		for (DWORD i = 0; ::SetupDiEnumDeviceInfo(hdi, i, &spdi_data); ++i) {
			wchar_t dev_obj_name[256] = {0};
			DWORD prdt, buf_size;
			::SetupDiGetDeviceRegistryProperty(hdi, &spdi_data,
				SPDRP_PHYSICAL_DEVICE_OBJECT_NAME, &prdt,
				(PBYTE)dev_obj_name, sizeof(dev_obj_name), &buf_size);
			if (lstrcmpi(dev_obj_name, device_name.c_str()) == 0) {
				DEVINST di;
				if (::CM_Get_Parent(&di, spdi_data.DevInst, 0) == CR_SUCCESS) {
					wchar_t buf2[256] = {0};
					::CM_Get_Device_ID(di, buf2, 256, 0);
					HDEVINFO hdi_child = ::SetupDiCreateDeviceInfoList(NULL, NULL);
					SP_DEVINFO_DATA spdi_data2 = {sizeof(SP_DEVINFO_DATA)};
					::SetupDiOpenDeviceInfo(hdi_child, buf2, NULL, 0, &spdi_data2);
					::SetupDiGetDeviceRegistryProperty(hdi_child, &spdi_data2,
						SPDRP_FRIENDLYNAME, &prdt, (PBYTE)buf2, sizeof(buf2), &buf_size);
					if (wcscmp(buf2, L"ROOT\\FTDISK\\0000") != 0) {
						result = buf2;
					}
					is_succeeded = true;
					::SetupDiDestroyDeviceInfoList(hdi_child);
				}
			}
			if (is_succeeded) {
				break;
			}
		}
		::SetupDiDestroyDeviceInfoList(hdi);
	}
	return result;
}
// {p[eBVׂ
bool isBasicPartition(wchar_t drive_letter)
{
	try {
		MyLib::StorageDeviceNumber sdn(drive_letter);
		MyLib::DriveLayoutInformation dli(sdn.getDeviceNumber());
		unsigned int const basic_partition_count =
			std::min(dli.getPartitionCount(), 4U);
		for (unsigned int i = 0; i < basic_partition_count; ++i) {
			unsigned int const partition_num =
				dli.getPartitionEntry(i)->PartitionNumber;
			if (partition_num == sdn.getPartitionNumber()) {
				return true;
			}
		}
	} catch (...) {
		// nop
	}
	return false;
}
bool isValidFileSystem(wchar_t const *drive_root)
{
	// t@CVXe𒲂ׂ
	wchar_t file_system_name[MAX_PATH];
	BOOL ret = ::GetVolumeInformation(drive_root, NULL, 0, NULL, NULL, NULL, file_system_name, MAX_PATH);
	if (ret) {
		if (wcscmp(file_system_name, L"FAT") == 0
		 || wcscmp(file_system_name, L"NTFS") == 0) {
			return true;
		}
	}
	return false;
}
};

// R{{bNXCDhCu񋓂
void USBSettingPage::enumerateDrives(/*int prev_selection*/)
{
	// ݁AIĂ鍀ڂL
	int const drive_selection = this->target_drive_list.GetCurSel();
	TCHAR const selected_drive_letter =
		(drive_selection != -1 ? static_cast<TCHAR>(this->target_drive_list.GetItemData(drive_selection)) : 0);
	// ڂS
	this->target_drive_list.ResetContent();
	// ڐݒpf[^
    struct DriveData {
        TCHAR drive_letter;
        HICON icon;
        std::wstring display_name;
    };
    std::vector<DriveData> cd_drives;

	// ݂{[(Only FIXED and REMOVABLE)
	std::set<std::wstring> device_names = enumerateDeviceName();
	DWORD const drive_mask = ::GetLogicalDrives();
	for (char i = 0; i < 26 && !device_names.empty(); ++i) {
		if (drive_mask & (1 << i)) {
			wchar_t const drive_letter0 = 'A' + i;
			wchar_t const drive_letter[] = {drive_letter0, ':', '\0'};
			wchar_t const root_path[] = {drive_letter0, ':', '\\', '\0'};
			wchar_t buf[MAX_PATH];
			if (::QueryDosDevice(drive_letter, buf, MAX_PATH) != 0) {
				std::set<std::wstring>::iterator it = device_names.find(buf);
				if (it != device_names.end()) {
					if (isBasicPartition(drive_letter0) && isValidFileSystem(root_path)) {
						DriveData drive_data;
						drive_data.drive_letter = drive_letter0;
						SHFILEINFO file_info;
						::SHGetFileInfo(root_path, 0, &file_info, sizeof(SHFILEINFO), SHGFI_DISPLAYNAME | SHGFI_ICON | SHGFI_SMALLICON);
						drive_data.icon = file_info.hIcon;
						drive_data.display_name = file_info.szDisplayName;
						std::wstring friendly_name = getFriendlyNameFromDeviceName(*it);
						if (!friendly_name.empty()) {
							drive_data.display_name += L' ' + friendly_name;
						}
						cd_drives.push_back(drive_data);
					}
					device_names.erase(it);
				}
			}
		}
	}
	// hCũACRC[WXgɒǉ
	CImageList image_list;
	image_list.Create(16, 16, ILC_COLOR8 | ILC_MASK, cd_drives.size(), 0);
	image_list.SetBkColor(RGB(255, 255, 255));
	for (size_t i = 0; i < cd_drives.size(); ++i) {
		image_list.AddIcon(cd_drives[i].icon);
	}
	this->target_drive_list.SetImageList(image_list);
	// hCuꗗɍڂǉ
	int new_drive_selection = 0;
	for (size_t i = 0; i < cd_drives.size(); ++i) {
		COMBOBOXEXITEM cbei;
		cbei.mask = CBEIF_TEXT | CBEIF_IMAGE | CBEIF_SELECTEDIMAGE;
		cbei.pszText = const_cast<wchar_t*>(cd_drives[i].display_name.c_str());
		cbei.cchTextMax = cd_drives[i].display_name.length();
		cbei.iItem = i;
		cbei.iImage = i;
		cbei.iSelectedImage = i;
		int current_item_index = this->target_drive_list.InsertItem(&cbei);
		// ڂ̕tf[^ɃhCu
		this->target_drive_list.SetItemData(current_item_index, cd_drives[i].drive_letter);
		// đIĂhCuƈv鍀
		if (cd_drives[i].drive_letter == selected_drive_letter) {
			new_drive_selection = current_item_index;
		}
	}
	// đIĂhCuĂёI悤ɂ
	const int cd_drive_count = this->target_drive_list.GetCount();
	if (cd_drive_count > 0) {
		this->target_drive_list.SetCurSel(new_drive_selection);
	}
}

void USBSettingPage::switchControlsByCheckbox()
{
	this->DoDataExchange(DDX_SAVE);
	this->target_drive_list.EnableWindow(this->is_checked);
	this->reload_button.EnableWindow(this->is_checked);
	this->setup_drive_list.EnableWindow(this->is_checked);
}
