#include "Driver32CabUpdater.h"
#include "TemporaryDirectory.h"
#include "mylib/exception.h"
#include "mylib/function.h"
#include "mylib/CabCompressor.h"
#include "mylib/CabDecompressor.h"
#include "mylib/FileCompressionInfo.h"
#include "mylib/FileVersionInfo.h"

namespace fs {
	using namespace boost::filesystem;
};

/////////////////////////////////////////////////////////////////////////////

Driver32CabUpdater::Driver32CabUpdater(boost::filesystem::wpath const &work_dir_path,
                                       std::wstring const &amd64_subdir_name)
: work_dir_path_(work_dir_path)
, amd64_subdir_name_(amd64_subdir_name)
{
}

//Driver32CabUpdater::~Driver32CabUpdater(void)
//{
//}

void Driver32CabUpdater::execute()
{
	// ꎞtH_쐬
	TemporaryDirectoryCloseDelete temp_dir_close_delete;
	TemporaryDirectory temp_dir1(this->work_dir_path_, &temp_dir_close_delete);
	fs::wpath const temp_dir1_path = temp_dir1.getDirectoryPath();
	TemporaryDirectory temp_dir2(this->work_dir_path_, &temp_dir_close_delete);
	fs::wpath const temp_dir2_path = temp_dir2.getDirectoryPath();
	// driver32.cab WJ
	static wchar_t const *DRIVER32_CAB_FILENAME = L"driver32.cab";
	fs::wpath const driver32_cab_path = this->work_dir_path_/this->amd64_subdir_name_/DRIVER32_CAB_FILENAME;
	MyLib::CabDecompressor decomp;
	if (!decomp.extractCabFile(driver32_cab_path.string(), temp_dir1_path.string())) {
		throw MY_WIN32_FILE_EXCEPTION_LAST_ERROR(driver32_cab_path.string());
	}
	// I386 ̊et@Cƃo[Wr
	bool needs_recompress = false;
	fs::wpath const i386_dir_path = this->work_dir_path_/L"I386";
	fs::wdirectory_iterator it_end;
	for (fs::wdirectory_iterator it(temp_dir1_path); it != it_end; ++it) {
		fs::wpath const file1_path = it->path();
		fs::wpath const file2_path0 = i386_dir_path/file1_path.filename();
		MyLib::FileCompressionInfo file2_info(file2_path0);
		if (file2_info.exists()) {
			fs::wpath const file2_path = file2_info.decompressOrCopyFile(temp_dir2_path);
			MyLib::FileVersionInfo ver1(file1_path);
			MyLib::FileVersionInfo ver2(file2_path);
			if (ver1.getFileVersion() < ver2.getFileVersion()) {
				MyLib::toFileWritable(file1_path);
				::CopyFile(file2_path.string().c_str(), file1_path.string().c_str(), FALSE);
				needs_recompress = true;
			}
		}
	}

	// ЂƂłÂt@C driver32.cab XV
	if (needs_recompress) {
		fs::wpath const driver_cab_temp_path = MyLib::createTemporaryFilePath(this->work_dir_path_);
		{
			MyLib::CabCompressor compressor(driver_cab_temp_path);
			fs::wdirectory_iterator it_end;
			for (fs::wdirectory_iterator it(temp_dir1_path); it != it_end; ++it) {
				fs::wpath file_path = it->path();
				// LZX21`ňk
				compressor.addFile(file_path, false, TCOMPfromLZXWindow(21));
			}
		}
		// overwrite driver32.cab
		MyLib::toFileWritable(driver32_cab_path);
		::CopyFile(driver_cab_temp_path.string().c_str(), driver32_cab_path.string().c_str(), FALSE);
		::DeleteFile(driver_cab_temp_path.string().c_str());
	}
}
