#include "DriverSetupData.h"
#include "mylib/InfFile.h"
#include "InfFileParser.h"

// INFt@C̓eIuWFNg\zRXgN^
DriverSetupData::DriverSetupData(
	std::wstring const &filename,
	int target_architecture,
	uint32_t target_version)
{
	InfFileParser parser(filename, target_architecture, target_version);
	// INFt@C̃pX
	this->inf_path = parser.getFullPathName();
	this->original_inf_file_title = this->inf_path.filename();
	// Class GUID
	this->class_uuid = MyUUID(parser.getClassGUID());
	// 
	this->provider_name = parser.getProviderName();
	// INFt@C̓t
	this->parseDateString(parser.getDateString());
	// INFt@C̃o[W
	this->inf_date = parseDateString(parser.getDateString());
	this->inf_ver = parseVersionString(parser.getVersionString());
	// Catalogt@C
	this->original_cat_file_title = parser.getCatalogFileTitle();
	std::wstring inf_path = parser.getFullPathName();
	this->catalog_file_path =
		inf_path.substr(0, inf_path.rfind(_T('\\')) + 1) +
		this->original_cat_file_title;
	// ManufacturerZNV
	InfFileParser::ModelsCollection models_collection = parser.getModelsSections();
	// ModelsZNVׂ̂Ă̍s
	DDInstallCollection dd_install_collection;
	for (InfFileParser::ModelsCollection::const_iterator it = models_collection.begin();
		it != models_collection.end();
		++it) {
			std::wstring models_section = it->first + it->second.getVerString();
			DDInstallCollection obj1 = parser.getDDInstallSections(models_section);
			this->ddinstall_collection.insert(obj1.begin(), obj1.end());
	}
	// DDInstall
	for (DDInstallCollection::iterator it = this->ddinstall_collection.begin();
		it != this->ddinstall_collection.end();
		++it) {
			// u.NTvu.NTx86v ȂǂtꂽDDInstallZNV擾
			std::wstring const decorated_section =
				parser.findDecoratedSectionNameWithNT(it->first);
			// DDInstall
			// CopyFiles
			{
				CopyFilesTags tags =
					parser.getCopyFilesTags(decorated_section);
				it->second.copy_files_tags.insert(
					tags.begin(), tags.end());
				// VǉꂽCopyFileTag̕t@Cf[^ǉ
				CopyFilesTags::const_iterator it2 = tags.begin();
				for (; it2 != tags.end(); ++it2) {
					if (this->file_data_collection.find(*it2)
						== this->file_data_collection.end()) {
							this->file_data_collection[*it2] =
								parser.getCopyFilesData(*it2);
					}
				}
			}
			// CopyInf
			CopyInfs copy_infs = parser.getCopyInfs(decorated_section);
			this->copy_infs.insert(copy_infs.begin(), copy_infs.end());
			// DDInstall.CoInstaller
			// CopyFiles
			{
				CopyFilesTags copy_files_tags =
					parser.getCopyFilesTags(decorated_section + _T(".CoInstallers"));
				it->second.copy_files_tags.insert(
					copy_files_tags.begin(), copy_files_tags.end());
				CopyFilesTags::const_iterator it2 = copy_files_tags.begin();
				for (; it2 != copy_files_tags.end(); ++it2) {
					if (this->file_data_collection.find(*it2)
						== this->file_data_collection.end()) {
							this->file_data_collection[*it2] =
								parser.getCopyFilesData(*it2);
					}
				}
			}
			// DDInstall.Service
			ServiceDataCollection::value_type service_data =
				parser.getServiceData(decorated_section + _T(".Services"));
			if (!service_data.first.empty()) {
				it->second.service_name = service_data.first;
				if (this->service_data_collection.find(service_data.first)
					== this->service_data_collection.end()) {
						this->service_data_collection.insert(service_data);
				}
			}
	}

	this->tag_files = parser.getTagFiles();
	this->new_cab_files = parser.getNewCabFiles();
}

// VXeɃCXg[ĂhCȍ񂩂
// IuWFNg\zRXgN^
DriverSetupData::DriverSetupData(
	const DriverSetupData::InstalledDriverInfo &installed_info,
	int target_architecture,
	uint32_t target_version)
{
	InfFileParser parser(
		installed_info.inf_file_title.c_str(),
		target_architecture, target_version);
	// INFt@C̃pX
	this->inf_path = parser.getFullPathName();
	//this->original_inf_file_title = InfFileParser::getOriginalFileTitleFromOEMInf(this->inf_path.string());
	this->original_inf_file_title = parser.getOriginalFileTitleFromOEMInf();
	if (this->original_inf_file_title.empty()) {
		this->original_inf_file_title = installed_info.inf_file_title;
	}
	// Class GUID
	this->class_uuid = MyUUID(parser.getClassGUID());
	// 
	this->provider_name = parser.getProviderName();
	// INFt@C̓tEo[W
	this->inf_date = parseDateString(parser.getDateString());
	std::wstring str001 = parser.getVersionString();
	this->inf_ver = parseVersionString(str001);
	// Catalogt@C
	this->original_cat_file_title = parser.getCatalogFileTitle();
	if (!this->original_cat_file_title.empty()) {
		HCATADMIN cat_admin_handle;
		if (::CryptCATAdminAcquireContext(&cat_admin_handle, NULL, 0)) {
			HANDLE file_handle = ::CreateFileW(this->inf_path.string().c_str(),
				GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
			if (file_handle != INVALID_HANDLE_VALUE) {
				DWORD dwCnt = 100;
				BYTE byHash[100];
				::CryptCATAdminCalcHashFromFileHandle(file_handle, &dwCnt, byHash, 0);
				::CloseHandle(file_handle);

				LPWSTR pszMemberTag = new wchar_t[dwCnt*2 + 1];
				for (DWORD dw = 0; dw < dwCnt; ++dw) {
					::wsprintfW(&pszMemberTag[dw*2], L"%02X", byHash[dw]);
				}

				HCATINFO cat_info_handle = ::CryptCATAdminEnumCatalogFromHash(
					cat_admin_handle, byHash, dwCnt, 0, NULL);
				if (cat_info_handle != NULL) {
					CATALOG_INFO cat_info;
					::CryptCATCatalogInfoFromContext(cat_info_handle, &cat_info, 0);
					this->catalog_file_path = cat_info.wszCatalogFile;
					::CryptCATAdminReleaseCatalogContext(cat_admin_handle, cat_info_handle, 0);
				}
			}
			::CryptCATAdminReleaseContext(cat_admin_handle, 0);
		}
	}

	// ܂AVXeɃCXg[Ă̂ƓDDInstallZNV̏擾

	std::wstring const decorated_section =
		installed_info.inf_section + installed_info.inf_section_ext;
	CopyFilesTags copy_files_tags;
	// DDInstall
	// CopyFiles
	{
		CopyFilesTags tags =
			parser.getCopyFilesTags(decorated_section);
		copy_files_tags.insert(
			tags.begin(), tags.end());
		CopyFilesTags::const_iterator it2 = tags.begin();
		for (; it2 != tags.end(); ++it2) {
			if (this->file_data_collection.find(*it2)
				== this->file_data_collection.end()) {
					this->file_data_collection[*it2] =
						parser.getCopyFilesData(*it2, true);
			}
		}
		// CopyInf
		CopyInfs copy_infs = parser.getCopyInfs(decorated_section, true);
		this->copy_infs.insert(copy_infs.begin(), copy_infs.end());
	}
	// DDInstall.CoInstaller
	// CopyFiles
	{
		CopyFilesTags tags =
			parser.getCopyFilesTags(decorated_section + _T(".CoInstallers"));
		copy_files_tags.insert(tags.begin(), tags.end());
		for (CopyFilesTags::const_iterator it2 = tags.begin(); it2 != tags.end(); ++it2) {
			if (this->file_data_collection.find(*it2)
				== this->file_data_collection.end()) {
					this->file_data_collection[*it2] =
						parser.getCopyFilesData(*it2, true);
			}
		}
	}

	// DDInstallZNVɂĂ\Ȍ擾

	// ManufacturerZNV
	InfFileParser::ModelsCollection models_collection = parser.getModelsSections();
	// ModelsZNVׂ̂Ă̍s
	DDInstallCollection dd_install_collection;
	for (InfFileParser::ModelsCollection::const_iterator it = models_collection.begin();
		it != models_collection.end();
		++it) {
			std::wstring models_section = it->first + it->second.getVerString();
			DDInstallCollection obj1 = parser.getDDInstallSections(models_section);
			this->ddinstall_collection.insert(obj1.begin(), obj1.end());
	}
	// DDInstall
	for (DDInstallCollection::iterator it = this->ddinstall_collection.begin();
		it != this->ddinstall_collection.end();
		++it) {
			// VXeɃCXg[Ăt@CŃJo[ł̂̂ݒǉ

			// u.NTvu.NTx86v ȂǂtꂽDDInstallZNV擾
			std::wstring const decorated_section =
				parser.findDecoratedSectionNameWithNT(it->first);
			// DDInstall
			// CopyFiles
			{
				CopyFilesTags tags =
					parser.getCopyFilesTags(decorated_section);
				{
					bool is_contained = true;
					CopyFilesTags::const_iterator it2 = tags.begin();
					for (; it2 != tags.end(); ++it2) {
						if (this->file_data_collection.find(*it2) == this->file_data_collection.end()) {
							is_contained = false;
						}
					}
					if (!is_contained) {
						continue;
					}
				}
				// CopyInf
				CopyInfs copy_infs = parser.getCopyInfs(decorated_section, true);
				{
					bool is_contained = true;
					CopyInfs::const_iterator it2 = copy_infs.begin();
					for (; it2 != copy_infs.end(); ++it2) {
						if (this->copy_infs.find(it2->first) == this->copy_infs.end()) {
							is_contained = false;
						}
					}
					if (!is_contained) {
						continue;
					}
				}
			}
			// DDInstall.CoInstaller
			// CopyFiles
			{
				CopyFilesTags tags =
					parser.getCopyFilesTags(decorated_section + _T(".CoInstaller"));
				{
					bool is_contained = true;
					CopyFilesTags::const_iterator it2 = tags.begin();
					for (; it2 != tags.end(); ++it2) {
						if (this->file_data_collection.find(*it2) == this->file_data_collection.end()) {
							is_contained = false;
						}
					}
					if (!is_contained) {
						continue;
					}
				}
			}
			// DDInstall.Service
			ServiceDataCollection::value_type service_data =
				parser.getServiceData(decorated_section + _T(".Services"));
			if (!service_data.first.empty()) {
				it->second.service_name = service_data.first;
				if (this->service_data_collection.find(service_data.first)
					== this->service_data_collection.end()) {
						this->service_data_collection.insert(service_data);
				}
			}
	}

	this->tag_files = parser.getTagFiles();
	this->new_cab_files = parser.getNewCabFiles();
}

// INFt@Cɏꂽt𐔒lɕϊ
uint32_t DriverSetupData::parseDateString(const std::wstring& str)
{
	// INFt@Cɏꂽt𐔒lɕϊ
	if (str.length() >= 10) {	// mm/dd/yyyy
		std::wstring month_str = str.substr(0, 2);
		unsigned int month = ::_tstoi(str.c_str());
		std::wstring day_str = str.substr(3, 2);
		unsigned int day = ::_tstoi(day_str.c_str());
		std::wstring year_str = str.substr(6, 4);
		unsigned int year = ::_tstoi(year_str.c_str());
		return
			((year << 16) & 0xFFFF0000) | ((month << 8) & 0xFF00) | (day & 0xFF);
	}
	return 0;
}
// INFt@Cɏꂽo[W𐔒lɕϊ
uint64_t DriverSetupData::parseVersionString(const std::wstring& str)
{
	const size_t str_length = str.length();
	// INFt@Cɏꂽo[W𐔒lɕϊ
	std::wstring::size_type period_pos = 0;
	uint64_t result = 0;
	for (unsigned int i = 4 - 1; i >= 0; --i) {
		const std::wstring::size_type target_pos = period_pos;
		period_pos = str.find(_T('.'), target_pos);
		const std::wstring::size_type length =
			(period_pos != std::wstring::npos ? period_pos - target_pos : std::wstring::npos);
		const uint64_t ver_part = ::_tstoi(str.substr(target_pos, length).c_str());
		result |= ((ver_part & 0xFFFF) << (i*16));
		if (period_pos == std::wstring::npos || str_length <= period_pos + 1) {
			break;
		}
		++period_pos;
	}
	return result;
}

boost::filesystem::wpath const& DriverSetupData::getInfFilePath() const
{
	return this->inf_path;
}

const std::wstring& DriverSetupData::getOriginalInfFileTitle() const
{
	return this->original_inf_file_title;
}

const std::wstring& DriverSetupData::getProviderName() const
{
	return this->provider_name;
}

const MyUUID& DriverSetupData::getClassUUID() const
{
	return this->class_uuid;
}

uint32_t DriverSetupData::getInfFileDate() const
{
	return this->inf_date;
}

uint64_t DriverSetupData::getInfFileVersion() const
{
	return this->inf_ver;
}

boost::filesystem::wpath const& DriverSetupData::getCatalogFilePath() const
{
	return this->catalog_file_path;
}

const std::wstring& DriverSetupData::getOriginalCatalogFileTitle() const
{
	return this->original_cat_file_title;
}

const DriverSetupData::DDInstallCollection& DriverSetupData::getDDInstallSectionCollection() const
{
	return this->ddinstall_collection;
}

const DriverSetupData::FileDataCollection& DriverSetupData::getFileDataCollection() const
{
	return this->file_data_collection;
}

const DriverSetupData::CopyInfs& DriverSetupData::getCopyInfs() const
{
	return this->copy_infs;
}

DriverSetupData::TagFileCollection DriverSetupData::getTagFiles() const
{
	return this->tag_files;
}

DriverSetupData::NewCabFileCollection DriverSetupData::getNewCabFiles() const
{
	return this->new_cab_files;
}

const DriverSetupData::ServiceDataCollection& DriverSetupData::getServiceDataCollection() const
{
	return this->service_data_collection;
}

// foCXp̃hCoǂԂ
bool DriverSetupData::isForSameDevice(const DriverSetupData &other_instance) const
{
	if (this->class_uuid == other_instance.class_uuid) {
		DriverSetupData::DDInstallCollection::const_iterator it1 =
			this->ddinstall_collection.begin();
		for (; it1 != this->ddinstall_collection.end(); ++it1) {
			DriverSetupData::HardwareIDCollection::const_iterator hwid_it =
				it1->second.hardware_ids.begin();
			for (; hwid_it != it1->second.hardware_ids.end(); ++hwid_it) {
				DriverSetupData::DDInstallCollection::const_iterator it2 =
					other_instance.ddinstall_collection.begin();
				for (; it2 != other_instance.ddinstall_collection.end(); ++it2) {
					if (it2->second.hardware_ids.find(*hwid_it) != it2->second.hardware_ids.end()) {
						// vHardwareID
						return isForSameDevice2(other_instance);
					}
				}
			}
		}
	}
	return false;
}

bool DriverSetupData::isForSameDevice2(const DriverSetupData &other_instance) const
{
	namespace fs = boost::filesystem;

	ServiceDataCollection const &sdc1 = this->getServiceDataCollection();
	ServiceDataCollection const &sdc2 = other_instance.getServiceDataCollection();
	if (sdc1.size() > 0 && sdc2.size() > 0) {
		// F6 driver
		return true;
	}

	if (MyLib::compareStringNoCase(this->getOriginalCatalogFileTitle(), other_instance.getOriginalCatalogFileTitle()) == 0) {
		// same catalog file
		return true;
	}

	// file
	FileDataCollection const &fdc1 = this->getFileDataCollection();
	FileDataCollection const &fdc2 = other_instance.getFileDataCollection();
	for (FileDataCollection::const_iterator it11 = fdc1.begin(); it11 != fdc1.end(); ++it11) {
		CopyFilesTagData const &cftd1 = it11->second;
		for (CopyFilesTagData::const_iterator it12 = cftd1.begin(); it12 != cftd1.end(); ++it12) {
			for (FileDataCollection::const_iterator it21 = fdc2.begin(); it21 != fdc2.end(); ++it21) {
				CopyFilesTagData const &cftd2 = it21->second;
				for (CopyFilesTagData::const_iterator it22 = cftd2.begin(); it22 != cftd2.end(); ++it22) {
					fs::wpath const file_path1 = it12->second.source_path;
					fs::wpath const file_path2 = it22->second.source_path;
					if (MyLib::compareStringNoCase(file_path1.filename(), file_path2.filename()) == 0) {
						return true;
					}
				}
			}
		}
	}

	return false;
}
