#pragma once

#include "mylib/function.h"

class MyUUID :
	public UUID
{
public:
	MyUUID()
	{
		this->Data1 = 0;
		this->Data2 = 0;
		this->Data3 = 0;
		::memset(this->Data4, 0, sizeof(UUID::Data4));
	}

	MyUUID(const UUID& uuid)
	{
		*static_cast<UUID*>(this) = uuid;
	}

	MyUUID(std::wstring const &guid)
	{
		if (guid.empty()) {
			this->Data1 = 0;
			this->Data2 = 0;
			this->Data3 = 0;
			::memset(this->Data4, 0, sizeof(UUID::Data4));
			return;
		}

		const size_t guid_length = guid.length();
		static const size_t EXPECTED_GUID_LENGTH = 36;
		TCHAR buf[EXPECTED_GUID_LENGTH + 1];
		switch (guid_length) {
		case EXPECTED_GUID_LENGTH:
			::memcpy(buf, guid.c_str(), EXPECTED_GUID_LENGTH*sizeof(TCHAR));
			break;
		case EXPECTED_GUID_LENGTH + 2:	// "{}"
			::memcpy(buf, guid.c_str() + 1, EXPECTED_GUID_LENGTH*sizeof(TCHAR));
			break;
		//case EXPECTED_GUID_LENGTH - 4:	// "-"Ȃ
		//	break;
		default:
			// ̃jo[TŗLʎq (UUID) łB
			throw MY_WIN32_EXCEPTION(0x06A9);
		}
		buf[EXPECTED_GUID_LENGTH] = '\0';
		UUID temp;
		const long ret = ::UuidFromString(reinterpret_cast<RPC_WSTR>(buf), &temp);
		if (ret != NO_ERROR) {
			throw MY_WIN32_EXCEPTION(ret);
		}
		*this = temp;
	}

	bool operator==(const UUID& other_instance)
	{
		return this->Data1 == other_instance.Data1
			&& this->Data2 == other_instance.Data2
			&& this->Data3 == other_instance.Data3
			&& ::memcmp(this->Data4, other_instance.Data4, sizeof(UUID::Data4)) == 0;
	}

	bool operator!=(const UUID& other_instance)
	{
		return !(*this == other_instance);
	}
};


class DriverSetupData
{
public:
	// INFt@C̓eIuWFNg\zRXgN^
	DriverSetupData(
		std::wstring const &filename,
		int architecture_id,
		uint32_t target_version);
	// VXeɃCXg[ĂhCȍ
	struct InstalledDriverInfo {
		std::wstring inf_file_title;
		std::wstring inf_section;
		std::wstring inf_section_ext;
	};
	// VXeɃCXg[ĂhCȍ񂩂
	// IuWFNg\zRXgN^
	DriverSetupData(
		const InstalledDriverInfo& installed_info,
		int architecture_id,
		uint32_t target_version);

	// Rs[t@C̃f[^
	struct FileData {
		boost::filesystem::wpath source_path;  // ݂̃t@C̍ݏ(tpX)
		std::wstring installation_path;        // Installation Root ̑΃pX
	};
	// t@Ct@C̃Rs[Ɋւ
	typedef std::map<std::wstring, FileData, MyLib::LessNoCase> CopyFilesTagData;
	typedef std::map<
		std::wstring,
		CopyFilesTagData,
		MyLib::LessNoCase> FileDataCollection;
	typedef std::map<std::wstring, std::wstring, MyLib::LessNoCase> CopyInfs;

	// CopyFiles
	typedef std::set<std::wstring, MyLib::LessNoCase> CopyFilesTags;
	// Description
	typedef std::set<std::wstring, MyLib::LessNoCase> DescriptionCollection;
	// HardwareID
	typedef std::set<std::wstring, MyLib::LessNoCase> HardwareIDCollection;

	// DDInstall̃f[^
	struct DDInstallData {
		// hCo\t@C̏W
		CopyFilesTags copy_files_tags;
		// hCo̊ȒPȐ̏W
		DescriptionCollection descriptions;
		// n[hEFAID̏W
		HardwareIDCollection hardware_ids;
		// T[rX
		std::wstring service_name;
	};
	typedef std::map<std::wstring, DDInstallData, MyLib::LessNoCase> DDInstallCollection;

	// WXgf[^
	struct RegData {
		HKEY root;	// HKEY_LOCAL_MACHINEƂ
		std::wstring subkey;		// L[
		std::wstring entry;	// O
		uint32_t flags;		// 
		std::vector<char> value;		// f[^
	};
	typedef std::vector<RegData> RegDataCollection;

	struct ServiceData {
		std::wstring service_binary;	// SCSI Miniport hCo(.syst@C)̃t@C TXTSETUP.SIF [SCSI.Load]
		RegDataCollection add_regs;
	};
	// T[rXT[rX̏
	typedef std::map<std::wstring, ServiceData, MyLib::LessNoCase> ServiceDataCollection;
	// [SourceDisksNames] tag-file
	typedef std::map<std::wstring, std::wstring> TagFileCollection;
	// [SourceDisksNames] cab-file
	typedef std::map<std::wstring, boost::filesystem::wpath, MyLib::LessNoCase> FileInCabInfo;
	typedef std::map<std::wstring, FileInCabInfo, MyLib::LessNoCase> NewCabFileCollection;

	boost::filesystem::wpath const& getInfFilePath() const;

	const std::wstring& getOriginalInfFileTitle() const;

	const std::wstring& getProviderName() const;

	const MyUUID& getClassUUID() const;

	uint32_t getInfFileDate() const;

	uint64_t getInfFileVersion() const;

	boost::filesystem::wpath const& getCatalogFilePath() const;

	const std::wstring& getOriginalCatalogFileTitle() const;

	const DDInstallCollection& getDDInstallSectionCollection() const;

	const FileDataCollection& getFileDataCollection() const;

	const CopyInfs& getCopyInfs() const;

	const ServiceDataCollection& getServiceDataCollection() const;

	TagFileCollection getTagFiles() const;
	NewCabFileCollection getNewCabFiles() const;

	// foCXp̃hCoǂԂ
	bool isForSameDevice(const DriverSetupData& other_instance) const;
	bool isForSameDevice2(const DriverSetupData& other_instance) const;

private:
	// INFt@C̃pX
	boost::filesystem::wpath inf_path;
	// INFt@C
	std::wstring original_inf_file_title;
	// 
	std::wstring provider_name;
	// Class GUID
	MyUUID class_uuid;
	// INFt@C̍XV
	// ʂAN16bitA8bitA8bit
	uint32_t inf_date;
	// INFt@C̃o[W
	uint64_t inf_ver;
	// CATt@C̃pX
	boost::filesystem::wpath catalog_file_path;
	std::wstring original_cat_file_title;
	// DDInstall
	DDInstallCollection ddinstall_collection;
	// Rs[t@C̈ꗗ
	FileDataCollection file_data_collection;
	CopyInfs copy_infs;
	// SCSI Miniport T[rXɊ֘AÂꂽhCof[^
	ServiceDataCollection service_data_collection;

	//
	TagFileCollection tag_files;
	NewCabFileCollection new_cab_files;

	// INFt@CɏꂽtEo[W𐔒lɕϊ
	static uint32_t parseDateString(const std::wstring& str);
	static uint64_t parseVersionString(const std::wstring& str);
};
