#include "mylib/function.h"
#include "mylib/InfFile.h"
#include "DriverSetupData.h"
#include "InfFileParser.h"
#include "mylib/FileCompressionInfo.h"

using MyLib::compareStringNoCase;
using MyLib::InfContextIterator;

// RXgN^
InfFileParser::InfFileParser(
		const std::wstring& inf_file_name,
		int target_architecture,
		uint32_t target_version)
: inf_file(inf_file_name.c_str())
, version(target_version)
{
	
	switch (target_architecture) {
	case PROCESSOR_ARCHITECTURE_INTEL:
		this->architecture_str = _T("X86");
		break;
	case PROCESSOR_ARCHITECTURE_AMD64:
		this->architecture_str = _T("AMD64");
		break;
	case PROCESSOR_ARCHITECTURE_IA64:
		this->architecture_str = _T("IA64");
		break;
	default:
		throw std::runtime_error("̃vZbTɂ͑ΉĂ܂");
	}
}

std::wstring InfFileParser::getFullPathName()
{
	return this->inf_file.getFullPathName();
}

std::wstring InfFileParser::getProviderName()
{
	return this->inf_file.getVersionInfo(_T("Provider"));
}
std::wstring InfFileParser::getClassGUID()
{
	return this->inf_file.getVersionInfo(_T("ClassGuid"));
}
std::wstring InfFileParser::getDateString()
{
	return this->inf_file.getVersionInfo(_T("DriverVer"));
}
std::wstring InfFileParser::getVersionString()
{
	const InfContextIterator inf_end;
	InfContextIterator ver_key_it =
		this->inf_file.findFirstLine(_T("Version"), _T("DriverVer"));
	if (ver_key_it != inf_end && ver_key_it->getFieldCount() >= 2) {
		return ver_key_it->getStringField(2);
	}
	return L"";
}

std::wstring InfFileParser::getCatalogFileTitle()
{
	std::wstring key[] = {
		_T("CatalogFile"),
		_T("CatalogFile.NT"),
		_T("CatalogFile.NT") + this->architecture_str,
	};

	const InfContextIterator inf_end;
	for (int i = _countof(key) - 1; i >= 0; --i) {
		InfContextIterator ver_key_it =
			this->inf_file.findFirstLine(_T("Version"), key[i].c_str());
		if (ver_key_it != inf_end) {
			return ver_key_it->getStringField(1);
		}
	}
	return _T("");
}

//
InfFileParser::ModelsCollection InfFileParser::getModelsSections()
{
	ModelsCollection models_collection;
	const InfContextIterator inf_end;
	const ModelsTargetOSVer models_no_decorated;
	const unsigned int no_decorated_architecture_flags =
		models_no_decorated.getArchitectureFlag();
	// ^[Qbg CPU Architecture tO (X86, AMD64, IA64)
	const unsigned int target_architecture_flag =
		ModelsTargetOSVer::getArchitectureFlagFromString(this->architecture_str);

	InfContextIterator inf_it = this->inf_file.findFirstLine(_T("Manufacturer"));
	for (; inf_it != inf_end; ++inf_it) {
		unsigned int field_count = inf_it->getFieldCount();
		// o[WȂǂtĂȂModels Section擾
		std::wstring models_base = inf_it->getStringField(field_count > 0 ? 1 : 0);
		ModelsCollection::iterator it = models_collection.find(models_base);
		if (it == models_collection.end()
			&& (no_decorated_architecture_flags & target_architecture_flag)) {
				it = models_collection.insert(
					std::make_pair(models_base, models_no_decorated)).first;
		}
		// XP(Ver.5.1)ȍ~TargetOSVersionw肳ĂƂ
		if (field_count >= 2 && this->version >= 0x00050001) {
			// TargetOSVersion(tB[hȍ~)̉
			for (unsigned int i = 2; i <= field_count; ++i) {
				ModelsTargetOSVer new_os_ver(inf_it->getStringField(i));
				// ɋ߂̂
				if (it == models_collection.end()) {
					if (matchesTargetMoreClosely(
						new_os_ver.getArchitectureFlag(),
						new_os_ver.getVersion(),
						models_no_decorated.getArchitectureFlag(),
						models_no_decorated.getVersion())) {
							it = models_collection.insert(
								std::make_pair(models_base, new_os_ver)).first;
					}
				} else {
					if (matchesTargetMoreClosely(
						new_os_ver.getArchitectureFlag(),
						new_os_ver.getVersion(),
						it->second.getArchitectureFlag(),
						it->second.getVersion())) {
							it->second = new_os_ver;
					}
				}
			}
		}
	}
	return models_collection;
}

//
bool InfFileParser::matchesTargetMoreClosely(
	unsigned int new_architecture_flag, uint32_t new_version,
	unsigned int current_architecture_flag, uint32_t current_version)
{
	unsigned int this_architecture_flag =
		ModelsTargetOSVer::getArchitectureFlagFromString(this->architecture_str);
	if ((new_architecture_flag & this_architecture_flag)) {
		if ((current_version == new_version && current_architecture_flag > new_architecture_flag) ||
			(current_version < new_version && new_version <= this->version)) {
				return true;
		}
	}
	return false;
}

// DDInstallZNVƂɕtDevice descriptionEHardwareID擾
DriverSetupData::DDInstallCollection
InfFileParser::getDDInstallSections(const std::wstring& models_section)
{
	DriverSetupData::DDInstallCollection result;

	InfContextIterator inf_it = this->inf_file.findFirstLine(models_section.c_str());
	const InfContextIterator inf_end;
	for (; inf_it != inf_end; ++inf_it) {
		// foCX̖O(L[)
		const std::wstring device_description = inf_it->getStringField(0);
		// DDInstallZNV(tB[hP)
		const std::wstring dd_install_section = inf_it->getStringField(1);
		// Hardware ID(tB[hQ)
		const std::wstring hardeware_id = inf_it->getStringField(2);
		// Device description̒ǉ
		result[dd_install_section].descriptions.insert(device_description);
		// HardwareID̒ǉ
		result[dd_install_section].hardware_ids.insert(hardeware_id);
	}
	return result;
}

// w肳ꂽZNVɁA.x86/.amd64/.ia64tꂽ̂ȂTB
InfContextIterator InfFileParser::findDecoratedSection(LPCTSTR section, LPCTSTR key)
{
	// SectionName.x86
	std::wostringstream ss;
	ss << section << _T('.') << this->architecture_str;

	std::wstring candidates[] = {ss.str(), section};

	const InfContextIterator it_end;
	for (size_t i = 0; i < _countof(candidates); ++i) {
		const InfContextIterator it = this->inf_file.findFirstLine(candidates[i].c_str(), key);
		if (it != it_end) {
			return it;
		}
	}
	return it_end;
}

// w肳ꂽZNVɁA.NT/.NTx86/.NTamd64/.NTia64tꂽ̂ȂTB
std::wstring InfFileParser::findDecoratedSectionNameWithNT(const std::wstring& section)
{
	std::vector<std::wstring> candidates;
	// SectionName.NTx86
	std::wostringstream ss;
	ss << section << _T(".NT") << this->architecture_str;
	candidates.push_back(ss.str());
	// SectionName.NT
	candidates.push_back(section + _T(".NT"));
	// SectionName
	candidates.push_back(section);

	std::vector<std::wstring>::const_iterator it = candidates.begin();
	for (; it != candidates.end(); ++it) {
		const InfContextIterator inf_end;
		if (this->inf_file.findFirstLine(it->c_str()) != inf_end) {
			return *it;
		}
	}
	return _T("");
}

// 
DriverSetupData::ServiceDataCollection::value_type
InfFileParser::getServiceData(const std::wstring& dd_install_section)
{
	enum AddServiceFields {
		SERVICE_NAME_FIELD	= 1,
		ADDSERVICE_FLAGS_FIELD,
		SERVICE_INSTALL_SECTION_FIELD,
		EVENTLOG_INSTALL_SECTION_FIELD,
		EVENTLOG_TYPE_FIELD,
		EVENT_NAME_FIELD
	};

	DriverSetupData::ServiceDataCollection::value_type result;

	const InfContextIterator inf_end;
	InfContextIterator inf_it =
		this->inf_file.findFirstLine(dd_install_section.c_str(), _T("AddService"));
	// DDInstall.ServiceZNVAddServiceL[T
	for (; inf_it != inf_end; inf_it = this->inf_file.findNextMatchLine(inf_it, _T("AddService"))) {
		if (inf_it->getFieldCount() < SERVICE_INSTALL_SECTION_FIELD) {
			continue;
		}
		// T[rX擾
		const std::wstring service_name = inf_it->getStringField(SERVICE_NAME_FIELD);
		// Service installZNV擾
		const std::wstring service_install_section =
			inf_it->getStringField(SERVICE_INSTALL_SECTION_FIELD);

		// Service installZNVLoadOrderGroupL[T
		InfContextIterator inf_it2 = this->inf_file.findFirstLine(
			service_install_section.c_str(), _T("LoadOrderGroup"));
		if (inf_it2 != inf_end) {
			// LoadOrderGroupL[̒l擾
			const std::wstring load_order_group = inf_it2->getStringField(1);
			// 啶EʂȂŔr邽߁Aɓ
			if (MyLib::compareStringNoCase(load_order_group, L"SCSI Miniport") == 0 || MyLib::compareStringNoCase(load_order_group, L"System Bus Extender") == 0) {
				// LoadOrderGroupL[̒l SCSI Miniport ܂ System Bus Extender Ȃ΁A
				// ServiceBinaryL[T
				InfContextIterator inf_it3 = this->inf_file.findFirstLine(
					service_install_section.c_str(), _T("ServiceBinary"));
				if (inf_it3 != inf_end) {
					const_cast<std::wstring&>(result.first) = service_name;
					// ServiceBinary
					const std::wstring service_binary_path = inf_it3->getStringField(1);
					const std::wstring::size_type filename_offset =
						service_binary_path.rfind(_T('\\')) + 1;
					result.second.service_binary =
						service_binary_path.substr(filename_offset);
					// AddReg
					result.second.add_regs =
						this->getAddregData(service_install_section, service_name);
					break;
				}
			}
		}
	}
	return result;
}

// WXg̒l擾
template <int VALUE_FIELD>
std::vector<char> getAddregValue(uint32_t flags, MyLib::InfContextIterator section_it)
{
	//static const int VALUE_FIELD = 5;
	std::vector<char> result;
	switch (flags & FLG_ADDREG_TYPE_MASK) {
	case FLG_ADDREG_TYPE_BINARY:
		{
			result = section_it->getBinaryField(VALUE_FIELD);
		}
		break;
	case FLG_ADDREG_TYPE_DWORD:
		{
			result.resize(sizeof(uint32_t));
			uint32_t* value_ptr = reinterpret_cast<uint32_t*>(&result[0]);
			*value_ptr = section_it->getIntField(VALUE_FIELD);
		}
		break;
	case FLG_ADDREG_TYPE_SZ:
	case FLG_ADDREG_TYPE_EXPAND_SZ:
		{
			const std::wstring str = section_it->getStringField(VALUE_FIELD);
			const size_t str_size = (str.length() + 1)*sizeof(TCHAR);
			result.resize(str_size);
			::memcpy(&result[0], str.c_str(), str_size);
		}
		break;
	case FLG_ADDREG_TYPE_MULTI_SZ:
		{
			// ̏W
			const std::vector<std::wstring> strs =
				section_it->getMultiSzField(VALUE_FIELD);
			// ̌
			const size_t str_count = strs.size();
			// ẽTCY(oCgP)擾
			std::vector<size_t> strs_size(str_count);
			size_t total_size = 0;
			for (size_t i = 0; i < str_count; ++i) {
				const size_t str_size =
					(strs[i].length() + 1)*sizeof(TCHAR);
				strs_size[i] = str_size;
				total_size += str_size;
			}
			result.resize(total_size + sizeof(TCHAR));
			// obt@ɃRs[
			size_t j = 0;
			for (size_t i = 0; i < str_count; ++i) {
				::memcpy(&result[j], strs[i].c_str(), strs_size[i]);
				j += strs_size[i];
			}
#ifdef UNICODE
			*reinterpret_cast<wchar_t*>(&result[j]) = L'\0';
#else
			result[j] = '\0';
#endif
		}
		break;
	default:
		break;
	}
	return result;
}

// 
DriverSetupData::RegDataCollection
InfFileParser::getAddregData(
	const std::wstring& dd_install_service,
	const std::wstring& service_name)
{
	enum AddregFields {
		REG_ROOT_FIELD	= 1,
		SUBKEY_FIELD,
		VALUE_ENTRY_NAME_FIELD,
		ADDREG_FLAGS_FIELD,
		VALUE_FIELD
	};

	DriverSetupData::RegDataCollection result;
	const InfContextIterator inf_end;

	// AddReg Directive
	InfContextIterator directive_it = this->inf_file.findFirstLine(
		dd_install_service.c_str(), _T("AddReg"));
	for (; directive_it != inf_end; directive_it = this->inf_file.findNextMatchLine(directive_it, _T("AddReg"))) {
		std::wstring addreg_section = directive_it->getStringField(1);
		// add-registory-section
		InfContextIterator section_it = this->inf_file.findFirstLine(addreg_section.c_str());
		for (; section_it != inf_end; ++section_it) {
			uint32_t field_count = section_it->getFieldCount();
			// ex.
			//	HKR,"Parameters\PnpInterface","5",0x00010001(FLG_ADDREG_TYPE_DWORD),0x00000001
			DriverSetupData::RegData reg_data;
			// INFt@Cł̃[gL[
			std::wstring inf_root = section_it->getStringField(REG_ROOT_FIELD);
			if (compareStringNoCase(inf_root, _T("HKR")) == 0) {
				// ݂́AService Install SectionłgpȂ̂ŁA
				// HKLM\SYSTEM\CurrentControlSet\Services\<Service Name>Ō߂
				reg_data.root = HKEY_LOCAL_MACHINE;
				//std::wostringstream ss;
				//ss << _T("ControlSet001\\Services\\") << service_name;
				//std::wstring key_name = L"ControlSet001\\Services\\";
				std::wstring key_name = service_name;
				if (field_count >= SUBKEY_FIELD) {
					//ss << _T('\\') << section_it->getStringField(SUBKEY_FIELD);
					std::wstring const subkey_name = section_it->getStringField(SUBKEY_FIELD);
					if (!subkey_name.empty()) {
						key_name += L'\\';
						key_name += subkey_name;
					}
				}
				reg_data.subkey = key_name;//ss.str();
			//} else if (inf_root == _T("HKLM") && field_count >= SUBKEY_FIELD) {
			//	std::wstring subkey = section_it->getStringField(SUBKEY_FIELD);
			//	// Setupreg.hivɕۑĂL[HKLM\SYSTEMȉ
			//	std::wstring::size_type backslash_pos = subkey.find(_T('\\'));
			//	if (compareStringNoCase(subkey.substr(0, backslash_pos), _T("SYSTEM")) == 0) {
			//		reg_data.root = HKEY_LOCAL_MACHINE;
			//		if (backslash_pos != std::wstring::npos) {
			//			reg_data.subkey = subkey.substr(backslash_pos + 1);
			//		}
			//	} else {
			//		continue;
			//	}
			} else {
				continue;
			}
			try {
				// Entry
				if (field_count >= VALUE_FIELD) {
					reg_data.entry = section_it->getStringField(VALUE_ENTRY_NAME_FIELD);
					// type & value
					if (field_count >= ADDREG_FLAGS_FIELD) {
						reg_data.flags = section_it->getIntField(ADDREG_FLAGS_FIELD);
						reg_data.value =
							getAddregValue<VALUE_FIELD>(reg_data.flags, section_it);
					}
				}
				//
				result.push_back(reg_data);
			} catch (...) {
				// nop
				// ُ킪ꍇ͖(si3112r.infƂ)
			}
		}
	}
	return result;
}

// 
std::set<std::wstring, MyLib::LessNoCase>
InfFileParser::getCopyFilesTags(const std::wstring& dd_install_section)
{
	std::set<std::wstring, MyLib::LessNoCase> result;
	const InfContextIterator inf_end;
	InfContextIterator directive_it = this->inf_file.findFirstLine(
		dd_install_section.c_str(), _T("CopyFiles"));
	// ׂĂ CopyFiles Directive ɂď
	for (; directive_it != inf_end; directive_it = this->inf_file.findNextMatchLine(directive_it, _T("CopyFiles"))) {
		// CopyFiles Directive ׂ̂ẴtB[h(E)
		for (unsigned i = 1; i <= directive_it->getFieldCount(); ++i) {
			std::wstring const str_field = directive_it->getStringField(i);
			if (!str_field.empty()) {
				result.insert(str_field);
			}
		}
	}
	return result;
}

DriverSetupData::CopyFilesTagData
InfFileParser::getCopyFilesData(const std::wstring& copy_files_tag, bool is_installed_file)
{
	assert(!copy_files_tag.empty());

	namespace fs = boost::filesystem;

	const InfContextIterator inf_end;
	DriverSetupData::CopyFilesTagData result;
	if (*copy_files_tag.begin() == _T('@')) {
		// P̂̃t@C
		//it@CEt@CfBNg/Cabt@C̃tpXECXg[DirIDj
		// [SourceDisksFiles]
		// xxx1.sys = 1
		// xxx2.dll = 1
		//
		// [SourceDisksNames]
		// 1 = "````````", <tagfile_or_cabfile>,,<tagfile_xp>
		//
		// [DestinationDirs]
		// DefaultDestDir = 12
		//
		//InfContextIterator destination_it = this->inf_file.findFirstLine(
		//	_T("DestinationDirs"), _T("DefaultDestDir"));
		//if (destination_it == inf_end) {
		//	throw MyLib::SetupAPIError(HRESULT_FROM_SETUPAPI(ERROR_LINE_NOT_FOUND));
		//}
		//// Dirid擾(P̂̃t@ĈƂ"DefaultDestDir")
		//const int dirid = destination_it->getIntField(1);
		// t@C̐擪́u@vƂ
		const std::wstring filename = copy_files_tag.substr(1);
		// Rs[t@CXgɒǉ
		if (is_installed_file) {
			DriverSetupData::FileData file_data;
			//file_data.dirid = dirid;
			DriverSetupData::FileData *ptr =
				this->getInstalledFileData(filename, this->inf_file.getTargetPath()/filename);
			//file_data.temporary = false;
			if (ptr != NULL) {
				result[filename] = *ptr;
			}
			delete ptr;
		}else {
			DriverSetupData::FileData *ptr = this->getFileData(filename);
			if (ptr != NULL) {
				result[filename] = *ptr;
			}
			delete ptr;
		}
	} else {
		// ̃t@C(file-list-section)
		//it@CEt@CfBNg/Cabt@C̃tpXECXg[DirIDj
		// [<CopyFilesSection>]
		// xxx1.sys
		// xxx2.dll
		//
		// [SourceDisksFiles]
		// xxx1.sys = 1
		// xxx2.dll = 1
		//
		// [SourceDisksNames]
		// 1 = "````````", <tagfile_or_cabfile>,,<tagfile_xp>
		//
		// [DestinationDirs]
		// DefaultDestDir = 12
		// <CopyFilesSection> = 11
		//
		//InfContextIterator destination_it = this->inf_file.findFirstLine(
		//	_T("DestinationDirs"), copy_files_tag.c_str());
		//if (destination_it == inf_end) {
		//	// file-list-sectionDirid`ĂȂꍇ"DefaultDestDir"
		//	destination_it = this->inf_file.findFirstLine(
		//		_T("DestinationDirs"), _T("DefaultDestDir"));
		//	if (destination_it == inf_end) {
		//		throw MyLib::SetupAPIError(HRESULT_FROM_SETUPAPI(ERROR_LINE_NOT_FOUND));
		//	}
		//}
		//// Dirid擾
		//const int dirid = destination_it->getIntField(1);
		// file-list-section̏
		InfContextIterator file_list_it =
			this->inf_file.findFirstLine(copy_files_tag.c_str());
		if (is_installed_file) {
			fs::wpath const target_path =
				this->inf_file.getTargetPath(copy_files_tag.c_str());
			for (; file_list_it != inf_end; ++file_list_it) {
				// CXg[t@C擾
				const std::wstring dest_file_name = file_list_it->getStringField(1);
				// Rs[t@CXgɒǉ
				DriverSetupData::FileData file_data;
				//file_data.dirid = dirid;
				// CXg[t@C
				std::wstring src_file_name;
				if (file_list_it->getFieldCount() >= 2) {
					src_file_name = file_list_it->getStringField(2);
				}
				if (src_file_name.empty()) {
					// tB[h < 2 ܂ tB[h2 Ȃ
					// CXg[ƃCXg[̃t@C͓
					src_file_name = dest_file_name;
				}
				DriverSetupData::FileData *ptr =
					this->getInstalledFileData(src_file_name, target_path/dest_file_name);
				//file_data.temporary = false;
				if (ptr != NULL) {
					result[src_file_name] = *ptr;
				}
				delete ptr;
			}
		} else {
			for (; file_list_it != inf_end; ++file_list_it) {
				//t@C擾
				std::wstring src_file_name;
				if (file_list_it->getFieldCount() >= 2) {
					src_file_name = file_list_it->getStringField(2);
				}
				const std::wstring filename =
					(src_file_name.empty() ? file_list_it->getStringField(1) : src_file_name);
				// Rs[t@CXgɒǉ
				DriverSetupData::FileData *ptr = this->getFileData(filename);
				if (ptr != NULL) {
					result[filename] = *ptr;
				}
				delete ptr;
			}
		}
	}
	return result;
}

namespace {
	std::wstring getOriginalFileTitleFromOEMInf0(SP_INF_INFORMATION *inf_info)
	{
		std::wstring result;

		//BOOL
	//WINAPI
	//SetupQueryInfOriginalFileInformationW(
	//    __in PSP_INF_INFORMATION InfInformation,
	//    __in UINT InfIndex,
	//    __in_opt PSP_ALTPLATFORM_INFO AlternatePlatformInfo,
	//    __out PSP_ORIGINAL_FILE_INFO_W OriginalFileInfo
	//    );

		// SP_ORIGINAL_FILE_INFO
		SP_ORIGINAL_FILE_INFO org_file_info;
		org_file_info.cbSize = sizeof(SP_ORIGINAL_FILE_INFO);

		static OSVERSIONINFO *os_ver_info_ptr = NULL;
		if (os_ver_info_ptr == NULL) {
			os_ver_info_ptr = new OSVERSIONINFO();
			os_ver_info_ptr->dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
			::GetVersionEx(os_ver_info_ptr);

		}
		ULARGE_INTEGER li;
		li.LowPart = os_ver_info_ptr->dwMinorVersion;
		li.HighPart = os_ver_info_ptr->dwMajorVersion;
		if (li.QuadPart >= 0x0000000500000001) {
			WORD processor_architecture = PROCESSOR_ARCHITECTURE_INTEL;
			typedef void (WINAPI* LPFN_GETNATIVESYSTEMINFO)(SYSTEM_INFO *lpSystemInfo);
			static LPFN_GETNATIVESYSTEMINFO fnGetNativeSystemInfo =
				reinterpret_cast<LPFN_GETNATIVESYSTEMINFO>(::GetProcAddress(
				::GetModuleHandle(L"kernel32.dll"), "GetNativeSystemInfo"));
			if (fnGetNativeSystemInfo != NULL) {
				SYSTEM_INFO sys_info;
				fnGetNativeSystemInfo(&sys_info);
				processor_architecture = sys_info.wProcessorArchitecture;
			}
			static SP_ALTPLATFORM_INFO_V2 *alt_platform_info_v2 = NULL;
			if (alt_platform_info_v2 == NULL) {
				alt_platform_info_v2 = new SP_ALTPLATFORM_INFO_V2();
				alt_platform_info_v2->cbSize = sizeof(SP_ALTPLATFORM_INFO_V2);
				alt_platform_info_v2->Platform = os_ver_info_ptr->dwPlatformId;
				alt_platform_info_v2->MajorVersion = os_ver_info_ptr->dwMajorVersion;
				alt_platform_info_v2->MinorVersion = os_ver_info_ptr->dwMinorVersion;
				alt_platform_info_v2->ProcessorArchitecture = processor_architecture;
				alt_platform_info_v2->Flags = 0x0001; // SP_ALTPLATFORM_FLAGS_VERSION_RANGE
				alt_platform_info_v2->FirstValidatedMajorVersion = os_ver_info_ptr->dwMajorVersion;
				alt_platform_info_v2->FirstValidatedMinorVersion = os_ver_info_ptr->dwMinorVersion;
			}
			typedef BOOL (WINAPI* LPFN_SETUPQUERYINFORIGINALFILEINFORMATION)(SP_INF_INFORMATION *InfInformation, unsigned int InfIndex, SP_ALTPLATFORM_INFO_V2 *AlternatePlatformInfo, SP_ORIGINAL_FILE_INFO_W *OriginalFileInfo);
			static LPFN_SETUPQUERYINFORIGINALFILEINFORMATION fnSetupQueryInfOriginalFileInformation =
				reinterpret_cast<LPFN_SETUPQUERYINFORIGINALFILEINFORMATION>(::GetProcAddress(
				::GetModuleHandle(L"SETUPAPI.DLL"), "SetupQueryInfOriginalFileInformationW"));
			BOOL ret = fnSetupQueryInfOriginalFileInformation(inf_info, 0, alt_platform_info_v2, &org_file_info);
			if (ret) {
				result = org_file_info.OriginalInfName;
			}
		} else {
			static SP_ALTPLATFORM_INFO_V1 *alt_platform_info_v1 = NULL;
			if (alt_platform_info_v1 == NULL) {
				alt_platform_info_v1 = new SP_ALTPLATFORM_INFO_V1();
				alt_platform_info_v1->cbSize = sizeof(SP_ALTPLATFORM_INFO_V1);
				alt_platform_info_v1->Platform = os_ver_info_ptr->dwPlatformId;
				alt_platform_info_v1->MajorVersion = os_ver_info_ptr->dwMajorVersion;
				alt_platform_info_v1->MinorVersion = os_ver_info_ptr->dwMinorVersion;
				alt_platform_info_v1->ProcessorArchitecture = PROCESSOR_ARCHITECTURE_INTEL;
				alt_platform_info_v1->Reserved = 0;
			}
			typedef BOOL (WINAPI* LPFN_SETUPQUERYINFORIGINALFILEINFORMATION)(SP_INF_INFORMATION *InfInformation, unsigned int InfIndex, SP_ALTPLATFORM_INFO_V1 *AlternatePlatformInfo, SP_ORIGINAL_FILE_INFO_W *OriginalFileInfo);
			static LPFN_SETUPQUERYINFORIGINALFILEINFORMATION fnSetupQueryInfOriginalFileInformation =
				reinterpret_cast<LPFN_SETUPQUERYINFORIGINALFILEINFORMATION>(::GetProcAddress(
				::GetModuleHandle(L"SETUPAPI.DLL"), "SetupQueryInfOriginalFileInformationW"));
			BOOL ret = fnSetupQueryInfOriginalFileInformation(inf_info, 0, alt_platform_info_v1, &org_file_info);
			if (ret) {
				result = org_file_info.OriginalInfName;
			}
		}


		//const size_t oem_inf_file_path_length = oem_inf_file_path.length();
		//if (oem_inf_file_path_length > 5) {
		//	// .pnft@C̃pX𐶐
		//	std::wstring pnf_file_path = oem_inf_file_path;
		//	pnf_file_path[oem_inf_file_path_length - 3] = 'P';
		//	// .pnft@C
		//	MyLib::AutoHandle<INVALID_HANDLE_VALUE> file_handle = ::CreateFile(
		//		pnf_file_path.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL,
		//		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		//	if (!file_handle.isInvalid()) {
		//		size_t file_size = ::GetFileSize(file_handle.get(), NULL);
		//		if (file_size > 0x54 + 4) {
		//			// .pnft@C̃ItZbg0x54ɁAINFt@CL^Ăꏊ̃ItZbgl
		//			::SetFilePointer(file_handle.get(), 0x54, NULL, FILE_BEGIN);
		//			unsigned int offset = 0;
		//			DWORD num_read;
		//			::ReadFile(file_handle.get(), &offset, 4, &num_read, NULL);
		//			if (offset != 0 && offset < file_size) {
		//				::SetFilePointer(file_handle.get(), offset, NULL, FILE_BEGIN);
		//				size_t const buf_size = file_size - offset;
		//				size_t const buf_count = buf_size/sizeof(wchar_t);
		//				boost::scoped_array<wchar_t> buf(new wchar_t[buf_count + 1]);
		//				::ReadFile(file_handle.get(), buf.get(), static_cast<DWORD>(buf_size), &num_read, NULL);
		//				if (::wcsnlen(buf.get(), buf_count) < buf_count) {
		//					result = buf.get();
		//				}
		//			}
		//		}
		//	}
		//}

		return result;
	}

	std::wstring getOriginalFileTitleFromOEMInf1(std::wstring const &oem_inf_file_path)
	{
		std::wstring result;

		// SP_INF_INFORMATION
		DWORD required_size;
		BOOL ret = ::SetupGetInfInformation(oem_inf_file_path.c_str(),
			INFINFO_INF_NAME_IS_ABSOLUTE, NULL, 0, &required_size);
		if (ret) {
			boost::scoped_array<char> buf1(new char[required_size]);
			SP_INF_INFORMATION *inf_info = reinterpret_cast<SP_INF_INFORMATION*>(buf1.get());
			ret = ::SetupGetInfInformation(oem_inf_file_path.c_str(),
				INFINFO_INF_NAME_IS_ABSOLUTE, inf_info, required_size, NULL);
			if (ret) {
				result = getOriginalFileTitleFromOEMInf0(inf_info);
			}
		}

		return result;
	}

	std::wstring getOEMInfFromOriginalFileTitle(const std::wstring& original_inf_file_title)
	{
		namespace fs = boost::filesystem;

		std::wstring result;
		TCHAR windir_path[MAX_PATH + 1];
		::GetWindowsDirectory(windir_path, MAX_PATH);
		fs::wpath const inf_dir_path = fs::wpath(windir_path)/L"inf";
		// ܂AVXeINFtH_ɁA̖Oő݂ĂȂ`FbN
		fs::wpath original_name_in_inf_dir = inf_dir_path/original_inf_file_title;
		if (fs::exists(original_name_in_inf_dir)) {
			result = original_inf_file_title;
		} else {
			fs::wpath wild_cards = inf_dir_path/L"oem*.inf";
			CFindFile finder;
			if (finder.FindFile(wild_cards.string().c_str())) {
				do {
					TCHAR oem_inf_file_path[MAX_PATH + 1];
					finder.GetFilePath(oem_inf_file_path, MAX_PATH);
					std::wstring orig_file_title = /*InfFileParser::*/getOriginalFileTitleFromOEMInf1(oem_inf_file_path);
					if (MyLib::compareStringNoCase(orig_file_title, original_inf_file_title) == 0) {
						TCHAR inf_file_title[_MAX_FNAME + 1];
						finder.GetFileName(inf_file_title, _MAX_FNAME);
						inf_file_title[::wcslen(inf_file_title) - 3] = L'I';
						result = inf_file_title;
						break;
					}
				} while (finder.FindNextFile());
			}
		}
		return result;
	}
};

// VXeINFtH_ɂ"oem0.inf"Ȃǂ̌̃t@C߂
std::wstring InfFileParser::getOriginalFileTitleFromOEMInf()
{
	std::wstring result;

	// SP_INF_INFORMATION
	DWORD required_size;
	BOOL ret = ::SetupGetInfInformation(this->inf_file,
		INFINFO_INF_SPEC_IS_HINF, NULL, 0, &required_size);
	if (ret) {
		boost::scoped_array<char> buf1(new char[required_size]);
		SP_INF_INFORMATION *inf_info = reinterpret_cast<SP_INF_INFORMATION*>(buf1.get());
		ret = ::SetupGetInfInformation(this->inf_file,
			INFINFO_INF_SPEC_IS_HINF, inf_info, required_size, NULL);
		if (ret) {
			result = getOriginalFileTitleFromOEMInf0(inf_info);
		}
	}

	return result;
}

// CopyInf directivẻ
DriverSetupData::CopyInfs
InfFileParser::getCopyInfs(const std::wstring& dd_install_section, bool is_installed_file)
{
	DriverSetupData::CopyInfs result;

	const InfContextIterator inf_end;
	InfContextIterator inf_it =
		this->inf_file.findFirstLine(dd_install_section.c_str(), _T("CopyInf"));
	for (; inf_it != inf_end; inf_it = this->inf_file.findNextMatchLine(inf_it, _T("CopyInf"))) {
		for (uint32_t i = 1; i <= inf_it->getFieldCount(); ++i) {
			std::wstring const original_inf_file_title = inf_it->getStringField(i);
			if (!original_inf_file_title.empty()) {
				if (is_installed_file) {
					std::wstring const oem_inf_file_title =
						getOEMInfFromOriginalFileTitle(original_inf_file_title);
					if (!oem_inf_file_title.empty()) {
						// Inft@CXgɒǉ
						result.insert(DriverSetupData::CopyInfs::value_type(original_inf_file_title, oem_inf_file_title));
					} else if (this->version >= 0x00050001) {
						throw MY_WIN32_FILE_EXCEPTION(original_inf_file_title, ERROR_FILE_NOT_FOUND);
					}
				} else {
					// Inft@CXgɒǉ
					result.insert(DriverSetupData::CopyInfs::value_type(original_inf_file_title, original_inf_file_title));
				}
			}
		}
	}
	return result;
}


// INFt@CAhCõt@CRs[f[^擾B
DriverSetupData::FileData*
InfFileParser::getFileData(const std::wstring& filename)
{
	namespace fs = boost::filesystem;
	typedef DriverSetupData::TagFileCollection TagFileCollection;

	// SourceDisksFiles Section T
	const InfContextIterator inf_end;
	InfContextIterator sdf_it = this->findDecoratedSection(_T("SourceDisksFiles"), filename.c_str());
	if (sdf_it != inf_end) {
		uint32_t field_count = sdf_it->getFieldCount();
		const std::wstring source_id = sdf_it->getStringField(1);
		std::wstring sdf_subdir;
		if (field_count >= 2) {
			sdf_subdir = sdf_it->getStringField(2);
		}
		// SourceDisksNames Section T
		InfContextIterator sdn_it = this->findDecoratedSection(_T("SourceDisksNames"), source_id.c_str());
		if (sdn_it != inf_end) {
			std::wstring tag_or_cab_file;
			if (sdn_it->getFieldCount() >= 2) {
				tag_or_cab_file = sdn_it->getStringField(2);
			}
			std::wstring sdn_subdir;
			if (sdn_it->getFieldCount() >= 4) {
				sdn_subdir = sdn_it->getStringField(4);
			}
			// installation_root (INFt@C̃fBNg)擾
			fs::wpath installation_root(this->inf_file.getFullPathName());
			installation_root.remove_filename();
			if (!tag_or_cab_file.empty()) {
				// tag-or-cab-file
				fs::wpath tag_or_cab_file_path;
				TagFileCollection::iterator tag_file_it = this->tag_files.find(tag_or_cab_file);
				if (tag_file_it == this->tag_files.end()) {
					tag_or_cab_file_path = installation_root/tag_or_cab_file;
					if (fs::exists(tag_or_cab_file_path)) {
						this->tag_files.insert(
							TagFileCollection::value_type(tag_or_cab_file, tag_or_cab_file));
					} else {
						fs::wpath const tag_or_cab_file_rel_path = fs::wpath(sdn_subdir)/tag_or_cab_file;
						tag_or_cab_file_path = installation_root/tag_or_cab_file_rel_path;
						if (fs::exists(tag_or_cab_file_path)) {
							this->tag_files.insert(
								TagFileCollection::value_type(tag_or_cab_file, tag_or_cab_file_rel_path.string()));//tag_or_cab_file_path.string().substr(installation_root.string().length() + 1)));
						} else {
							throw MY_WIN32_FILE_EXCEPTION(filename, ERROR_FILE_NOT_FOUND);
						}
					}
				} else {
					tag_or_cab_file_path = installation_root/(tag_file_it->second);
				}
				// tag-file (tB[h6)
				if (this->version >= 0x00050001 && sdn_it->getFieldCount() >= 6 && (sdn_it->getIntField(5) & 0x10)) {
					const std::wstring tag_file = sdn_it->getStringField(6);
					if (!tag_file.empty() && this->tag_files.find(tag_file) == this->tag_files.end()) {
						fs::wpath const path1 = installation_root/tag_file;
						if (fs::exists(path1)) {
							this->tag_files.insert(
								TagFileCollection::value_type(tag_file, tag_file));
						} else {
							fs::wpath const rel_path2 = fs::wpath(sdn_subdir)/tag_file;
							fs::wpath const path2 = installation_root/rel_path2;
							if (fs::exists(path2)) {
								this->tag_files.insert(
									TagFileCollection::value_type(tag_file, rel_path2.string()));//path.string().substr(installation_root.string().length() + 1)));
							} else {
								throw MY_WIN32_FILE_EXCEPTION(filename, ERROR_FILE_NOT_FOUND);
							}
						}
					}
					// mCABt@C
					//result.installation_path = tag_or_cab_file_path.str().substr(installation_root.str().length() + 1);
					return NULL;
				}
			}
			// t@CP̂ő݂Ăꍇ̃pX쐬
			fs::wpath const file_rel_path = fs::wpath(sdn_subdir)/sdf_subdir/filename;
			fs::wpath const file_path = installation_root/file_rel_path;
			MyLib::FileCompressionInfo info(file_path);
			if (info.exists()) {
				DriverSetupData::FileData *ptr = new DriverSetupData::FileData();
				ptr->source_path = info.getActualPath();//file_path;
				ptr->installation_path = file_rel_path.string();//file_path.string().substr(installation_root.string().length() + 1);
				return ptr;
			} else if (!tag_or_cab_file.empty()) {
				// t@CȂꍇACabt@CɓĂ\
				// ܂Ainstallation_rootcabt@CT
				fs::wpath cab_path =
					installation_root/tag_or_cab_file;
				if (!fs::exists(cab_path)) {
					// installation_rootɂȂꍇASourceDisksNamesZNVɏĂTufBNgT
					cab_path = installation_root/sdn_subdir/tag_or_cab_file;
				}
				if (fs::exists(cab_path)) {
					// tagfileƂCABt@CRs[΂悢
					//result.installation_path = cab_path.str().substr(installation_root.str().length() + 1);
					return NULL;
				}
			}
			throw MY_WIN32_FILE_EXCEPTION(filename, ERROR_FILE_NOT_FOUND);
		}
	}
	return NULL;
}

DriverSetupData::FileData*
InfFileParser::getInstalledFileData(std::wstring const &inst_file_title, boost::filesystem::wpath const &src_file_path)
{
	namespace fs = boost::filesystem;
	typedef DriverSetupData::FileInCabInfo FileInCabInfo;
	typedef DriverSetupData::NewCabFileCollection NewCabFileCollection;
	typedef DriverSetupData::TagFileCollection TagFileCollection;

	const InfContextIterator inf_end;
	// SourceDisksFiles
	InfContextIterator sdf_it = this->findDecoratedSection(_T("SourceDisksFiles"), inst_file_title.c_str());
	if (sdf_it != inf_end) {
		unsigned int field_count = sdf_it->getFieldCount();
		const std::wstring source_id = sdf_it->getStringField(1);
		std::wstring sdf_subdir;
		if (field_count >= 2) {
			sdf_subdir = sdf_it->getStringField(2);
		}
		// SourceDisksNames Section T
		InfContextIterator sdn_it = this->findDecoratedSection(_T("SourceDisksNames"), source_id.c_str());
		if (sdn_it != inf_end) {
			std::wstring tag_or_cab_file;
			if (sdn_it->getFieldCount() >= 2) {
				tag_or_cab_file = sdn_it->getStringField(2);
			}
			std::wstring sdn_subdir;
			if (sdn_it->getFieldCount() >= 4) {
				sdn_subdir = sdn_it->getStringField(4);
			}
			// installation_root (INFt@C̃fBNg)擾
			fs::wpath installation_root(this->inf_file.getFullPathName());
			installation_root.remove_filename();
			if (!tag_or_cab_file.empty()) {
				// tag-or-cab-file
				fs::wpath const rel_path = fs::wpath(sdn_subdir)/tag_or_cab_file;
				fs::wpath const tag_or_cab_file_path = installation_root/rel_path;
				//std::wstring const relative_path = tag_or_cab_file_path.string().substr(installation_root.string().length() + 1);
				//size_t const tag_or_cab_file_length = tag_or_cab_file.length();
				//if (tag_or_cab_file_length > 4 && MyLib::compareStringNoCase(_T(".CAB"), tag_or_cab_file.substr(tag_or_cab_file_length - 4)) == 0) {
				if (MyLib::compareStringNoCase(_T(".CAB"), rel_path.extension()) == 0) {
					NewCabFileCollection::iterator it = this->new_cab_files.find(rel_path.string());
					// new_cab_file t@Cǉ
					if (it == this->new_cab_files.end()) {
						it = this->new_cab_files.insert(std::make_pair(rel_path.string(), FileInCabInfo())).first;
						it->second.insert(std::make_pair(inst_file_title, src_file_path));
					} else {
						FileInCabInfo::iterator it2 = it->second.find(inst_file_title);
						if (it2 == it->second.end()) {
							it->second.insert(std::make_pair(inst_file_title, src_file_path));
						}
					}
					return NULL;
				}
				TagFileCollection::iterator tag_file_it = this->tag_files.find(tag_or_cab_file);
				if (tag_file_it == this->tag_files.end()) {
					//tag_or_cab_file_path;
					this->tag_files.insert(
						TagFileCollection::value_type(tag_or_cab_file, rel_path.string()));//tag_or_cab_file_path.string().substr(installation_root.string().length() + 1)));
				}
				// tag-file (tB[h6)
				if (this->version >= 0x00050001 && sdn_it->getFieldCount() >= 6 && (sdn_it->getIntField(5) & 0x10)) {
					const std::wstring tag_file = sdn_it->getStringField(6);
					if (!tag_file.empty() && this->tag_files.find(tag_file) == this->tag_files.end()) {
						fs::wpath const tag_rel_path = fs::wpath(sdn_subdir)/tag_file;
						//fs::wpath path = installation_root/tag_file;
						//fs::wpath const path = installation_root/rel_path;
						this->tag_files.insert(
							TagFileCollection::value_type(tag_file, tag_rel_path.string()));//path.string().substr(installation_root.string().length() + 1)));
					}
					// mCABt@C
					NewCabFileCollection::iterator it = this->new_cab_files.find(rel_path.string());
					// new_cab_file t@Cǉ
					if (it == this->new_cab_files.end()) {
						it = this->new_cab_files.insert(std::make_pair(rel_path.string(), FileInCabInfo())).first;
						it->second.insert(std::make_pair(inst_file_title, src_file_path));
					} else {
						FileInCabInfo::iterator it2 = it->second.find(inst_file_title);
						if (it2 == it->second.end()) {
							it->second.insert(std::make_pair(inst_file_title, src_file_path));
						}
					}
					return NULL;
				}
			}
			// t@CP̂ő݂Ăꍇ̃pX쐬
			fs::wpath file_rel_path = fs::wpath(sdn_subdir)/sdf_subdir/inst_file_title;
			fs::wpath file_path = installation_root/file_rel_path;
			DriverSetupData::FileData *ptr = new DriverSetupData::FileData();
			ptr->source_path = src_file_path;
			ptr->installation_path = file_rel_path.string();//file_path.string().substr(installation_root.string().length() + 1);
			return ptr;
		}
	}
	return NULL;
}

DriverSetupData::TagFileCollection InfFileParser::getTagFiles() const
{
	return this->tag_files;
}

DriverSetupData::NewCabFileCollection InfFileParser::getNewCabFiles() const
{
	return this->new_cab_files;
}
