#pragma once

#include "../../MyLib/Exception.h"
#include "../../MyLib/MyLib.h"
#include "DriverSetupData.h"

class IntegrationSetting
{
public:
	IntegrationSetting()
		: adapts_big_drive(false)
	{
	}

	MyLib::PathString getWorkDirPath() const
	{
		return this->work_dir_path;
	}

	MyLib::PathString getSPFilePath() const
	{
		return this->sp_file_path;
	}

	bool adaptsBigDrive() const
	{
		return this->adapts_big_drive;
	}

	bool isCleanInstallOnly() const
	{
		return this->is_clean_install_only;
	}

	bool recompressesDriverCab() const
	{
		return this->recompresses_driver_cab;
	}

	const std::vector<DriverSetupData>& getDriverSetupDataCollection() const
	{
		return this->driver_data;
	}

	MyLib::String getVolumeLabel() const
	{
		return this->volume_label;
	}

	MyLib::PathString getISOImagePath() const
	{
		return this->iso_image_path;
	}

	MyLib::String getRecordDeviceString() const
	{
		return this->record_device_str;
	}

	int getRecordSpeed() const
	{
		return this->record_speed;
	}

	bool erasesRWMedia() const
	{
		return this->erases_rw_media;
	}

	bool ejectsMedia() const
	{
		return this->ejects_media;
	}

	enum SetWorkDirReturnCode {
		SET_WORK_DIR_SUCCESS,
		SET_WORK_DIR_NOT_EMPTY,
		SET_WORK_DIR_NOT_DIRECTORY,
		SET_WORK_DIR_CREATION_FAILED
	};
	// 
	enum SetWorkDirReturnCode setWorkDirPath(const MyLib::PathString& path, bool fails_if_not_empty)
	{
		enum SetWorkDirReturnCode result = SET_WORK_DIR_SUCCESS;
		//CPath dir_full_path = getFullPath(dir_path);
		if (path.exists()) {
			// ݂Ă΁AfBNgA󂩃`FbN
			if (path.isDirectory()) {
				// 󂩂ǂ
				if (!isDirectoryEmpty(path) && fails_if_not_empty) {
					result = SET_WORK_DIR_NOT_EMPTY;
				} else {
					this->work_dir_path = path;
				}
			} else {
				// fBNgȂ
				result = SET_WORK_DIR_NOT_DIRECTORY;
			}
		} else {
			// ݂Ȃ΁AV
			int err_code =
				MyLib::createDirectoryPath(path.str().c_str());
			if (err_code == NO_ERROR) {
				this->work_dir_path = path;
			} else {
				result = SET_WORK_DIR_CREATION_FAILED;
			}
		}
		return result;
	}

	bool setSPFilePath(const MyLib::PathString& path)
	{
		if (path.exists() && !path.isDirectory()) {
			this->sp_file_path = path;
			return true;
		} else {
			return false;
		}
	}

	void setBigDrive(bool adapts)
	{
		this->adapts_big_drive = adapts;
	}

	void setCleanInstallOnly(bool is_clean_install_only)
	{
		this->is_clean_install_only = is_clean_install_only;
	}

	void setDriverCabCompression(bool recompresses)
	{
		this->recompresses_driver_cab = recompresses;
	}

	void setDriverSetupData(const std::vector<DriverSetupData>& data)
	{
		this->driver_data = data;
	}

	void setVolumeLabel(const MyLib::String& volume_label)
	{
		this->volume_label = volume_label;
	}

	void setISOImagePath(const MyLib::PathString& path)
	{
		this->iso_image_path = path;
	}

	void setRecordDeviceString(const MyLib::String& str)
	{
		this->record_device_str = str;
	}

	void setRecordSpeed(int n)
	{
		this->record_speed = n;
	}

	void setErasesRWMedia(bool erases)
	{
		this->erases_rw_media = erases;
	}

	void setEjectsMedia(bool ejects)
	{
		this->ejects_media = ejects;
	}

private:
	// 
	MyLib::PathString work_dir_path;
	MyLib::PathString sp_file_path;
	bool adapts_big_drive;
	bool is_clean_install_only;
	bool recompresses_driver_cab;
	std::vector<DriverSetupData> driver_data;
	MyLib::String volume_label;
	MyLib::PathString iso_image_path;
	MyLib::String record_device_str;
	int record_speed;
	bool erases_rw_media;
	bool ejects_media;

	// isDirectoryEmpty
	//  fBNg󂩒ׂ
	static bool isDirectoryEmpty(const MyLib::PathString& dir_path)
	{
		bool result = true;
		const MyLib::PathString wild_card = dir_path + _T("*.*");
		CFindFile finder;
		if (finder.FindFile(wild_card.str().c_str())) {
			do {
				if (!finder.IsDots()) {
					result = false;
					break;
				}
			} while (finder.FindNextFile());
		}
		return result;
	}
};
