#include "mylib/exception.h"
#include "mylib/function.h"
#include "mylib/CabCompressor.h"
#include "mylib/CabDecompressor.h"
#include "mylib/FileCompressionInfo.h"
#include "mylib/SetupFileModifier.h"
#include "mylib/inifile.h"
#include "DriverSetupData.h"
#include "DriverFileData.h"
#include "DriverServiceData.h"
#include "Integrator.h"
#include "TemporaryDirectory.h"
#include "FileCopyManager.h"
#include "mylib/FileVersionInfo.h"

extern void deleteDirectory(boost::filesystem::wpath const &dir_path);
extern void removeFileForcibly(boost::filesystem::wpath const &file_path);

MYFREEFUNC MyFree = (MYFREEFUNC)GetProcAddress(GetModuleHandle(L"SETUPAPI.DLL"), "MyFree");

namespace fs {
	using namespace boost::filesystem;
};

/////////////////////////////////////////////////////////////////////////////


class MyCabDecompressor :
	public MyLib::CabDecompressorT<MyCabDecompressor>
{
public:
	CopyFileFlags onCopyFile(
		const std::wstring& file_name_in_cab,
		size_t /*uncompressed_file_size*/,
		FILETIME file_time,
		unsigned short /*attribs*/)
	{
		fs::wpath dest_path = this->target_dir_path/file_name_in_cab;
		if (fs::exists(dest_path)) {
			HANDLE file_handle = ::CreateFile(dest_path.string().c_str(), GENERIC_READ,
				FILE_SHARE_READ, NULL, OPEN_EXISTING,
				FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
			if (file_handle != INVALID_HANDLE_VALUE) {
				FILETIME last_write_time;
				::GetFileTime(file_handle, NULL, NULL, &last_write_time);
				::CloseHandle(file_handle);
				FILETIME existing_file_time;
				::FileTimeToLocalFileTime(&last_write_time, &existing_file_time);
				if (::CompareFileTime(&file_time, &existing_file_time) > 0) {
					MyLib::toFileWritable(dest_path);
					if (!::DeleteFile(dest_path.string().c_str())) {
						::MessageBox(NULL, dest_path.string().c_str(), _T("DeleteFile"), 0);
						return ABORT;
					}
					return EXTRACT;
				} else {
					return SKIP;
				}
			} else {
				return ABORT;
			}
		} else {
			return EXTRACT;
		}
	}
};

//uint64_t getFileVersion(std::wstring const &path)
//{
//	DWORD info_size = ::GetFileVersionInfoSize(path.c_str(), NULL);
//	MyLib::TemporaryBuffer<char> buf(info_size);
//	if (::GetFileVersionInfo(path.c_str(), NULL, info_size, buf.getPointer())) {
//		VS_FIXEDFILEINFO *fixed_file_info;
//		unsigned int len;
//		::VerQueryValue(buf.getPointer(), _T("\\"), reinterpret_cast<void**>(&fixed_file_info), &len);
//		return (static_cast<uint64_t>(fixed_file_info->dwFileVersionMS) << 32) | fixed_file_info->dwFileVersionLS;
//	}
//	return 0;
//}

//FILETIME getFileTime(std::wstring const &path)
//{
//	FILETIME result = {0};
//	HANDLE file_handle = ::CreateFile(path.c_str(), GENERIC_READ,
//		FILE_SHARE_READ, NULL, OPEN_EXISTING,
//		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
//	if (file_handle != INVALID_HANDLE_VALUE) {
//		FILETIME last_write_time;
//		::GetFileTime(file_handle, NULL, NULL, &last_write_time);
//		::CloseHandle(file_handle);
//		::FileTimeToLocalFileTime(&last_write_time, &result);
//	}
//	return result;
//}

namespace {
bool isServiceBinary(DriverSetupData::ServiceDataCollection const &service_data_collection,
                     std::wstring const &file_title)
{
	DriverSetupData::ServiceDataCollection::const_iterator it =
		service_data_collection.begin();
	for (; it != service_data_collection.end(); ++it) {
		if (MyLib::compareStringNoCase(file_title, it->second.service_binary) == 0) {
			return true;
		}
	}
	return false;
}

boost::filesystem::wpath changeExtensionForCompressedFile(boost::filesystem::wpath const &file_path, wchar_t tail_char)
{
	std::wstring const path_str = file_path.string();
	std::wstring const ext = fs::extension(file_path);

	if (ext.empty()) {
		// gqȂ
		wchar_t new_ext[] = {L'.', tail_char, L'\0'};
		return fs::change_extension(file_path, new_ext);
	} else {
		size_t const ext_len = ext.length();
		if (ext_len <= 3) {
			// gqhbg3ȉ
			return fs::change_extension(file_path, ext + tail_char);
		} else {
			std::wstring new_ext = ext;
			new_ext[ext_len - 1] = tail_char;
			return fs::change_extension(file_path, new_ext);
		}
	}
}

// copyFileWithCompress
//  source_file_path and dest_file_path have no-compressed filename.
void copyFileWithCompress(boost::filesystem::wpath const &source_file_path,
                          boost::filesystem::wpath const &dest_file_path,
                          std::wstring const &file_title)
{
	fs::wpath const dest_path =
		changeExtensionForCompressedFile(dest_file_path, L'_');

	MyLib::FileCompressionInfo info(source_file_path);
	if (info.getCompressionType() != FILE_COMPRESSION_NONE) {
		// t@CkĂ΂̂܂܃Rs[
		std::wstring const src_path_str = info.getActualPath().string();
		std::wstring const dest_path_str = dest_path.string();
		MyLib::toFileWritable(dest_path);
		if (!::CopyFile(src_path_str.c_str(), dest_path_str.c_str(), FALSE)) {
			throw MY_WIN32_FILE_TRANSFER_EXCEPTION_LAST_ERROR(src_path_str, dest_path_str);
		}
	} else {
		// t@CkĂȂΈk
		MyLib::CabCompressor compress(dest_path);
		compress.addFile(source_file_path, file_title);
		compress.flush(false);
	}
}
};

/////////////////////////////////////////////////////////////////////////////

Integrator::Integrator(MyLib::SetupFileModifier *txtsetup_sif0,
                       MyLib::SetupFileModifier *dosnet_inf0,
                       MyLib::RegHive *setupreg_hiv0,
                       MyLib::SetupFileModifier *hivesys_inf0,
                       boost::filesystem::wpath const &work_dir_path0,
                       std::wstring const &setup_subdir_name0,
                       int architecture_id,
                       uint32_t os_ver0)
	: txtsetup_sif(txtsetup_sif0)
	, dosnet_inf(dosnet_inf0)
	, setupreg_hiv(setupreg_hiv0)
	, hivesys_inf(hivesys_inf0)
	, work_dir_path(work_dir_path0)
	, setup_subdir_name(setup_subdir_name0)
	, target_path(work_dir_path/setup_subdir_name)
	, architecture_id(architecture_id)
	, os_ver(os_ver0)
{
	using MyLib::SetupFileModifier;
	//using MyLib::SvcpackInfModifier;
	using MyLib::RegHive;

	//fs::wpath target_file_path;
	//{
	//	target_file_path = this->target_path/L"txtsetup.sif";
	//	this->txtsetup_sif.reset(new SetupFileModifier(target_file_path.string()));
	//	target_file_path = this->target_path/L"dosnet.inf";
	//	this->dosnet_inf.reset(new SetupFileModifier(target_file_path.string()));
	//	target_file_path = this->target_path/L"setupreg.hiv";
	//	this->setupreg_hiv.reset(new RegHive(target_file_path.string()));
	//}

	switch (architecture_id) {
	case PROCESSOR_ARCHITECTURE_INTEL:
		this->section_decorate = _T(".x86");
		break;
	case PROCESSOR_ARCHITECTURE_AMD64:
		this->section_decorate = _T(".amd64");
		break;
	case PROCESSOR_ARCHITECTURE_IA64:
		this->section_decorate = _T(".ia64");
		break;
	default:
		assert(false);
	}
}

Integrator::~Integrator()
{
}

//// ύXt@Cɔf
//void Integrator::flush()
//{
//	this->fireEvent(_T("TXTSETUP.SIF CĂ܂BBB"));
//	this->txtsetup_sif->flush();
//	this->fireEvent(_T("\n"));
//
//	this->fireEvent(_T("DOSNET.INF CĂ܂BBB"));
//	this->dosnet_inf->flush();
//	this->fireEvent(_T("\n"));
//}

namespace {
	fs::wpath *drivers_dir_path_ptr = NULL;
};

void Integrator::registFilesForSetupFiles(std::wstring const &filename/*, bool not_compressed*/)
{
	// TXTSETUP.SIF ̏
	MyLib::SetupFileLine::FieldCollection fields;
	fields[0] = filename;
	fields[1] = _T('1');
	//if (not_compressed) {
	//	fields[7] = L"_x";
	//}
	fields[8] = _T('2');
	fields[9] = _T('0');
	fields[10] = _T('0');
	// XPȍ~ŐVǉꂽtB[h
	if (this->os_ver >= 0x00050001) {
		fields[12] = _T('1');
		fields[13] = _T('2');
	}
	this->txtsetup_sif->insertLine(_T("SourceDisksFiles") + this->section_decorate, fields);
	// DOSNET.INF̏
	fields.clear();
	fields[1] = _T("d1");
	fields[2] = filename;
	this->dosnet_inf->insertLine(L"Files", fields);
}

void Integrator::prepareDriverIntegration()
{
	// ǉhCoǗptH_̃pX
	if (!drivers_dir_path_ptr) {
		drivers_dir_path_ptr = new fs::wpath(this->work_dir_path/L"$OEM$/Drivers");  // TODO: INIt@Cǂݍ
	}
	fs::wpath const drivers_dir_path = *drivers_dir_path_ptr;
	if (!fs::exists(drivers_dir_path)) {
		fs::create_directories(drivers_dir_path);
	}

	// F6TOOLSB.EXE
	// F6TOOLSB.DLL
	// F6TOOLSB.CMD
	//  - F6TOOLSB.EXE  WINNT.SIF [GuiUnattended] DetachedProgram Arguments ̓e
	//  SETUP.EXE Ŵ̂ł΁ASETUPORG.EXE ɕύX
	fs::wpath const setup_exe_path = this->target_path/L"SETUP.EXE";
	MyLib::FileVersionInfo setup_exe_verinfo(setup_exe_path);
	if (setup_exe_verinfo.queryValue(L"CompanyName") == L"Microsoft Corporation") {
		// SETUP.EX_ WJ
		MyLib::FileCompressionInfo setup_exe_info(setup_exe_path.string().c_str());
		TemporaryDirectoryCloseDelete temp_dir_close_del;
		TemporaryDirectory temp_dir(&temp_dir_close_del);
		fs::wpath const setup_exe_temp_dir = temp_dir.getDirectoryPath();
		fs::wpath extract_path = setup_exe_info.decompressOrCopyFile(setup_exe_temp_dir);
		// SETUPORG.EX_ 쐬
		fs::wpath const dest_path = this->target_path/L"SETUPORG.EX_";
		try {
			MyLib::CabCompressor compress(dest_path.string().c_str());
			compress.addFile(extract_path, std::wstring(L"SETUPORG.EXE"));
			compress.flush(false);
		} catch (const std::runtime_error& e) {
			throw std::runtime_error(std::string("SETUPORG.EX_\r\n") + e.what());
		}
		//// TXTSETUP.SIF ̏
		//MyLib::SetupFileLine::FieldCollection fields;
		//fields[0] = _T("setuporg.exe");
		//fields[1] = _T('1');
		//fields[8] = _T('2');
		//fields[9] = _T('0');
		//fields[10] = _T('0');
		//// XPȍ~ŐVǉꂽtB[h
		//if (this->os_ver >= 0x00050001) {
		//	fields[12] = _T('1');
		//	fields[13] = _T('2');
		//}
		//this->txtsetup_sif->insertLine(_T("SourceDisksFiles") + this->section_decorate, fields);
		//// DOSNET.INF̏
		//fields.clear();
		//fields[1] = _T("d1");
		//fields[2] = _T("setuporg.exe");
		//this->dosnet_inf->insertLine(_T("Files"), fields);
		registFilesForSetupFiles(L"setuporg.exe");
	}

	// setup.ex_ ̃Rs[
	// (setup.x86 t@CύXăRs[)
	FileCopyManager bin_to_target(*BIN_DIR_PATH, this->target_path);
	bin_to_target.copyFile(L"setup" + this->section_decorate, L"SETUP.EX_");

	//FileCopyManager bin_to_drivers(*BIN_DIR_PATH, drivers_dir_path);
	//// PRESETUP.CMD
	//if (!fs::exists(this->target_path/L"PRESETUP.CMD")) {
	//	registFilesForSetupFiles(L"presetup.cmd", true);
	//}
	//bin_to_target.copyFile(L"PRESETUP.CMD");
	//// SetDevicePath.exe
	//bin_to_drivers.copyFile(L"SetDevicePath" + this->section_decorate, L"SETDEVP.EXE");
	//// nosigdrv.exe
	//if (!fs::exists(this->target_path/L"NOSIGDRV.EX_")) {
	//	registFilesForSetupFiles(L"nosigdrv.exe");
	//}
	//bin_to_target.copyFile(L"nosigdrv.exe" + this->section_decorate, L"NOSIGDRV.EX_");

	// nosigdrv.dll
	if (!fs::exists(this->target_path/L"NOSIGDRV.DL_")) {
		registFilesForSetupFiles(L"nosigdrv.dll");
	}
	bin_to_target.copyFile(L"nosigdrv" + this->section_decorate, L"NOSIGDRV.DL_");

	fs::wpath const svcpack_dir_path = this->work_dir_path/this->setup_subdir_name/L"SVCPACK";
	// svcpack fBNg쐬
	fs::create_directory(svcpack_dir_path);

	FileCopyManager bin_to_svcpack(*BIN_DIR_PATH, svcpack_dir_path);
	// ResetDevicePath.exe
	bin_to_svcpack.copyFile(L"ResetDevicePath" + this->section_decorate, L"RSTDEVP.EXE");

	// DOSNET.INF [OptionalSrcDirs] svcpack
	if (!this->dosnet_inf->existsMatchLine(_T("OptionalSrcDirs"), 1, _T("svcpack"))) {
		this->dosnet_inf->insertLine(_T("OptionalSrcDirs"), _T("svcpack"));
	}
	// SVCPACK.INF [SetupHotfixesToRun] rstdevp.exe
	MyLib::SvcpackInfModifier svcpack_inf((this->target_path/L"SVCPACK.INF").string(),
	                                      this->os_ver,
	                                      L'\\' + this->setup_subdir_name + L"\\SVCPACK");
	if (!svcpack_inf.existsMatchLine(_T("SetupHotfixesToRun"), 1, _T("rstdevp.exe"))) {
		svcpack_inf.insertLine(_T("SetupHotfixesToRun"), _T("rstdevp.exe"));
	}
	svcpack_inf.flush();
}

void Integrator::integrateDriver(DriverSetupData const &data)
{
	assert(drivers_dir_path_ptr);
	fs::wpath const drivers_dir_path = *drivers_dir_path_ptr;

	// CINFt@C̃pX擾
	fs::wpath const inf_src_path = data.getInfFilePath();
	// Installation Root
	fs::wpath installation_root = inf_src_path;
	installation_root.remove_filename();

	bool is_new_dir_created = false;
	fs::wpath dest_dir_path;
	std::wstring const installation_root_str = installation_root.string();
	DriverDestinations::const_iterator dests_it = this->driver_dests.find(installation_root_str);
	if (dests_it != this->driver_dests.end()) {
		dest_dir_path = dests_it->second;
	} else {
		// ehCo16i̔ԍĊǗ
		for (int i = 0; i < 1000; ++i) {
			TCHAR buf[4];
			::wsprintf(buf, _T("%X"), i);
			fs::wpath const path = drivers_dir_path/buf;
			if (!fs::exists(path) || fs::is_empty(path)) {
				dest_dir_path = path;
				break;
			}
		}
		// hCõTutH_쐬
		fs::create_directories(dest_dir_path);
		is_new_dir_created = true;
	}

	// CINFt@C̏
	fs::wpath const inf_dest_path = dest_dir_path/data.getOriginalInfFileTitle();
	// INFt@C͂̂܂܃Rs[
	{
		std::wstring const src_path_str = inf_src_path.string();
		std::wstring const dest_path_str = inf_dest_path.string();
		if (!::CopyFile(src_path_str.c_str(), dest_path_str.c_str(), FALSE)) {
			throw MY_WIN32_FILE_TRANSFER_EXCEPTION_LAST_ERROR(src_path_str, dest_path_str);
		}
	}
	// CINFt@CCopyInfsŎw肳Ăt@C̏
	DriverSetupData::CopyInfs::const_iterator it3 = data.getCopyInfs().begin();
	for (; it3 != data.getCopyInfs().end(); ++it3) {
		// INFt@C͂̂܂܃Rs[
		fs::wpath const src_path = installation_root/it3->second;
		fs::wpath const dest_path = dest_dir_path/it3->first;
		// t@Cłɑ݂Ă΃XLbv
		if (!fs::exists(dest_path)) {
			std::wstring const src_path_str = src_path.string();
			std::wstring const dest_path_str = dest_path.string();
			if (!::CopyFile(src_path_str.c_str(), dest_path_str.c_str(), FALSE)) {
				throw MY_WIN32_FILE_TRANSFER_EXCEPTION_LAST_ERROR(src_path_str, dest_path_str);
			}
		}
	}

	// NewCabFile 쐬
	DriverSetupData::NewCabFileCollection const new_cab_files = data.getNewCabFiles();
	DriverSetupData::NewCabFileCollection::const_iterator it36 = new_cab_files.begin();
	for (; it36 != new_cab_files.end(); ++it36) {
		fs::wpath const new_cab_file_path = dest_dir_path/it36->first;
		// t@Cłɑ݂Ă΃XLbv
		if (!fs::exists(new_cab_file_path)) {
			// fBNg쐬
			fs::wpath dest_subdir_path = new_cab_file_path;
			dest_subdir_path.remove_filename();
			fs::create_directories(dest_subdir_path);
			std::wstring const file_title = new_cab_file_path.filename();
			MyLib::CabCompressor compress(new_cab_file_path.string().c_str());
			DriverSetupData::FileInCabInfo::const_iterator it37 = it36->second.begin();
			for (; it37 != it36->second.end(); ++it37) {
				compress.addFile(it37->second, it37->first);
			}
			compress.flush(false);
		}
	}

	// TagFilẽRs[
	DriverSetupData::TagFileCollection const tag_files = data.getTagFiles();
	DriverSetupData::TagFileCollection::const_iterator it35 = tag_files.begin();
	for (; it35 != tag_files.end(); ++it35) {
		fs::wpath const src_path = installation_root/it35->first;
		fs::wpath const dest_path = dest_dir_path/it35->second;
		// t@Cłɑ݂Ă΃XLbv
		if (!fs::exists(dest_path)) {
			// fBNg쐬
			fs::wpath dest_subdir_path = dest_path;
			dest_subdir_path.remove_filename();
			fs::create_directories(dest_subdir_path);

			if (fs::exists(src_path)) {
				std::wstring const src_path_str = src_path.string();
				std::wstring const dest_path_str = dest_path.string();
				if (!::CopyFile(src_path_str.c_str(), dest_path_str.c_str(), FALSE)) {
					throw MY_WIN32_FILE_TRANSFER_EXCEPTION_LAST_ERROR(src_path_str, dest_path_str);
				}
			} else {
				// ̃t@C݂ȂꍇAƂ肠̃t@C쐬
				::CloseHandle(::CreateFile(dest_path.string().c_str(), GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL));
			}
		}
	}
	// CINFt@CCopyFilesŎw肳Ăt@C̏
	DriverSetupData::FileDataCollection::const_iterator it =
		data.getFileDataCollection().begin();
	for (; it != data.getFileDataCollection().end(); ++it) {
		DriverSetupData::CopyFilesTagData::const_iterator it2 =
			it->second.begin();
		for (; it2 != it->second.end(); ++it2) {
			std::wstring const file_title = it2->first;
			fs::wpath const source_file_path = it2->second.source_path;
			fs::wpath const dest_file_path = dest_dir_path/it2->second.installation_path;
			// t@Cłɑ݂Ă΃XLbv
			MyLib::FileCompressionInfo info(dest_file_path);
			if (!info.exists()) {
				// fBNg쐬
				fs::wpath const dest_subdir_path = dest_file_path.parent_path();
				fs::create_directories(dest_subdir_path);
				// gq`FbN
				std::wstring const ext = fs::extension(dest_file_path);
				// Rs[
				if (MyLib::compareStringNoCase(ext, L".ini") != 0) {
					copyFileWithCompress(source_file_path, dest_file_path, dest_file_path.filename());
				} else {
					// init@C͂̂܂܃Rs[
					std::wstring const src_path_str = source_file_path.string();
					std::wstring const dest_path_str = dest_file_path.string();
					if (!::CopyFile(src_path_str.c_str(), dest_path_str.c_str(), FALSE)) {
						throw MY_WIN32_FILE_TRANSFER_EXCEPTION_LAST_ERROR(src_path_str, dest_path_str);
					}
				}
			}
			// F6hCo
			if (isServiceBinary(data.getServiceDataCollection(), file_title)) {
				this->installFile(source_file_path, file_title);
			}
		}
	}

	// TXTSETUP.SIF[SCSI]Ɏgp镶ێRei
	typedef std::map<std::wstring, std::wstring> MapStringToString;
	MapStringToString msts;
	// HardwareIdsDatabase
	DriverSetupData::DDInstallCollection::const_iterator it4 =
		data.getDDInstallSectionCollection().begin();
	for (; it4 != data.getDDInstallSectionCollection().end(); ++it4) {
		// T[rX
		std::wstring const &service_name = it4->second.service_name;
		if (!service_name.empty()) {
			// TXTSETUP.SIF[HardwareIdsDatabase]ɏ
			DriverSetupData::HardwareIDCollection::const_iterator it5 =
				it4->second.hardware_ids.begin();
			for (; it5 != it4->second.hardware_ids.end(); ++it5) {
				std::wstring const &hardware_id = *it5;
				this->txtsetup_sif->eraseExistingLine(L"HardwareIdsDatabase", 0, hardware_id);
				this->txtsetup_sif->insertLine(
					L"HardwareIdsDatabase", hardware_id, L'\"' + service_name + L'\"');
			}
			// TXTSETUP.SIF[SCSI]Ɏgp镶擾
			// ݂̓CINFt@CModelsZNVɂ
			// device-descriptionԍŏ̂̂gp
			// http://msdn.microsoft.com/en-us/library/ms794357.aspx
			MapStringToString::iterator it45 = msts.find(service_name);
			if (it45 == msts.end()) {
				DriverSetupData::DescriptionCollection::const_iterator it46 = it4->second.descriptions.begin();
				if (it46 != it4->second.descriptions.end()) {
					msts.insert(std::make_pair(service_name, *it46));
				}
			}
		}
	}
	// SCSI SCSI.Load
	DriverSetupData::ServiceDataCollection::const_iterator it6 =
		data.getServiceDataCollection().begin();
	for (; it6 != data.getServiceDataCollection().end(); ++it6) {
		std::wstring const &service_name = it6->first;
		// TXTSETUP.SIF[SCSI.Load]ɏ
		// service_name = service_binary, 4
		std::wstring const &service_binary = it6->second.service_binary;
		this->txtsetup_sif->eraseExistingLine(L"SCSI.Load", 0, service_name);
		MyLib::SetupFileLine::FieldCollection fields;
		fields[0] = service_name;
		fields[1] = service_binary;
		fields[2] = L"4";
		this->txtsetup_sif->insertLine(L"SCSI.Load", fields);
		// TXTSETUP.SIF[SCSI]ɏ
		// service_name = "Device Description"
		std::wstring const device_desc = msts[it6->first];
		this->txtsetup_sif->eraseExistingLine(L"SCSI", 0, service_name);
		this->txtsetup_sif->insertLine(L"SCSI", service_name, L'\"' + device_desc + L'\"');
		// SETUPREG.HIV
		this->writeServiceRegistory(it6->second.add_regs);
	}

	//
	// catt@CRs[
	//
	std::wstring const cat_file_title = data.getOriginalCatalogFileTitle();
	fs::wpath const dest_cat_file_path = dest_dir_path/cat_file_title;
	fs::wpath const src_cat_file_path = data.getCatalogFilePath();
	// ̃t@C݂Ȃ΃XLbv
	// (݂Ȃt@CINFt@Cɏ[J[̂ŁBIntelƂAIntelƂAc)
	if (fs::exists(src_cat_file_path) && !fs::exists(dest_cat_file_path)) {
		std::wstring const src_path_str = src_cat_file_path.string();
		std::wstring const dest_path_str = dest_cat_file_path.string();
		if (!::CopyFile(src_path_str.c_str(), dest_path_str.c_str(), FALSE)) {
			throw MY_WIN32_FILE_TRANSFER_EXCEPTION_LAST_ERROR(src_path_str, dest_path_str);
		}
	}

	if (is_new_dir_created) {
		this->driver_dests[installation_root_str] = dest_dir_path;
	}
}

/////////////////////////////////////////////////////////////////////////////
// hCo폜֌W
/////////////////////////////////////////////////////////////////////////////

namespace {
bool isFileInList(fs::wpath const &file_path,
                  std::list<fs::wpath> const &file_list,
                  fs::wpath const &dir_path)
{
	fs::wpath const abs_file_path = dir_path/file_path;
	std::list<fs::wpath>::const_iterator it = file_list.begin();
	for (; it != file_list.end(); ++it) {
		fs::wpath const abs_path_in_list = dir_path/(*it);
		if (fs::equivalent(abs_file_path, abs_path_in_list)) {
			return true;
		}
	}
	return false;
}
};


void Integrator::removeDriver(DriverFileDataMap const &leave_drivers,
                              DriverFileDataMap const &del_drivers,
                              DriverServiceData const &leave_services,
                              DriverServiceData const &del_services)
{
	// ǉhCoǗptH_̃pX
	fs::wpath const drivers_dir_path = this->work_dir_path/L"$OEM$\\Drivers";

	// t@C폜
	{
	DriverFileDataMap::const_iterator it1 = del_drivers.begin();
	for (; it1 != del_drivers.end(); ++it1) {
		std::wstring const drivers_sub_dir_name = it1->first;
		fs::wpath const drivers_sub_dir_path = drivers_dir_path/drivers_sub_dir_name;
		DriverFileDataMap::const_iterator it2 = leave_drivers.find(drivers_sub_dir_name);
		if (it2 == leave_drivers.end()) {
			deleteDirectory(drivers_sub_dir_path);
		} else {
			std::list<fs::wpath> const &del_file_list = it1->second.getFileList();
			std::list<fs::wpath> const &leave_file_list = it2->second.getFileList();
			std::list<fs::wpath>::const_iterator it11 = del_file_list.begin();
			for (; it11 != del_file_list.end(); ++it11) {
				fs::wpath const rel_p = *it11;
				if (!isFileInList(rel_p, leave_file_list, drivers_sub_dir_path)) {
					fs::wpath const p = drivers_sub_dir_path/rel_p;
					MyLib::toFileWritable(p);
					fs::remove(p);
				}
			}
		}
	}
	}


	StringSetNoCase del_service_binaries;
	{
	StringSetNoCase const &del_service_list = del_services.getServiceList();
	StringSetNoCase const &leave_service_list = leave_services.getServiceList();
	if (!del_service_list.empty()) {
		CRegKey reg;
		reg.Open(*this->setupreg_hiv, L"ControlSet001\\Services");
		StringSetNoCase::const_iterator it1 = del_service_list.begin();
		for (; it1 != del_service_list.end(); ++it1) {
			std::wstring const service_name = *it1;
			if (leave_service_list.find(service_name) == leave_service_list.end()) {
				// HardwareIdsDatabase
				this->txtsetup_sif->eraseExistingLine(L"HardwareIdsDatabase", 1, service_name);
				// SCSI
				this->txtsetup_sif->eraseExistingLine(L"SCSI", 0, service_name);
				// SCSI.Load
				const MyLib::SetupFileLineIterator2 it01_end;
				MyLib::SetupFileLineIterator2 it01 = this->txtsetup_sif->findFirstLine(L"SCSI.Load");
				it01 = this->txtsetup_sif->findMatchLine(it01, 0, service_name.c_str());
				for (; it01 != it01_end; it01 = this->txtsetup_sif->findMatchLine(it01, 0, service_name.c_str())) {
					std::wstring const service_binary = it01->getStringField(1);
					if (del_service_binaries.find(service_binary) == del_service_binaries.end()) {
						del_service_binaries.insert(service_binary);
					}
					it01 = this->txtsetup_sif->eraseLine(it01);
				}
				// SETUPREG.HIV̍XV
				reg.RecurseDeleteKey(service_name.c_str());
			}
		}
	}
	}

	{
	StringSetNoCase::const_iterator it1 = del_service_binaries.begin();
	for (; it1 != del_service_binaries.end(); ++it1) {
		std::wstring const filename = *it1;
		// TXTSETUP.SIF ̏C
		std::wstring const decorated_sdf = L"SourceDisksFiles" + this->section_decorate;
		MyLib::SetupFileLineIterator2 it =
			this->txtsetup_sif->findFirstLine(decorated_sdf.c_str(), filename.c_str());
		const MyLib::SetupFileLineIterator2 it_end;
		if (it == it_end) {
			it = this->txtsetup_sif->findFirstLine(L"SourceDisksFiles", filename.c_str());
		}
		if (it != it_end) {
			std::wstring const sdn_id = it->getStringField(1);
			MyLib::SetupFileLineIterator2 it2 =
				this->txtsetup_sif->findFirstLine((L"SourceDisksNames" + this->section_decorate).c_str(), sdn_id.c_str());
			if (it2 == it_end) {
				it2 = this->txtsetup_sif->findFirstLine(L"SourceDisksNames", sdn_id.c_str());
			}
			if (it2 != it_end && it2->getFieldCount() >= 4) {
				std::wstring const subdir_name = it2->getStringField(4);
				fs::wpath const existing_path = this->work_dir_path/subdir_name/filename;
				// gq̍Ōオ_ɂȂĂ邩H
				MyLib::FileCompressionInfo info(existing_path.string().c_str());
				if (info.exists()) {
					// ̃t@C폜
					removeFileForcibly(info.getActualPath());
				}
			}
			this->txtsetup_sif->eraseLine(it);
		}
		// DOSNET.INF ̏C
		MyLib::SetupFileLineIterator2 it12 = this->dosnet_inf->findFirstLine(L"FloppyFiles.3");
		it12 = this->dosnet_inf->findMatchLine(it12, 2, filename.c_str());
		if (it12 != it_end) {
			this->dosnet_inf->eraseLine(it12);
		}
	}
	}
}

/////////////////////////////////////////////////////////////////////////////

// F6 driver service binary
bool Integrator::installFile(boost::filesystem::wpath const &source_file_path,
                             std::wstring const &file_title)
{
	std::wstring const decorated_sdf = L"SourceDisksFiles" + this->section_decorate;
	// ds폜
	// check file version
	MyLib::SetupFileLineIterator2 it =
		this->txtsetup_sif->findFirstLine(decorated_sdf.c_str(), file_title.c_str());
	const MyLib::SetupFileLineIterator2 it_end;
	if (it == it_end) {
		it = this->txtsetup_sif->findFirstLine(L"SourceDisksFiles", file_title.c_str());
	}
	if (it != it_end) {
		std::wstring const sdn_id = it->getStringField(1);
		MyLib::SetupFileLineIterator2 it2 =
			this->txtsetup_sif->findFirstLine((L"SourceDisksNames" + this->section_decorate).c_str(), sdn_id.c_str());
		if (it2 == it_end) {
			it2 = this->txtsetup_sif->findFirstLine(L"SourceDisksNames", sdn_id.c_str());
		}
		if (it2 != it_end && it2->getFieldCount() >= 4) {
			std::wstring const subdir_name = it2->getStringField(4);
			fs::wpath const existing_path = this->work_dir_path/subdir_name/file_title;
			// gq̍Ōオ_ɂȂĂ邩H
			MyLib::FileCompressionInfo info(existing_path.string().c_str());
			if (info.exists()) {
				//bool const is_compressed = info.isCabinet();
				//fs::wpath existing_file_path;
				//// t@C
				//std::wstring const temp_dir_path = MyLib::createTemporaryFileName();
				//if (is_compressed) {
				//	fs::create_directories(temp_dir_path);
				//}
				//existing_file_path = info.decompressOrCopyFile(temp_dir_path).str();
				MyLib::FileVersionInfo ver_info_existing(existing_path);
				MyLib::FileVersionInfo ver_info_new(source_file_path);
				// versionr
				uint64_t existing_ver = ver_info_existing.getFileVersion();///getFileVersion(existing_file_path.string());
				uint64_t new_ver = ver_info_new.getFileVersion();//getFileVersion(source_file_path.string());
				//if (existing_ver == 0 && new_ver == 0) {
					//// versionȂ΍XVr
					//FILETIME existing_time = getFileTime(existing_file_path.string());
					//FILETIME new_time = getFileTime(source_file_path.string());
					//if (::CompareFileTime(&existing_time, &new_time) >= 0) {
					//	//if (is_compressed) {
					//	//	deleteDirectory(temp_dir_path);
					//	//}
					//	return false;
					//}
				//} else {
					if (existing_ver >= new_ver) {
						//if (is_compressed) {
						//	deleteDirectory(temp_dir_path);
						//}
						return false;
					}
				//}
				//// 𓀂t@C폜
				//if (is_compressed) {
				//	deleteDirectory(temp_dir_path);
				//}
				// ̃t@C폜
				removeFileForcibly(info.getActualPath());
			}
		}
		this->txtsetup_sif->eraseLine(it);
	}

	copyFileWithCompress(source_file_path, this->target_path/file_title, file_title);
	//// t@C̖"_"ɒu
	//std::wstring name_compressed = file_title;
	//name_compressed[name_compressed.length() - 1] = L'_';
	//// t@C̈k
	//fs::wpath const dest_path = this->target_path/name_compressed;
	//MyLib::CabCompressor compress(dest_path.string().c_str());
	//compress.addFile(source_file_path, file_title);
	//compress.flush(false);

	//
	// TXTSETUP.SIF̏
	//
	// s̒ǉ
	{
		MyLib::SetupFileLine::FieldCollection fields;
		fields[0] = file_title;
		fields[1] = L'1';
		fields[7] = L"4_";
		fields[8] = L'4';
		fields[9] = L'1';
		// XPȍ~ŐVǉꂽtB[h
		if (this->os_ver >= 0x00050001) {
			fields[12] = L'1';
			fields[13] = L'4';
		}
		this->txtsetup_sif->insertLine(decorated_sdf, fields);
	}
	//
	// DOSNET.INF̏
	//
	{
		MyLib::SetupFileLine::FieldCollection fields;
		fields[1] = L"d1";
		fields[2] = file_title;
		if (!this->dosnet_inf->existsMatchLine(L"FloppyFiles.3", fields)) {
			this->dosnet_inf->insertLine(L"FloppyFiles.3", fields);
		}
	}
	return true;
}

bool Integrator::writeServiceRegistory(const DriverSetupData::RegDataCollection &regs)
{
	// SETUPREG.HIV̍XV
	DriverSetupData::RegDataCollection::const_iterator it3 = regs.begin();
	for (; it3 != regs.end(); ++it3) {
		MyLib::SetupFileLine::FieldCollection fields;
		fields[1] = L"HKLM";
		fields[2] = L"\"SYSTEM\\CurrentControlSet\\Services\\" + it3->subkey + L"\"";
		if ((it3->flags & FLG_ADDREG_KEYONLY) == 0) {
			fields[3] = L'"' + it3->entry + L'"';
			MyLib::SetupFileLine::FieldCollection fields2;
			fields2[1] = L"HKLM";
			fields2[2] = L"SYSTEM\\CurrentControlSet\\Services\\" + it3->subkey;
			fields2[3] = it3->entry;
			MyLib::SetupFileLineIterator2 it = this->hivesys_inf->findFirstLine(L"AddReg");
			MyLib::SetupFileLineIterator2 it_end;
			while ((it = this->hivesys_inf->findMatchLine(it, fields2)) != it_end) {
				//int n = it->getFieldCount();
				//for (int i = 1; i <= n; ++i) {
				//	std::wstring str = it->getStringField(i);
				//	int l = str.length();
				//}
				it = this->hivesys_inf->eraseLine(it);
			}
		} else {
			MyLib::SetupFileLine::FieldCollection fields2;
			fields2[1] = L"HKLM";
			fields2[2] = L"SYSTEM\\CurrentControlSet\\Services\\" + it3->subkey;
			MyLib::SetupFileLineIterator2 it = this->hivesys_inf->findFirstLine(L"AddReg");
			MyLib::SetupFileLineIterator2 it_end;
			while ((it = this->hivesys_inf->findMatchLine(it, fields2)) != it_end) {
				it = this->hivesys_inf->eraseLine(it);
			}
		}
		wchar_t buf[11];
		::wsprintf(buf, L"0x%08X", it3->flags);
		fields[4] = buf;

		CRegKey reg;
		std::wstring const key_name = L"ControlSet001\\Services\\" + it3->subkey;
		reg.Create(*this->setupreg_hiv, key_name.c_str());
		if (!it3->value.empty()) {
			switch (it3->flags & FLG_ADDREG_TYPE_MASK) {
#if (_ATL_VER >= 0x0700)
			case FLG_ADDREG_TYPE_BINARY:
				{
					reg.SetBinaryValue(it3->entry.c_str(),
						&it3->value.front(), static_cast<uint32_t>(it3->value.size()));
					for (int i = 0; i < it3->value.size(); ++i) {
						::wsprintf(buf, L"%02X", it3->value[i]);
						fields[i + 5] = buf;
					}
				}
				break;
			case FLG_ADDREG_TYPE_DWORD:
				{
					DWORD v = *reinterpret_cast<const DWORD*>(&it3->value.front());
					reg.SetDWORDValue(it3->entry.c_str(),
						v);
					::wsprintf(buf, L"0x%08X", v);
					fields[5] = buf;
				}
				break;
			case FLG_ADDREG_TYPE_SZ:
			case FLG_ADDREG_TYPE_EXPAND_SZ:
				{
					wchar_t const *v = reinterpret_cast<wchar_t const*>(&it3->value.front());
					reg.SetStringValue(it3->entry.c_str(), v);
					fields[5] = std::wstring(L"\"") + v + L"\"";
				}
				break;
			case FLG_ADDREG_TYPE_MULTI_SZ:
				{
					wchar_t const *v = reinterpret_cast<wchar_t const*>(&it3->value.front());
					reg.SetMultiStringValue(it3->entry.c_str(), v);
					wchar_t const *p = v;
					size_t l = wcslen(p);
					for (int i = 0; l > 0 ; ++i) {
						fields[i + 5] = std::wstring(L"\"") + v + L"\"";
						p += l + 1;
						l = wcslen(p);
					}
				}
				break;
#else
			case FLG_ADDREG_TYPE_BINARY:
				{
					::RegSetValueEx(reg, it3->entry.c_str(), 0, REG_BINARY,
						reinterpret_cast<const BYTE*>(&it3->value.front()), static_cast<DWORD>(it3->value.size()));
					for (int i = 0; i < it3->value.size(); ++i) {
						::wsprintf(buf, L"%02X", it3->value[i]);
						fields[i + 5] = buf;
					}
				}
				break;
			case FLG_ADDREG_TYPE_DWORD:
				{
					DWORD v = *reinterpret_cast<const DWORD*>(&it3->value.front());
					reg.SetValue(v, it3->entry.c_str());
					::wsprintf(buf, L"0x%08X", v);
					fields[5] = buf;
				}
				break;
			case FLG_ADDREG_TYPE_SZ:
			case FLG_ADDREG_TYPE_EXPAND_SZ:
				{
					wchar_t const *v = reinterpret_cast<wchar_t const*>(&it3->value.front());
					reg.SetValue(v, it3->entry.c_str());
					fields[5] = std::wstring(L"\"") + v + L"\"";
				}
				break;
			case FLG_ADDREG_TYPE_MULTI_SZ:
				{
					wchar_t const *v = reinterpret_cast<wchar_t const*>(&it3->value.front());
					::RegSetValueEx(reg, it3->entry.c_str(), 0, REG_MULTI_SZ,
						reinterpret_cast<BYTE const*>(v), static_cast<DWORD>(it3->value.size()));
					wchar_t const *p = v;
					size_t l = wcslen(p);
					for (int i = 0; l > 0 ; ++i) {
						fields[i + 5] = std::wstring(L"\"") + v + L"\"";
						p += l + 1;
						l = wcslen(p);
					}
				}
				break;
#endif
			}
		}
		this->hivesys_inf->insertLine(L"AddReg", fields);
	}

	return true;
}

void Integrator::adaptBigDrive()
{
    CRegKey reg;
    const long ret1 =
		reg.Create(*this->setupreg_hiv, L"ControlSet001\\Services\\atapi\\Parameters");
    if (ret1 != NO_ERROR) {
        //throw Win32Error(ret1);
		throw MY_WIN32_EXCEPTION(ret1);
    }
#if (_ATL_VER >= 0x0700)
    const long ret2 = reg.SetDWORDValue(_T("EnableBigLba"), 1);
#else
	const long ret2 = reg.SetValue(1, L"EnableBigLba");
#endif
    if (ret2 != NO_ERROR) {
        //throw Win32Error(ret2);
		throw MY_WIN32_EXCEPTION(ret2);
    }
}

void Integrator::mergeCabFiles(std::wstring const &temp_dir)
{
	// ꎞtH_쐬
	TemporaryDirectoryCloseDelete temp_dir_close_delete;
	TemporaryDirectory temp_dir1(temp_dir, &temp_dir_close_delete);
	fs::wpath const temp_dir_path = temp_dir1.getDirectoryPath();

	//if (this->architecture_id == PROCESSOR_ARCHITECTURE_AMD64
	// && this->architecture_id == PROCESSOR_ARCHITECTURE_IA64) {
	//	// 64bitł̏ꍇAdriver.cabdriver32.cabŉ𓀃fBNg𕪂
	//	fs::wpath const dir_path_64 = temp_dir_path/L"64";
	//	fs::create_directory(dir_path_64);
	//	fs::wpath const dir_path_32 = temp_dir_path/L"32";
	//	fs::create_directory(dir_path_32);

	//	// DRVINDEX.INFJ
	//	fs::wpath const drvindex_inf_path = this->target_path/L"drvindex.inf";
	//	std::vector<std::wstring> cab_file_titles;
	//	{
	//		MyCabDecompressor decomp;
	//		MyLib::SetupFileReader reader(drvindex_inf_path.string());
	//		MyLib::SetupFileLineIterator it = reader.findFirstLine(L"Cabs");
	//		const MyLib::SetupFileLineIterator it_end;
	//		for (; it != it_end; ++it) {
	//			std::wstring const cab_file_title = it->getStringField(1);
	//			cab_file_titles.push_back(cab_file_title);
	//			fs::wpath const cab_file_path = this->target_path/cab_file_title;
	//			// 𓀂TemptH_ɒu
	//			// TODO: driver32.cab ͕ʃfBNgɉ
	//			// TODO: o[W`FbN(I386ɂt@CƂr)
	//			if (!decomp.extractCabFile(cab_file_path.string(), temp_dir_path.string())) {
	//				deleteDirectory(temp_dir_path);
	//				throw MY_WIN32_FILE_EXCEPTION_LAST_ERROR(cab_file_path.string());
	//			}
	//		}
	//	}
	//}

	// DRVINDEX.INFJ
	fs::wpath const drvindex_inf_path = this->target_path/L"drvindex.inf";
	static wchar_t const *DRIVER_CAB_FILENAME = L"driver.cab";
	std::wstring driver_cab_tag;
	std::wstring spx_cab_filename;
	std::wstring spx_cab_tag;
	// new driver.cab
	fs::wpath const driver_cab_temp_path = MyLib::createTemporaryFilePath(temp_dir);

	boost::scoped_ptr<MyLib::SetupFileReader> reader(new MyLib::SetupFileReader(drvindex_inf_path.string()));
	MyLib::SetupFileLineIterator it = reader->findFirstLine(L"Cabs");
	const MyLib::SetupFileLineIterator it_end;
	for (; it != it_end; ++it) {
		std::wstring const cab_file_title = it->getStringField(1);
		// r₷悤ɏ
		size_t const cab_file_title_len = cab_file_title.length();
		size_t const cab_file_title_buf_size = (cab_file_title_len + 1)*sizeof(wchar_t);
		boost::scoped_array<wchar_t> cab_file_title_buf(new wchar_t[cab_file_title_buf_size]);
		std::memcpy(cab_file_title_buf.get(), cab_file_title.c_str(), cab_file_title_buf_size);
		::CharLower(cab_file_title_buf.get());
		if (cab_file_title_len == 10 && std::memcmp(cab_file_title_buf.get(), DRIVER_CAB_FILENAME, cab_file_title_buf_size) == 0) {
			driver_cab_tag = it->getStringField(0);
		} else if (cab_file_title_len == 7
		        && std::memcmp(cab_file_title_buf.get(), L"sp", 2*sizeof(wchar_t)) == 0
		        && L'1' <= cab_file_title_buf[2] && cab_file_title_buf[2] <= L'9'
		        && std::memcmp(cab_file_title_buf.get() + 3, L".cab", 4*sizeof(wchar_t)) == 0) {
			spx_cab_filename = cab_file_title;
			spx_cab_tag = it->getStringField(0);
		}
	}

	// extract driver.cab
	//if (driver_cab_tag.empty()) {
	//	// error
	//}
	fs::wpath const driver_cab_path = this->target_path/DRIVER_CAB_FILENAME;
	MyCabDecompressor decomp;
	if (!decomp.extractCabFile(driver_cab_path.string(), temp_dir_path.string())) {
		throw MY_WIN32_FILE_EXCEPTION_LAST_ERROR(driver_cab_path.string());
	}
	if (!spx_cab_filename.empty()) {
		// extract spX.cab
		fs::wpath const spx_cab_path = this->target_path/spx_cab_filename;
		if (!decomp.extractCabFile(spx_cab_path.string(), temp_dir_path.string())) {
			throw MY_WIN32_FILE_EXCEPTION_LAST_ERROR(spx_cab_path.string());
		}
	}

	{
		MyLib::CabCompressor compressor(driver_cab_temp_path);
		fs::wdirectory_iterator it_end;
		for (fs::wdirectory_iterator it(temp_dir_path); it != it_end; ++it) {
			fs::wpath file_path = it->path();
			// LZX21`ňk
			compressor.addFile(file_path, false, TCOMPfromLZXWindow(21));
		}
		//compressor.flush(false);
	}	// compressor ̃fXgN^Ăяo
	// overwrite driver.cab
	MyLib::toFileWritable(driver_cab_path);
	::CopyFile(driver_cab_temp_path.string().c_str(), driver_cab_path.string().c_str(), FALSE);
	::DeleteFile(driver_cab_temp_path.string().c_str());

	if (!spx_cab_filename.empty()) {
		// modify drvindex.inf
		// new drvindex.inf
		fs::wpath const drvindex_inf_temp_path = MyLib::createTemporaryFilePath(temp_dir);
		{
			fs::wofstream ofs(drvindex_inf_temp_path);
			ofs << L"[Version]\nSignature=\"$Windows NT$\"\nCabFiles=";
			it = reader->findFirstLine(L"Version", L"CabFiles");
			unsigned int const cabs_tag_count = it->getFieldCount();
			std::vector<std::wstring> cabs_tags;
			for (unsigned int i = 1; i <= cabs_tag_count; ++i) {
				std::wstring const cabs_tag = it->getStringField(i);
				if (MyLib::compareStringNoCase(cabs_tag, spx_cab_tag) != 0) {
					if (!cabs_tags.empty()) {
						ofs << L',';
					}
					ofs << cabs_tag;
					cabs_tags.push_back(cabs_tag);
				}
			}
			ofs << L"\n[" << driver_cab_tag << L"]\n";
			for (it = reader->findFirstLine(driver_cab_tag.c_str()); it != it_end; ++it) {
				ofs << it->getStringField(1) << L'\n';
			}
			for (it = reader->findFirstLine(spx_cab_tag.c_str()); it != it_end; ++it) {
				ofs << it->getStringField(1) << L'\n';
			}
			for (std::vector<std::wstring>::const_iterator it1 = cabs_tags.begin(); it1 != cabs_tags.end(); ++it1) {
				if (MyLib::compareStringNoCase(*it1, driver_cab_tag) != 0) {
					ofs << L'[' << *it1 << L"]\n";
					for (it = reader->findFirstLine(it1->c_str()); it != it_end; ++it) {
						ofs << it->getStringField(1) << L'\n';
					}
				}
			}
			ofs << L"[Cabs]\n";
			for (it = reader->findFirstLine(L"Cabs"); it != it_end; ++it) {
				std::wstring const cabs_tag = it->getStringField(0);
				if (MyLib::compareStringNoCase(cabs_tag, spx_cab_tag) != 0) {
					ofs << cabs_tag << L'=' << it->getStringField(1) << L'\n';
				}
			}
		}
		// overwrite drvindex.inf
		reader.reset();
		MyLib::toFileWritable(drvindex_inf_path);
		::CopyFile(drvindex_inf_temp_path.string().c_str(), drvindex_inf_path.string().c_str(), FALSE);
		::DeleteFile(drvindex_inf_temp_path.string().c_str());

		// txtsetup.sif
		// [SetupData]
		// DriverCabName = driver.cab, sp2.cab, ...
		const MyLib::SetupFileLineIterator2 it_end;
		MyLib::SetupFileLineIterator2 it1 =
			this->txtsetup_sif->findFirstLine(L"SetupData", L"DriverCabName");
		unsigned int const n = it1->getFieldCount();
		MyLib::SetupFileLine::FieldCollection field_collection;
		int cab_count = 0;
		field_collection[0] = L"DriverCabName";
		for (unsigned int i = 1; i <= n; ++i) {
			std::wstring const cab_name = it1->getStringField(i);
			if (MyLib::compareStringNoCase(cab_name, spx_cab_filename) != 0) {
				field_collection[++cab_count] = cab_name;
			}
		}
		this->txtsetup_sif->eraseLine(it1);
		this->txtsetup_sif->insertLine(L"SetupData", field_collection);
		// TXTSETUP.SIF  SPx.CAB̃Gg폜
		this->txtsetup_sif->eraseExistingLine(L"SourceDisksFiles" + this->section_decorate, 0, spx_cab_filename);
		this->txtsetup_sif->eraseExistingLine(L"SourceDisksFiles", 0, spx_cab_filename);
		this->txtsetup_sif->eraseExistingLine(L"FileFlags", 0, spx_cab_filename);
		// DOSNET.INF  SPx.CAB̃Gg폜
		this->dosnet_inf->eraseExistingLine(L"Files", 2, spx_cab_filename);
		this->dosnet_inf->eraseExistingLine(L"ForceCopyDriverCabFiles", 1, spx_cab_filename);
		// SPx.CAB 폜
		fs::wpath const p = this->target_path/spx_cab_filename;
		removeFileForcibly(p);
	}

	//std::vector<std::wstring> cab_file_titles;
	//{
	//	MyCabDecompressor decomp;
	//	MyLib::SetupFileReader reader(drvindex_inf_path.string());
	//	MyLib::SetupFileLineIterator it = reader.findFirstLine(L"Cabs");
	//	const MyLib::SetupFileLineIterator it_end;
	//	for (; it != it_end; ++it) {
	//		std::wstring const cab_file_title = it->getStringField(1);
	//		cab_file_titles.push_back(cab_file_title);
	//		fs::wpath const cab_file_path = this->target_path/cab_file_title;
	//		// 𓀂TemptH_ɒu
	//		// TODO: driver32.cab ͕ʃfBNgɉ
	//		// TODO: o[W`FbN(I386ɂt@CƂr)
	//		if (!decomp.extractCabFile(cab_file_path.string(), temp_dir_path.string())) {
	//			//deleteDirectory(temp_dir_path);
	//			throw MY_WIN32_FILE_EXCEPTION_LAST_ERROR(cab_file_path.string());
	//		}
	//	}
	//}

	//// new driver.cab
	//std::wstring const driver_cab_temp_path = MyLib::createTemporaryFilePath(temp_dir).string();
	//// new drvindex.inf
	//std::wstring const drvindex_inf_temp_path = MyLib::createTemporaryFilePath(temp_dir).string();
	//{
	//	fs::wofstream ofs(drvindex_inf_temp_path);
	//	ofs << L"[Version]\nSignature = \"$Windows NT$\"\nCabFiles = driver\n";
	//	ofs << L"[Cabs]\ndriver = driver.cab\n[driver]\n";
	//	//
	//	MyLib::CabCompressor compressor(driver_cab_temp_path.c_str());
	//	fs::wdirectory_iterator it_end;
	//	for (fs::wdirectory_iterator it(temp_dir_path); it != it_end; ++it) {
	//		fs::wpath file_path = it->path();
	//		ofs << file_path.filename() << L"\n";
	//		// LZX21`ňk
	//		compressor.addFile(file_path, false, TCOMPfromLZXWindow(21));
	//	}
	//	compressor.flush(false);
	//}	// compressor  ofs ̃fXgN^Ăяo

	//// overwrite driver.cab
	//fs::wpath const driver_cab_path = this->target_path/L"driver.cab";
	//MyLib::toFileWritable(driver_cab_path);
	//::CopyFile(driver_cab_temp_path.c_str(), driver_cab_path.string().c_str(), FALSE);
	//::DeleteFile(driver_cab_temp_path.c_str());
	//// overwrite drvindex.inf
	//MyLib::toFileWritable(drvindex_inf_path);
	//::CopyFile(drvindex_inf_temp_path.c_str(), drvindex_inf_path.string().c_str(), FALSE);
	//::DeleteFile(drvindex_inf_temp_path.c_str());

	//// [SetupData]
	//// DriverCabName = driver.cab, sp2.cab, ...
	//const MyLib::SetupFileLineIterator2 it_end;
	//MyLib::SetupFileLineIterator2 it1 =
	//	this->txtsetup_sif->findFirstLine(L"SetupData", L"DriverCabName");
	//if (it1 != it_end) {
	//	this->txtsetup_sif->eraseLine(it1);
	//	this->txtsetup_sif->insertLine(L"SetupData", L"DriverCabName", L"driver.cab");
	//}
	//// DRIVER.CAB ȊO(SPx.CAB Ȃ)̏
	//for (size_t i = 0; i < cab_file_titles.size(); ++i) {
	//	if (MyLib::compareStringNoCase(cab_file_titles[i], L"driver.cab") != 0) {
	//		// DRIVER.CAB ȊÕGg폜
	//		// TODO: driver32.cab ͍폜Ȃ
	//		this->txtsetup_sif->eraseExistingLine(L"SourceDisksFiles" + this->section_decorate, 0, cab_file_titles[i]);
	//		this->txtsetup_sif->eraseExistingLine(L"SourceDisksFiles", 0, cab_file_titles[i]);
	//		this->txtsetup_sif->eraseExistingLine(L"FileFlags", 0, cab_file_titles[i]);
	//		this->dosnet_inf->eraseExistingLine(L"Files", 2, cab_file_titles[i]);
	//		this->dosnet_inf->eraseExistingLine(L"ForceCopyDriverCabFiles", 1, cab_file_titles[i]);
	//		fs::wpath const p = this->target_path/cab_file_titles[i];
	//		// DRIVER.CAB ȊÕt@C폜
	//		removeFileForcibly(p);
	//	}
	//}
}
