#include "TemporaryDirectory.h"
#include "mylib/function.h"

extern void deleteDirectory(boost::filesystem::wpath const &dir_path);

/////////////////////////////////////////////////////////////////////////////

void TemporaryDirectoryCloseDelete::execute(boost::filesystem::wpath const &dir_path) const
{
	deleteDirectory(dir_path);
}

/////////////////////////////////////////////////////////////////////////////

TemporaryDirectoryCloseUndoMovedFiles::TemporaryDirectoryCloseUndoMovedFiles(boost::filesystem::wpath const &undo_dir_path)
	: undo_dir_path_(undo_dir_path)
{
}

void TemporaryDirectoryCloseUndoMovedFiles::execute(boost::filesystem::wpath const &dir_path) const
{
	namespace fs = boost::filesystem;
	fs::wdirectory_iterator it_end;
	for (fs::wdirectory_iterator it(dir_path); it != it_end; ++it) {
		fs::wpath const p = it->path();
		fs::wpath const dst_path = this->undo_dir_path_/p.filename();
		fs::rename(it->path(), dst_path);
	}
	fs::remove(dir_path);
}

/////////////////////////////////////////////////////////////////////////////

TemporaryDirectory::TemporaryDirectory(TemporaryDirectoryCloseBehavior const *close_behavior)
	: dir_path_(MyLib::createTemporaryFilePath())
	, close_behavior_(close_behavior)
{
	boost::filesystem::create_directories(this->dir_path_);
}

TemporaryDirectory::TemporaryDirectory(boost::filesystem::wpath const &parent_dir_path,
                                       TemporaryDirectoryCloseBehavior const *close_behavior)
	: dir_path_(MyLib::createTemporaryFilePath(parent_dir_path))
	, close_behavior_(close_behavior)
{
	boost::filesystem::create_directories(this->dir_path_);
}

TemporaryDirectory::~TemporaryDirectory()
{
	this->close_behavior_->execute(this->dir_path_);
}

boost::filesystem::wpath TemporaryDirectory::getDirectoryPath() const
{
	return this->dir_path_;
}

void TemporaryDirectory::setCloseBehavior(TemporaryDirectoryCloseBehavior const *close_behavior)
{
	this->close_behavior_ = close_behavior;
}

//	void close()
//		case DELETE_DIR:
//			deleteDirectory(this->dir_path_);
//		case UNDO_MOVED_FILES:
//			{
//				FileUndoInfo::const_iterator it = this->file_undo_info.begin();
//				for (; it != this->file_undo_info.end(); ++it) {
//					const PathString src_path = this->dir_path_ + it->first;
//					::MoveFile(src_path.str().c_str(), it->second.str().c_str());
//				}
//			}
