#include "stdafx.h"
#include "storagedevicenumber.h"
#include "drivelayoutinformation.h"
#include "filetowritable.h"
#include "inifile.h"

namespace {
size_t getFileSize(HANDLE h)
{
	LARGE_INTEGER li;
	if (!::GetFileSizeEx(h, &li)) {
		throw MyLib::Win32Exception();
	}
	return static_cast<size_t>(li.QuadPart);
}

bool areFilesEqual(std::wstring const &file1_path,
                   std::wstring const &file2_path)
{
	bool result = false;

	AutoHandle<INVALID_HANDLE_VALUE> h1(::CreateFile(file1_path.c_str(),
	                                                 GENERIC_READ,
	                                                 FILE_SHARE_READ,
	                                                 NULL,
	                                                 OPEN_EXISTING,
	                                                 FILE_ATTRIBUTE_NORMAL,
	                                                 NULL));
	if (h1.isInvalid()) {
		throw MyLib::Win32FileException(file1_path);
	}
	AutoHandle<INVALID_HANDLE_VALUE> h2(::CreateFile(file2_path.c_str(),
	                                                 GENERIC_READ,
	                                                 FILE_SHARE_READ,
	                                                 NULL,
	                                                 OPEN_EXISTING,
	                                                 FILE_ATTRIBUTE_NORMAL,
	                                                 NULL));
	if (h2.isInvalid()) {
		throw MyLib::Win32FileException(file2_path);
	}

	size_t file1_size = getFileSize(h1.get());
	size_t file2_size = getFileSize(h2.get());
	if (file1_size == file2_size) {
		DWORD bytes_read;
		boost::scoped_array<char> file1_buf(new char[file1_size]);
		boost::scoped_array<char> file2_buf(new char[file2_size]);
		if (!::ReadFile(h1.get(), file1_buf.get(), file1_size, &bytes_read, NULL)) {
			throw MyLib::Win32FileException(file1_path);
		}
		if (!::ReadFile(h2.get(), file2_buf.get(), file2_size, &bytes_read, NULL)) {
			throw MyLib::Win32FileException(file2_path);
		}
		if (std::memcmp(file1_buf.get(), file2_buf.get(), file1_size) == 0) {
			result = true;
		}
	}
	return result;
}
//=======================================================================
// 啶EʂȂ̔r֐
//=======================================================================
int compareStringNoCase(std::wstring const &left, std::wstring const &right)
{
	// lstrcmpi()ECompareString()gȂ̂́A
	// 啶E𓯈ꎋt@C̋KƈقȂ邽߁B
	// [http://d.hatena.ne.jp/hasegawayosuke/20060125/1138185283]
	// tolower()gȂ̂́A}`oCgɑΉ邽߁B

	// ׂ̕đ啶ɂ
	size_t left_buf_count = left.length() + 1;
	boost::scoped_array<wchar_t> left_buf(new wchar_t[left_buf_count]);
	std::memcpy(left_buf.get(), left.c_str(), left_buf_count*sizeof(wchar_t));
	::CharUpper(left_buf.get());
	// ׂ̕đ啶ɂ
	size_t right_buf_count = right.length() + 1;
	boost::scoped_array<wchar_t> right_buf(new wchar_t[right_buf_count]);
	std::memcpy(right_buf.get(), right.c_str(), right_buf_count*sizeof(wchar_t));
	::CharUpper(right_buf.get());
	// r
	size_t const comp_size = std::min(left_buf_count, right_buf_count)*sizeof(wchar_t);
	int const result = std::memcmp(&left_buf[0], &right_buf[0], comp_size);
	if (result == 0) {
		if (left_buf_count > right_buf_count) {
			return 1;
		}
		if (left_buf_count < right_buf_count) {
			return -1;
		}
	}
	return result;
}
};

void fixBootIni(std::wstring const &src, std::wstring const &dst);
void copyBootFiles(std::wstring const &src, std::wstring const &dst);

int WINAPI wWinMain(HINSTANCE /*hInstance*/, HINSTANCE /*hPrevInstance*/, wchar_t * /*lpstrCmdLine*/, int /*nCmdShow*/)
{
	try {

	wchar_t path_buf[MAX_PATH];
	// ZbgAbvphCȕ擾
	GetModuleFileName(NULL, path_buf, MAX_PATH);
	wchar_t const src_drive_letter = path_buf[0];
	StorageDeviceNumber sdn_src(src_drive_letter);

	wchar_t log_file_path[] = L"F:\\ldrfix.log";
	log_file_path[0] = src_drive_letter;
	std::wofstream ofs(log_file_path);
	ofs << path_buf << L'\n';
	ofs << L"Device Type : " << sdn_src.getDeviceType() << L'\n';
	ofs << L"Device Number : " << sdn_src.getDeviceNumber() << L'\n';
	ofs << L"Partition Number : " << sdn_src.getPartitionNumber() << std::endl;

	// ZbgAbvhCȕ擾
	GetSystemDirectory(path_buf, MAX_PATH);
	wchar_t const dst_drive_letter = path_buf[0];
	StorageDeviceNumber sdn_dst(dst_drive_letter);

	ofs << path_buf << L'\n';
	ofs << L"Device Type : " << sdn_dst.getDeviceType() << L'\n';
	ofs << L"Device Number : " << sdn_dst.getDeviceNumber() << L'\n';
	ofs << L"Partition Number : " << sdn_dst.getPartitionNumber() << std::endl;

	// ZbgAbvphCuƃZbgAbvhCufBXNɂ
	if (sdn_src.getDeviceNumber() == sdn_dst.getDeviceNumber()) {
		// ZbgAbvphCu boot.ini 
		// ZbgAbvphCu boot.ini Ă
		wchar_t src_boot_ini_path[] = L"A:\\boot.ini";
		src_boot_ini_path[0] = src_drive_letter;

		ofs << L"Pattern1\n";
		ofs << src_boot_ini_path << std::endl;

		FileToWritable ftw(src_boot_ini_path);
		IniFile src_boot_ini(src_boot_ini_path);
		src_boot_ini.setString(L"operating systems", L"C:\\$WIN_NT$.~BT\\BOOTSECT.DAT", L"\"Windows ZbgAbv Jn\"");
		return 0;
	}

	// fBXÑp[eBV\擾
	unsigned int dst_device_num = sdn_dst.getDeviceNumber();
	DriveLayoutInformation dli(dst_device_num);
	int active_partition = dli.getActivePartition();
	// ANeBup[eBVAFł
	if (active_partition != -1
	 && IsRecognizedPartition(dli.getPartitionEntry(active_partition)->PartitionType)) {
		wchar_t src_drive_root[] = L"A:";
		src_drive_root[0] = src_drive_letter;
		std::wstringstream ss;
		ss << L"\\\\?\\GLOBALROOT\\Device\\Harddisk"
		   << dst_device_num
		   << L"\\Partition"
		   << dli.getPartitionEntry(active_partition)->PartitionNumber;

		ofs << L"Pattern2\n";
		ofs << L"Source : " << src_drive_root << L'\n';
		ofs << L"Destination : " << ss.str() << std::endl;

		// ANeBup[eBV ntldr ȂǂRs[
		copyBootFiles(src_drive_root, ss.str());
		// ANeBup[eBV boot.ini 
		fixBootIni(src_drive_root, ss.str());
		return 0;
	}
	// FłANeBup[eBVȂ
	for (unsigned int i = 0; i < std::min(dli.getPartitionCount(), (unsigned int)4); ++i) {
		// FłAł擪ɂp[eBVANeBuɂ
		if (IsRecognizedPartition(dli.getPartitionEntry(i)->PartitionType)) {
			ofs << L"active partition : " << i << std::endl;
			dli.setActivePartiton(i);

			wchar_t src_drive_root[] = L"A:";
			src_drive_root[0] = src_drive_letter;
			std::wstringstream ss;
			ss << L"\\\\?\\GLOBALROOT\\Device\\Harddisk"
			   << dst_device_num
			   << L"\\Partition"
			   << dli.getPartitionEntry(i)->PartitionNumber;

			ofs << L"Pattern3\n";
			ofs << L"Source : " << src_drive_root << L'\n';
			ofs << L"Destination : " << ss.str() << std::endl;

			// ANeBup[eBV ntldr ȂǂRs[
			copyBootFiles(src_drive_root, ss.str());
			// ANeBup[eBV boot.ini 
			fixBootIni(src_drive_root, ss.str());
			break;
		}
	}

	} catch (MyLib::FileTransferException<wchar_t>& e) {
		std::wstringstream ss;
		ss << L"From: " << e.getSourceFileName() << L"\n";
		ss << L"To: " << e.getDestinationFileName() << L"\n";
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText() << L'[' << e.getErrorCode() << L']';
		::MessageBox(NULL, ss.str().c_str(), NULL, 0);
	} catch (MyLib::FileException<wchar_t>& e) {
		std::wstringstream ss;
		ss << e.getFileName() << L"\n";
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText() << L'[' << e.getErrorCode() << L']';
		::MessageBox(NULL, ss.str().c_str(), NULL, 0);
	} catch (MyLib::Exception<wchar_t>& e) {
		std::wstringstream ss;
		if (e.getErrorFile() != NULL) {
			ss << e.getErrorFile() << L"(" << e.getErrorLine() << L")";
		}
		ss << e.getErrorText() << L'[' << e.getErrorCode() << L']';
		::MessageBox(NULL, ss.str().c_str(), NULL, 0);
	} catch (const std::runtime_error& e) {
		::MessageBoxA(NULL, e.what(), NULL, 0);
	}

	return 0;
}

void copyBootFiles(std::wstring const &src, std::wstring const &dst)
{
	static wchar_t const *NTLDR_FILENAME = L"\\NTLDR";
	static wchar_t const *NTDETECT_COM_FILENAME = L"\\NTDETECT.COM";
	static wchar_t const *BOOTFONT_BIN_FILENAME = L"\\BOOTFONT.BIN";
	static wchar_t const *NTBOOTDD_SYS_FILENAME = L"\\NTBOOTDD.SYS";

	std::wstring const src_ntldr_path = src + NTLDR_FILENAME;
	{
		std::wstring const dst_ntldr_path = dst + NTLDR_FILENAME;
		FileToWritable ftw(dst_ntldr_path);
		if (!::CopyFile(src_ntldr_path.c_str(), dst_ntldr_path.c_str(), FALSE)) {
			throw MyLib::Win32FileTransferException(src_ntldr_path, dst_ntldr_path);
		}
	}
	std::wstring const src_ntdetect_com_path = src + NTDETECT_COM_FILENAME;
	{
		std::wstring const dst_ntdetect_com_path = dst + NTDETECT_COM_FILENAME;
		FileToWritable ftw(dst_ntdetect_com_path);
		if (!::CopyFile(src_ntdetect_com_path.c_str(), dst_ntdetect_com_path.c_str(), FALSE)) {
			throw MyLib::Win32FileTransferException(src_ntdetect_com_path, dst_ntdetect_com_path);
		}
	}
	std::wstring const src_bootfont_bin_path = src + BOOTFONT_BIN_FILENAME;
	if (::GetFileAttributes(src_bootfont_bin_path.c_str()) != 0xFFFFFFFF) {
		std::wstring const dst_bootfont_bin_path = dst + BOOTFONT_BIN_FILENAME;
		FileToWritable ftw(dst_bootfont_bin_path);
		if (!::CopyFile(src_bootfont_bin_path.c_str(), dst_bootfont_bin_path.c_str(), FALSE)) {
			throw MyLib::Win32FileTransferException(src_bootfont_bin_path, dst_bootfont_bin_path);
		}
	}
	std::wstring const src_ntbootdd_sys_path = src + NTBOOTDD_SYS_FILENAME;
	if (::GetFileAttributes(src_ntbootdd_sys_path.c_str()) != 0xFFFFFFFF) {
		std::wstring const dst_ntbootdd_sys_path = dst + NTBOOTDD_SYS_FILENAME;
		FileToWritable ftw(dst_ntbootdd_sys_path);
		if (!::CopyFile(src_ntbootdd_sys_path.c_str(), dst_ntbootdd_sys_path.c_str(), FALSE)) {
			throw MyLib::Win32FileTransferException(src_ntbootdd_sys_path, dst_ntbootdd_sys_path);
		}
	}
}

void fixBootIni(std::wstring const &src, std::wstring const &dst)
{
	std::wstring const src_boot_ini_path = src + L"\\boot.ini";
	std::wstring const dst_boot_ini_path = dst + L"\\boot.ini";
	std::wstring const bootorig_ini_path = src + L"\\bootorig.ini";
	if (areFilesEqual(src_boot_ini_path, bootorig_ini_path)) {
		// XP/2003 USB
		IniFile dst_boot_ini(dst_boot_ini_path);
		std::wstring const default_entry =
			dst_boot_ini.getString(L"boot loader", L"default", L"");
		std::wstring const entry_desc =
			dst_boot_ini.getString(L"operating systems", default_entry, L"");
		std::wstring const bus_type = default_entry.substr(0, 4);
		if (compareStringNoCase(bus_type, L"mult") == 0) {
			FileToWritable ftw(dst_boot_ini_path);
			// ȑÕGg폜
			dst_boot_ini.deleteKey(L"operating systems", default_entry);
			// rdisk l 0 ɂ
			std::wstring new_entry = default_entry;
			new_entry[21] = L'0';
			dst_boot_ini.setString(L"boot loader", L"default", new_entry);
			dst_boot_ini.setString(L"operating systems", new_entry, entry_desc);
		}
	} else {
		// another HDD or 2000 USB
		// ZbgAbvphCu boot.ini Ă
		{
			FileToWritable ftw(dst_boot_ini_path);
			if (!::CopyFile(src_boot_ini_path.c_str(), dst_boot_ini_path.c_str(), FALSE)) {
				throw MyLib::Win32FileTransferException(src_boot_ini_path, dst_boot_ini_path);
			}
		}
		{
			IniFile dst_boot_ini(dst_boot_ini_path);
			std::wstring const default_entry =
				dst_boot_ini.getString(L"boot loader", L"default", L"");
			std::wstring const entry_desc =
				dst_boot_ini.getString(L"operating systems", default_entry, L"");
			std::wstring const bus_type = default_entry.substr(0, 4);
			if (compareStringNoCase(bus_type, L"mult") == 0) {
				FileToWritable ftw(dst_boot_ini_path);
				// ȑÕGg폜
				dst_boot_ini.deleteKey(L"operating systems", default_entry);
				// rdisk l 0 ɂ
				std::wstring new_entry = default_entry;
				new_entry[21] = L'0';
				dst_boot_ini.setString(L"boot loader", L"default", new_entry);
				dst_boot_ini.setString(L"operating systems", new_entry, entry_desc);
			}
		}
		{
			FileToWritable ftw(src_boot_ini_path);
			::CopyFile(bootorig_ini_path.c_str(), src_boot_ini_path.c_str(), FALSE);
		}
	}
}
