#ifndef CAB_DECOMPRESSOR_H__15B97EE0_A438_472F_A63F_C3B89F55C51D__INCLUDED
#define CAB_DECOMPRESSOR_H__15B97EE0_A438_472F_A63F_C3B89F55C51D__INCLUDED
// M$ compatible compilers support #pragma once
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

/////////////////////////////////////////////////////////////////////////////
// CabDecompressor.h
//	Cabinet.dllFDI(File Decompression Interface)gpāA
//	t@CCabinetWJ
// XVF
//	H20(2008)/05/20 쐬
//	H21(2009)/06/06 boost::filesystem, boost::scoped_array ̓
// Classes in this file:
//	FDIError
//	CabDecompressorT
//	CabDecompressor
/////////////////////////////////////////////////////////////////////////////

	//MyLib::CabDecompressor decomp;
	//boost::filesystem::wpath target_path = L"D:\\driver_cab";
	//boost::filesystem::wpath path = L"D:\\temp\\wxpvol_ja0\\i386\\driver.cab";
	//FDICABINETINFO fdici = decomp.getCabFileInfo(path);
	//printf(
	//	"   Total length of cabinet file : %d\n"
	//	"   Number of folders in cabinet : %d\n"
	//	"   Number of files in cabinet   : %d\n"
	//	"   Cabinet set ID               : %d\n"
	//	"   Cabinet number in set        : %d\n"
	//	"   RESERVE area in cabinet?     : %s\n"
	//	"   Chained to prev cabinet?     : %s\n"
	//	"   Chained to next cabinet?     : %s\n"
	//	"\n",
	//	fdici.cbCabinet,
	//	fdici.cFolders,
	//	fdici.cFiles,
	//	fdici.setID,
	//	fdici.iCabinet,
	//	fdici.fReserve == TRUE ? "yes" : "no",
	//	fdici.hasprev == TRUE ? "yes" : "no",
	//	fdici.hasnext == TRUE ? "yes" : "no");

	//if (decomp.extractCabFile(path, target_path)) {
	//	printf("succeeded\n");
	//} else {
	//	printf("failed\n");
	//}

namespace MyLib {
//=======================================================================
// FDIError
//	FDIG[\ERF̗ONX
//=======================================================================
class FDIError :
	public std::runtime_error
{
	ERF err_code;

public:
	FDIError(const ERF& code);
	static const std::string makeErrorMessage(const ERF& code);
};

inline FDIError::FDIError(const ERF& code)
: std::runtime_error(makeErrorMessage(code))
, err_code(code)
{
}

// FDĨG[Ԃ
inline const std::string FDIError::makeErrorMessage(const ERF& code)
{
	//std::string result;
	switch (code.erfOper) {
	case FDIERROR_NONE:
		return "No error";
	case FDIERROR_CABINET_NOT_FOUND:
		return "Cabinet not found";
	case FDIERROR_NOT_A_CABINET:
		return "Not a cabinet";
	case FDIERROR_UNKNOWN_CABINET_VERSION:
		return "Unknown cabinet version";
	case FDIERROR_CORRUPT_CABINET:
		return "Corrupt cabinet";
	case FDIERROR_ALLOC_FAIL:
		return "Memory allocation failed";
	case FDIERROR_BAD_COMPR_TYPE:
		return "Unknown compression type";
	case FDIERROR_MDI_FAIL:
		return "Failure decompressing data";
	case FDIERROR_TARGET_FILE:
		return "Failure writing to target file";
	case FDIERROR_RESERVE_MISMATCH:
		return "Cabinets in set have different RESERVE sizes";
	case FDIERROR_WRONG_CABINET:
		return "Cabinet returned on fdintNEXT_CABINET is incorrect";
	case FDIERROR_USER_ABORT:
		return "User aborted";
	default:
		return "Unknown error";
	}
	//return result;
}

//=======================================================================
// CabDecompressorT
//	CabinetWJ̊NX
//=======================================================================
template <class T>
class CabDecompressorT
{
public:
	CabDecompressorT();
	virtual ~CabDecompressorT();

	bool isCabinet(boost::filesystem::wpath const &path);
	FDICABINETINFO getCabFileInfo(boost::filesystem::wpath const &path);
	bool extractCabFile(boost::filesystem::wpath const &cab_file_path, boost::filesystem::wpath const &target_dir_path);

protected:
	boost::filesystem::wpath target_dir_path;

	// overrides
	void onCabinetInfo(
		const char *next_cabinet,
		const char *next_disk,
		const char *cabinet_path,
		unsigned short cabinet_set_id,
		unsigned short cabinet_number_in_set);

	void onPartialFile(
		const char *continued_file_name,
		const char *file_start_cabinet_name,
		const char *file_start_disk_name);

	enum CopyFileFlags {
		EXTRACT,
		SKIP,
		ABORT
	};
	enum CopyFileFlags onCopyFile(
		const std::wstring& file_name_in_cab,
		size_t uncompressed_file_size,
		FILETIME file_time,
		unsigned short attribs);

	bool onCloseFileInfo(
		const char *file_name_in_cab,
		intptr_t file_handle,
		FILETIME file_time,
		unsigned short attribs,
		bool executes_after_extract);

private:
	HFDI fdi_handle;
	ERF err_code;

	static FNALLOC(onAlloc);
	static FNFREE(onFree);
	static FNOPEN(onOpen);
	static FNREAD(onRead);
	static FNWRITE(onWrite);
	static FNCLOSE(onClose);
	static FNSEEK(onSeek);
	static FNFDINOTIFY(onNotify);
};

template <class T>
inline CabDecompressorT<T>::CabDecompressorT()
{
	this->fdi_handle = ::FDICreate(T::onAlloc, T::onFree, T::onOpen,
		T::onRead, T::onWrite, T::onClose, T::onSeek, 0, &this->err_code);
	if (this->fdi_handle == NULL) {
		throw FDIError(this->err_code);
	}
}

template <class T>
inline CabDecompressorT<T>::~CabDecompressorT()
{
	::FDIDestroy(this->fdi_handle);
}

template <class T>
inline bool CabDecompressorT<T>::isCabinet(boost::filesystem::wpath const &file_path)
{
	size_t const n = MyLib::convertUTF16ToUTF8(file_path.string().c_str(), NULL);
	boost::scoped_array<char> buf(new char[n]);
	MyLib::convertUTF16ToUTF8(file_path.string().c_str(), buf.get());
	intptr_t hf = T::onOpen(
		buf.get(), _O_BINARY | _O_RDONLY | _O_SEQUENTIAL, 0);
	FDICABINETINFO cab_info;
	bool result = (::FDIIsCabinet(this->fdi_handle, hf, &cab_info) != 0);
	T::onClose(hf);
	return result;
}

template <class T>
inline FDICABINETINFO CabDecompressorT<T>::getCabFileInfo(boost::filesystem::wpath const &file_path)
{
	FDICABINETINFO result;
	size_t const n = MyLib::convertUTF16ToUTF8(file_path.string().c_str(), NULL);
	boost::scoped_array<char> buf(new char[n]);
	MyLib::convertUTF16ToUTF8(file_path.string().c_str(), buf.get());
	intptr_t hf = T::onOpen(
		buf.get(), _O_BINARY | _O_RDONLY | _O_SEQUENTIAL, 0);
	if (!::FDIIsCabinet(this->fdi_handle, hf, &result)) {
		T::onClose(hf);
		throw std::runtime_error("not cabinet");
	}
	T::onClose(hf);
	return result;
}

template <class T>
inline bool CabDecompressorT<T>::extractCabFile(
	boost::filesystem::wpath const &cab_file_path, boost::filesystem::wpath const &target_dir_path)
{
	namespace fs = boost::filesystem;

	std::wstring const cab_file_path_str = cab_file_path.string();
	boost::scoped_array<wchar_t> cab_file_path_buf(new wchar_t[cab_file_path_str.length() + 1]);
	::lstrcpy(cab_file_path_buf.get(), cab_file_path_str.c_str());

	// Cabinett@C
	std::wstring const file_title = cab_file_path.filename();
	size_t const file_title_count = convertUTF16ToUTF8(file_title.c_str(), NULL);
	boost::scoped_array<char> title_buf(new char[file_title_count]);
	convertUTF16ToUTF8(file_title.c_str(), title_buf.get());
	// Cabinett@CtH_̃pX
	fs::wpath cab_dir_path = cab_file_path;
	cab_dir_path.remove_filename();
	size_t const dir_count = convertUTF16ToUTF8(cab_dir_path.string().c_str(), NULL);
	boost::scoped_array<char> dir_buf(new char[dir_count]);
	convertUTF16ToUTF8(cab_dir_path.string().c_str(), dir_buf.get());
	// WJ̃tH_̃pX
	fs::create_directories(target_dir_path);
	this->target_dir_path = target_dir_path;
	// WJ
	return ::FDICopy(this->fdi_handle, title_buf.get(),
		dir_buf.get(), 0, T::onNotify, NULL, this) != 0;
}

template <class T>
inline void CabDecompressorT<T>::onCabinetInfo(
	const char * /*next_cabinet*/,
	const char * /*next_disk*/,
	const char * /*cabinet_path*/,
	unsigned short /*cabinet_set_id*/,
	unsigned short /*cabinet_number_in_set*/)
{
}

template <class T>
inline void CabDecompressorT<T>::onPartialFile(
	const char * /*continued_file_name*/,
	const char * /*file_start_cabinet_name*/,
	const char * /*file_start_disk_name*/)
{
}

template <class T>
inline typename CabDecompressorT<T>::CopyFileFlags CabDecompressorT<T>::onCopyFile(
	std::wstring const & /*file_name_in_cab*/,
	size_t /*uncompressed_file_size*/,
	FILETIME /*file_time*/,
	unsigned short /*attribs*/)
{
	return EXTRACT;
}

template <class T>
inline bool CabDecompressorT<T>::onCloseFileInfo(
	const char *file_name_in_cab, intptr_t file_handle, FILETIME file_time,
	unsigned short attribs, bool executes_after_extract)
{
	return true;
}

/////////////////////////////////////////////////////////////////////////

template <class T>
inline void * DIAMONDAPI CabDecompressorT<T>::onAlloc(unsigned long cb)
{
	return operator new(cb);
}

template <class T>
inline void DIAMONDAPI CabDecompressorT<T>::onFree(void *pv)
{
	operator delete(pv);
}

template <class T>
inline intptr_t DIAMONDAPI CabDecompressorT<T>::onOpen(char *pszFile, int oflag, int pmode)
{
	int result = -1;
	// pszFileUTF-8
	size_t const n = MyLib::convertUTF8ToUTF16(pszFile, NULL);
	boost::scoped_array<wchar_t> buf(new wchar_t[n]);
	MyLib::convertUTF8ToUTF16(pszFile, buf.get());
#	if defined(_MSC_VER) && (_MSC_VER >= 1400) /* VC 2005 */
	::_wsopen_s(&result, buf.get(), oflag, _SH_DENYNO, pmode);
#	else	/* others */
	result = ::_wsopen(buf.get(), oflag, _SH_DENYNO, pmode);
#	endif
	return result;
}

template <class T>
inline int DIAMONDAPI CabDecompressorT<T>::onClose(intptr_t hf)
{
	return ::_close(static_cast<int>(hf));
}

template <class T>
inline unsigned int DIAMONDAPI CabDecompressorT<T>::onRead(intptr_t hf, void *pv, unsigned int cb)
{
	return ::_read(static_cast<int>(hf), pv, cb);
}

template <class T>
inline unsigned int DIAMONDAPI CabDecompressorT<T>::onWrite(intptr_t hf, void *pv, unsigned int cb)
{
	return ::_write(static_cast<int>(hf), pv, cb);
}

template <class T>
inline long DIAMONDAPI CabDecompressorT<T>::onSeek(intptr_t hf, long dist, int seektype)
{
	return ::_lseek(static_cast<int>(hf), dist, seektype);
}

template <class T>
inline intptr_t DIAMONDAPI CabDecompressorT<T>::onNotify(FDINOTIFICATIONTYPE fdint, FDINOTIFICATION *pfdin)
{
	namespace fs = boost::filesystem;
	T *instance = reinterpret_cast<T*>(pfdin->pv);

	switch (fdint) {
	case fdintCABINET_INFO: // general information about the cabinet
		//	printf(
		//		"fdintCABINET_INFO\n"
		//		"  next cabinet     = %s\n"
		//		"  next disk        = %s\n"
		//		"  cabinet path     = %s\n"
		//		"  cabinet set ID   = %d\n"
		//		"  cabinet # in set = %d (zero based)\n"
		//		"\n",
		//		pfdin->psz1,
		//		pfdin->psz2,
		//		pfdin->psz3,
		//		pfdin->setID,
		//		pfdin->iCabinet
		//	);
		instance->onCabinetInfo(pfdin->psz1, pfdin->psz2, pfdin->psz3,
			pfdin->setID, pfdin->iCabinet);
		return 0;
	case fdintCOPY_FILE:	// file to be copied
		{
			FILETIME ft;
			::DosDateTimeToFileTime(pfdin->date, pfdin->time, &ft);
			FILETIME lft;
			::FileTimeToLocalFileTime(&ft, &lft);
			std::wstring const file_name =
				((pfdin->attribs & _A_NAME_IS_UTF) != 0 ?
				convertFromUTF8(pfdin->psz1) : convertFromSJIS(pfdin->psz1));
			enum CopyFileFlags flag = instance->onCopyFile(
				file_name, pfdin->cb, lft, pfdin->attribs);
			if (flag == EXTRACT) {
				fs::wpath const extract_file_path = instance->target_dir_path/file_name;
				fs::wpath const parent_dir_path = extract_file_path.parent_path();
				if (!fs::exists(parent_dir_path)) {
					fs::create_directories(parent_dir_path);
				}
#if defined(_MSC_VER) && (_MSC_VER >= 1400) /* VC 2005 */
				int handle;
				::_tsopen_s(&handle,
					extract_file_path.string().c_str(),
					_O_BINARY | _O_CREAT | _O_WRONLY | _O_SEQUENTIAL,
					_SH_DENYNO, _S_IREAD | _S_IWRITE);
				return handle;
#else
				return ::_tsopen(
					extract_file_path.string().c_str(),
					_O_BINARY | _O_CREAT | _O_WRONLY | _O_SEQUENTIAL,
					_SH_DENYNO, _S_IREAD | _S_IWRITE);
#endif
			} else if (flag == SKIP) {
				return 0;
			} else {
				return -1;
			}
		}
	case fdintPARTIAL_FILE: // first file in cabinet is continuation
		//	printf(
		//		"fdintPARTIAL_FILE\n"
		//		"   name of continued file            = %s\n"
		//		"   name of cabinet where file starts = %s\n"
		//		"   name of disk where file starts    = %s\n",
		//		pfdin->psz1,
		//		pfdin->psz2,
		//		pfdin->psz3
		//	);
		instance->onPartialFile(pfdin->psz1, pfdin->psz2, pfdin->psz3);
		return 0;
	case fdintCLOSE_FILE_INFO:	// close the file, set relevant info
		{
			::_close(static_cast<int>(pfdin->hf));
			std::wstring const file_name =
				((pfdin->attribs & _A_NAME_IS_UTF) != 0 ?
				convertFromUTF8(pfdin->psz1) : convertFromSJIS(pfdin->psz1));
			boost::filesystem::wpath const file_path = instance->target_dir_path/file_name;
			/*
			 * Set date/time
			 *
			 * Need Win32 type handle for to set date/time
			 */
			HANDLE handle = CreateFile(
				file_path.string().c_str(),
				GENERIC_READ | GENERIC_WRITE,
				FILE_SHARE_READ,
				NULL,
				OPEN_EXISTING,
				FILE_ATTRIBUTE_NORMAL,
				NULL);
			if (handle != INVALID_HANDLE_VALUE) {
				FILETIME    datetime;
				if (::DosDateTimeToFileTime(pfdin->date, pfdin->time, &datetime)) {
					FILETIME    local_filetime;
					if (::LocalFileTimeToFileTime(&datetime, &local_filetime)) {
						::SetFileTime(handle, &local_filetime, NULL, &local_filetime);
					}
				}
				::CloseHandle(handle);
			}
			/*
			 * Mask out attribute bits other than readonly,
			 * hidden, system, and archive, since the other
			 * attribute bits are reserved for use by
			 * the cabinet format.
			 */
			DWORD attribs = pfdin->attribs & (_A_RDONLY | _A_HIDDEN | _A_SYSTEM | _A_ARCH);
			::SetFileAttributes(file_path.string().c_str(), attribs);
			return TRUE;
		}
	case fdintNEXT_CABINET:	// file continued to next cabinet
	//	printf(
	//		"fdintNEXT_CABINET\n"
	//		"   name of next cabinet where file continued = %s\n"
	//		"   name of next disk where file continued    = %s\n"
	//		"   cabinet path name                         = %s\n"
	//		"\n",
	//		pfdin->psz1,
	//		pfdin->psz2,
	//		pfdin->psz3
	//		pdfin->fdie
	//	);
		return 0;
	}
	return 0;
}

//=======================================================================
// CabDecompressor
//=======================================================================
class CabDecompressor :
	public CabDecompressorT<CabDecompressor>
{
};
};	// namespace MyLib

#endif	// CAB_DECOMPRESSOR_H__15B97EE0_A438_472F_A63F_C3B89F55C51D__INCLUDED
