#ifndef CATALOG_FILE_INFO_H__8E601E83_F294_4CEA_A583_10F0A0335699__INCLUDED
#define CATALOG_FILE_INFO_H__8E601E83_F294_4CEA_A583_10F0A0335699__INCLUDED
// M$ compatible compilers support #pragma once
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

/////////////////////////////////////////////////////////////////////////////
// CatalogFileInfo.h
//	H20(2008)/05/09 쐬
// Classes in this file:
//	CatalogFileInfo
/////////////////////////////////////////////////////////////////////////////

namespace MyLib {
	//=======================================================================
	// CatalogFileInfo
	//	catt@C擾
	// QlF
	//	How To Get Information from Authenticode Signed Executables
	//	http://support.microsoft.com/kb/323809/en-us
	// gpF
	//	MyLib::CatalogFileInfo cat_info(_T("D:\\ext64\\si3112r.cat"));
	//	std::wcout.imbue(std::locale(""));
	//	std::wcout << L"ҁF" << cat_info.getSubjectName() << std::endl;
	//	FILETIME ft = cat_info.getTimeStamp();
	//	FILETIME lft;
	//	::FileTimeToLocalFileTime(&ft, &lft);
	//	SYSTEMTIME st;
	//	::FileTimeToSystemTime(&lft, &st);
	//	std::wcout << L"F" << st.wYear << L'/' << st.wMonth << L'/' << st.wDay << L' ';
	//	std::wcout << st.wHour << L':' << st.wMinute << L':' << st.wSecond << std::endl;
	//=======================================================================
	class CatalogFileInfo
	{
		enum { ENCODING = X509_ASN_ENCODING | PKCS_7_ASN_ENCODING };

	public:
		// RXgN^
		CatalogFileInfo(const String& file_path)
			: crypt_msg(NULL)
			, cert_store(NULL)
			, signer_info_ptr(NULL)
		{
			// Get message handle and store handle from the signed file.
			DWORD encoding_type, content_type, format_type;
			bool ret1 = (::CryptQueryObject(CERT_QUERY_OBJECT_FILE,
				file_path.c_str(), CERT_QUERY_CONTENT_FLAG_PKCS7_SIGNED_EMBED,
				CERT_QUERY_FORMAT_FLAG_BINARY, 0, &encoding_type, &content_type,
				&format_type, &this->cert_store, &this->crypt_msg, NULL) != 0);
			if (!ret1) {
				throw MY_WIN32_EXCEPTION_LAST_ERROR;
			}
			// Get signer information size.
			DWORD signer_info_size;
			bool ret2 = (::CryptMsgGetParam(crypt_msg, CMSG_SIGNER_INFO_PARAM,
				0, NULL, &signer_info_size) != 0);
			if (!ret2) {
				this->close();
				throw MY_WIN32_EXCEPTION_LAST_ERROR;
			}
			// Get signer information.
			this->signer_info_ptr =
				reinterpret_cast<CMSG_SIGNER_INFO*>(new char[signer_info_size]);
			bool ret3 = (::CryptMsgGetParam(crypt_msg, CMSG_SIGNER_INFO_PARAM, 0,
				this->signer_info_ptr, &signer_info_size) != 0);
			if (!ret3) {
				this->close();
				throw MY_WIN32_EXCEPTION_LAST_ERROR;
			}
		}

		// fXgN^
		~CatalogFileInfo()
		{
			this->close();
		}

		// n
		void close()
		{
			if (this->crypt_msg != NULL) {
				::CryptMsgClose(this->crypt_msg);
				this->crypt_msg = NULL;
			}
			if (this->cert_store != NULL) {
				::CertCloseStore(this->cert_store, 0);
				this->cert_store = NULL;
			}
			if (this->signer_info_ptr != NULL) {
				delete[] reinterpret_cast<char*>(this->signer_info_ptr);
				this->signer_info_ptr = NULL;
			}
		}

		// (foCX}l[W̃vpeB)fW^҂擾
		String getSubjectName()
		{
			// Search for the signer certificate in the temporary 
			// certificate store.
			CERT_INFO info;
			info.Issuer = this->signer_info_ptr->Issuer;
			info.SerialNumber = this->signer_info_ptr->SerialNumber;
			const CERT_CONTEXT *context_ptr = ::CertFindCertificateInStore(
				cert_store, ENCODING, 0, CERT_FIND_SUBJECT_CERT, &info, NULL);
			if (context_ptr != NULL) {
				// Get subject name size
				const DWORD buf_count = ::CertGetNameString(context_ptr,
					CERT_NAME_SIMPLE_DISPLAY_TYPE, 0, NULL, NULL, 0);
				if (buf_count > 0) {
					// Get subject name
					MyLib::TemporaryBuffer<TCHAR> buf(buf_count);
					const DWORD	ret = ::CertGetNameString(context_ptr,
						CERT_NAME_SIMPLE_DISPLAY_TYPE, 0, NULL,
						buf.getPointer(), buf_count);
					if (ret > 0) {
						::CertFreeCertificateContext(context_ptr);
						return buf.getPointer();
					}
				}
				::CertFreeCertificateContext(context_ptr);
			}
			throw MY_WIN32_EXCEPTION_LAST_ERROR;
		}

		// 擾
		FILETIME getTimeStamp()
		{
			CMSG_SIGNER_INFO *counter_ptr =
				this->getTimeStampSignerInfo(this->signer_info_ptr);
			// Get the timestamp certificate signerinfo structure.
			if (counter_ptr != NULL) {
				// Search for Timestamp certificate in the temporary
				// certificate store.
				CERT_INFO info;
				info.Issuer = counter_ptr->Issuer;
				info.SerialNumber = counter_ptr->SerialNumber;
				const CERT_CONTEXT *context_ptr = ::CertFindCertificateInStore(
					this->cert_store, ENCODING, 0, CERT_FIND_SUBJECT_CERT, &info, NULL);
				if (context_ptr != NULL) {
					// Loop through authenticated attributes and find
					// szOID_RSA_signingTime OID.
					FILETIME result = {0};
					for (unsigned int n = 0; n < counter_ptr->AuthAttrs.cAttr; ++n) {
						const int comp_result = ::strcmp(szOID_RSA_signingTime,
							counter_ptr->AuthAttrs.rgAttr[n].pszObjId);
						if (comp_result == 0) {
							// Decode and get FILETIME structure.
							DWORD buf_size = sizeof(FILETIME);
							bool ret1 = (::CryptDecodeObject(ENCODING,
								szOID_RSA_signingTime,
								counter_ptr->AuthAttrs.rgAttr[n].rgValue[0].pbData,
								counter_ptr->AuthAttrs.rgAttr[n].rgValue[0].cbData,
								0, &result, &buf_size) != 0);
							if (ret1) {
								break;
							} else {
								delete[] reinterpret_cast<char*>(counter_ptr);
								::CertFreeCertificateContext(context_ptr);
								throw MY_WIN32_EXCEPTION_LAST_ERROR;
							}
						}
					}	// for
					delete[] reinterpret_cast<char*>(counter_ptr);
					::CertFreeCertificateContext(context_ptr);
					return result;
				}
				delete[] reinterpret_cast<char*>(counter_ptr);
			}
			throw MY_WIN32_EXCEPTION_LAST_ERROR;
		}

	private:
		HCRYPTMSG crypt_msg;
		HCERTSTORE cert_store;
		CMSG_SIGNER_INFO *signer_info_ptr;

		CMSG_SIGNER_INFO *getTimeStampSignerInfo(
			const CMSG_SIGNER_INFO *signer_info_ptr)
		{
			// Loop through unathenticated attributes for
			// szOID_RSA_counterSign OID.
			for (unsigned int n = 0; n < signer_info_ptr->UnauthAttrs.cAttr; ++n) {
				const int comp_result = std::strcmp(szOID_RSA_counterSign,
					signer_info_ptr->UnauthAttrs.rgAttr[n].pszObjId);
				if (comp_result == 0) {
					// Get size of CMSG_SIGNER_INFO structure.
					DWORD buf_size;
					const bool ret1 = (::CryptDecodeObject(ENCODING,
						PKCS7_SIGNER_INFO,
						signer_info_ptr->UnauthAttrs.rgAttr[n].rgValue[0].pbData,
						signer_info_ptr->UnauthAttrs.rgAttr[n].rgValue[0].cbData,
						0, NULL, &buf_size) != 0);
					if (!ret1) {
						//_tprintf(_T("CryptDecodeObject failed with %x\n"),
						throw MY_WIN32_EXCEPTION_LAST_ERROR;
					}

					// Allocate memory for CMSG_SIGNER_INFO.
					CMSG_SIGNER_INFO *counter_signer_info_ptr =
						reinterpret_cast<CMSG_SIGNER_INFO*>(new char[buf_size]);

					// Decode and get CMSG_SIGNER_INFO structure
					// for timestamp certificate.
					const bool ret2 = (::CryptDecodeObject(ENCODING,
						PKCS7_SIGNER_INFO,
						signer_info_ptr->UnauthAttrs.rgAttr[n].rgValue[0].pbData,
						signer_info_ptr->UnauthAttrs.rgAttr[n].rgValue[0].cbData,
						0, counter_signer_info_ptr, &buf_size) != 0);
					if (!ret2) {
						//_tprintf(_T("CryptDecodeObject failed with %x\n"),
						delete[] reinterpret_cast<char*>(counter_signer_info_ptr);
						throw MY_WIN32_EXCEPTION_LAST_ERROR;
					}
					return counter_signer_info_ptr; // Break from for loop.
				}
			}
			return NULL;
		}
	}; // class CatalogFileInfo
}; // namespace MyLib

#endif // CATALOG_FILE_INFO_H__8E601E83_F294_4CEA_A583_10F0A0335699__INCLUDED
