#pragma once

namespace MyLib {
	static std::map<HWND, int> caret_pos_map;

	//==========================================================================
	// EditCtrlStreamBuf
	//	GfBbgRg[o̓Xg[obt@
	//==========================================================================
	template <class E, class T = std::char_traits<E> >
	class EditCtrlStreamBuf :
		public std::basic_streambuf<E, T>
	{
		CEdit edit_ctrl; //GfBbgRg[
		std::basic_string<char_type> m_strBuffer; //obt@

	public:
		EditCtrlStreamBuf(HWND handle)
			: edit_ctrl(handle)
		{
			if (caret_pos_map.find(handle) == caret_pos_map.end()) {
				caret_pos_map[handle] = edit_ctrl.GetWindowTextLength();
			}
		}

		virtual ~EditCtrlStreamBuf()
		{
		}

	protected:
		virtual int sync()
		{
			std::basic_string<char_type> temp_buf;
			const size_t buffer_length = this->m_strBuffer.length();
			for (size_t i = 0; i < buffer_length; ++i) {
				TCHAR c = this->m_strBuffer[i];
				switch (c) {
				case '\b':	// Back Space
					{
						if (!temp_buf.empty()) {
							this->appendText(temp_buf);
							temp_buf.erase();
						}
						const int caret_pos = caret_pos_map[this->edit_ctrl.m_hWnd];
						if (caret_pos > 0) {
							this->edit_ctrl.SetSel(caret_pos - 1, caret_pos);
							this->edit_ctrl.ReplaceSel(_T(""));

							caret_pos_map[this->edit_ctrl.m_hWnd] = caret_pos - 1;
						}
					}
					break;
				case '\r':	// CR
					if (!temp_buf.empty()) {
						this->appendText(temp_buf);
						temp_buf.erase();
					}
					// CR-LF
					{
						int caret_pos = caret_pos_map[this->edit_ctrl.m_hWnd];
						int line_num = this->edit_ctrl.LineFromChar(caret_pos);
						int line_start_pos = this->edit_ctrl.LineIndex(line_num);
						this->edit_ctrl.SetSel(line_start_pos, line_start_pos);

						caret_pos_map[this->edit_ctrl.m_hWnd] = line_start_pos;
					}
					break;
				case '\n':	// LF
					if (!temp_buf.empty()) {
						this->appendText(temp_buf);
						temp_buf.erase();
					}
					{
						int caret_pos = caret_pos_map[this->edit_ctrl.m_hWnd];
						int line_num = this->edit_ctrl.LineFromChar(caret_pos);
						int line_start_pos = this->edit_ctrl.LineIndex(line_num);
						int line_len = this->edit_ctrl.LineLength(line_start_pos);
						int line_end_pos = line_start_pos + line_len;
						this->edit_ctrl.SetSel(line_end_pos, line_end_pos);
						this->edit_ctrl.ReplaceSel(_T("\r\n"));

						caret_pos_map[this->edit_ctrl.m_hWnd] = line_end_pos + 2;
					}
					break;
				default:
					temp_buf.push_back(c);
					break;
				}
			}
			m_strBuffer.erase();
			if (!temp_buf.empty()) {
				this->appendText(temp_buf);
			}
			return traits_type::eof();
		}

		virtual int_type overflow(int_type c = traits_type::eof())
		{
			if (traits_type::eq_int_type(traits_type::eof(), c)) {
				return traits_type::eof();
			}
			m_strBuffer += traits_type::to_char_type(c);
			return traits_type::not_eof(c);
		}

	private:
		void appendText(const std::basic_string<char_type>& str)
		{
			int caret_pos = caret_pos_map[this->edit_ctrl.m_hWnd];
			int line_num = this->edit_ctrl.LineFromChar(caret_pos);
			int line_start_pos = this->edit_ctrl.LineIndex(line_num);
			int line_len = this->edit_ctrl.LineLength(line_start_pos);
			this->edit_ctrl.SetSel(caret_pos, line_start_pos + line_len);
			this->edit_ctrl.ReplaceSel(str.c_str());

			caret_pos_map[this->edit_ctrl.m_hWnd] = caret_pos + str.length();
		}
	};

	//==========================================================================
	// BasicEditCtrlStream
	//	GfBbgRg[o̓Xg[
	//==========================================================================
	template<class E, class T = std::char_traits<E> >
	class BasicEditCtrlStream :
		public std::basic_iostream<E, T>
	{
	public:
		BasicEditCtrlStream(HWND handle)
			: basic_iostream<E, T>(new EditCtrlStreamBuf<E, T>(handle))
		{
		}

		virtual ~BasicEditCtrlStream()
		{
		}
	};
	typedef BasicEditCtrlStream<TCHAR, std::char_traits<TCHAR> > EditCtrlStream;
};
