﻿#ifndef NUMERIC_EDIT_H__08FFB2E9_98DC_473B_9A6B_4358004A0725__INCLUDED
#define NUMERIC_EDIT_H__08FFB2E9_98DC_473B_9A6B_4358004A0725__INCLUDED
// MS compatible compilers support #pragma once
#if defined(_MSC_VER) && (_MSC_VER >= 1020 )
#pragma once
#endif

/////////////////////////////////////////////////////////////////////////////
// NumericEditCtrl
//  数値のみを入力できるテキストボックス
//  リソースプロパティの「Number」では
//	マイナス記号や小数点も入力できないため
//  H19(2007)/02/14 作成
//  H19(2007)/12/13 cppファイルを統合
//	H19(2007)/12/14 二重include防止処理を変更
//	H20(2008)/12/05 小数・負数の許可・禁止を変更できるようにした
//	                Ctrl+V などのショートカットキーが使えなかったのを修正
//	                フォーカスを持つときは IME を無効にするようにした
/////////////////////////////////////////////////////////////////////////////

#pragma comment(lib, "imm32.lib")

class NumericEditCtrl
	: public CWindowImpl<NumericEditCtrl, CEdit>
{
	bool	m_bIsFloat;		//true:小数点数 / false:整数
	bool	m_bIsSigned;	//true:符号あり / false:符号なし
	HIMC h_imc_;

	struct LocaleInfo
	{
		TCHAR decimal_symbol;
		TCHAR negation_symbol;
		LocaleInfo()
			: decimal_symbol(_T('.'))   // reasonable default!
			, negation_symbol(_T('-'))  // reasonable default!
		{
			//
			// you MAY initialize these variables to 
			// any other value you like, for e.g., zero
			//
			TCHAR buf[2];
			/**
			 * Retrieve the symbol used as the decimal separator
			 */
			int ret = ::GetLocaleInfo(LOCALE_USER_DEFAULT, 
			                          LOCALE_SDECIMAL, 
			                          buf,
			                          _countof(buf));
			if (ret) {
				this->decimal_symbol = buf[0];
			}
			/**
			 * retrieve the symbol used as the negative sign
			 */
			ret = ::GetLocaleInfo(LOCALE_USER_DEFAULT, 
			                      LOCALE_SNEGATIVESIGN, 
			                      buf,
			                      _countof(buf));
			if (ret) {
				this->negation_symbol = buf[0];
			}
		}
	};

	static LocaleInfo const default_user_locale;

public:
	DECLARE_WND_SUPERCLASS(NULL, CEdit::GetWndClassName());

	// コンストラクション
	NumericEditCtrl(bool uses_signed = true, bool uses_float = true) :
		m_bIsSigned(uses_signed),
		m_bIsFloat(uses_float),
		h_imc_(NULL)
	{
	}

	// 負数の許可
	void enableSigned() { this->m_bIsSigned = true; }
	// 小数の許可
	void enableFloat() { this->m_bIsFloat = true; }
	// 負数の禁止
	void disableSigned() { this->m_bIsSigned = false; }
	// 小数の禁止
	void disableFloat() { this->m_bIsFloat = false; }

    // メッセージ マップ
	BEGIN_MSG_MAP_EX(NumericEditCtrl)
		MSG_WM_CHAR(OnChar)
		MSG_WM_PASTE(OnPaste)
		REFLECTED_COMMAND_CODE_HANDLER_EX(EN_SETFOCUS, OnSetFocus)
		REFLECTED_COMMAND_CODE_HANDLER_EX(EN_KILLFOCUS, OnKillFocus)
		DEFAULT_REFLECTION_HANDLER()
    END_MSG_MAP_EX()

	void OnChar(TCHAR nChar, unsigned int nRepCnt, unsigned int nFlags);
	void OnPaste();
	LRESULT OnSetFocus(unsigned int uNotifyCode, int nID, HWND wndCtl);
	LRESULT OnKillFocus(unsigned int uNotifyCode, int nID, HWND wndCtl);
};

#endif	// NUMERIC_EDIT_H__08FFB2E9_98DC_473B_9A6B_4358004A0725__INCLUDED
